History { 
	classvar <>forwardFunc, <date, <sortDate, <time0;
	classvar <lines, <lineShorts;
	classvar <>title = "History repeats";
	classvar <w, <bounds, <doc, pop, startbut;
	classvar <docFlag = \sameDoc, <hasMovedOn = true;
	classvar <>verbose = false, <>recordLocally = true;
	
	*initClass { 
		bounds = Rect(0, 0, 300, 200); 
		this.clear;
	}
	*clear { 
		lines = List(1024);
		lineShorts = List(1024); 
	}
			// replace (\n ..." with (... ))!!!
	*start { 
		var interp = thisProcess.interpreter;
		interp.codeDump = { |str, val, func| 
			if ( func.notNil 
					and: { str.notEmpty } 
					and: { str != "\n" } 
					and: { str.keep(7) != "History" }) 
				{ 
					if (verbose, { [str, val, func].postcs });
					if (recordLocally, { this.enter(interp.cmdLine) });
					forwardFunc.value(str, val, func);
				}
		};
		hasMovedOn = true;
	}
	*stop { 
		thisProcess.interpreter.codeDump = nil; 
		hasMovedOn = true;
	}

	*localOn { recordLocally = true }
	*localOff { recordLocally = false }
	
	*network { }
	
	*makeWin { 
		var closebut, v, f, font;
		font = Font("Osaka", 9);
		w = SCWindow("History", bounds).front;
	//	w.view.decorator = FlowLayout(w, w.bounds.moveTo(0,0), 0@0, 0@0);
		
		startbut = SCButton(w, Rect(0, 0, 80, 20))
			.states_([["start"], ["stop"]])
			.action_({ |btn|
				switch(btn.value, 
					0, { this.stop }, 
					1, { this.start }
				);
			});

		pop = SCPopUpMenu(w, Rect(84,0,80,20))
			.items_([\sameDoc, \newDoc])
			.value_([\sameDoc, \newDoc].indexOf(docFlag))
			.action_({ |pop| docFlag = [\sameDoc, \newDoc][pop.value] });
		
		closebut = SCButton(w, Rect(168, 0, 80, 20))
			.states_([["closeAll"]]).action_({ 
				Document.allDocuments.do { |doc| if (doc.title == title) { doc.close } };
			});
		
	//	w.view.decorator.nextLine; 
		v = SCListView(w,bounds.moveTo(0, 20))
			.font_(font)
			.items_([]).resize_(5).background_(Color.grey(0.62))
			.action_({ |entry| 
				var val = entry.value; 
				this.postDoc(val, docFlag) 
			})
			.enterKeyAction_({ History.doc.front });
		hasMovedOn = true;
		
		SkipJack({ 
			if (hasMovedOn) { 
				v.items_(lineShorts.array).value_(0);
				if (this.hasMovedOn, { startbut.value_(0) });
				hasMovedOn = false;
			};
		}, 1, { w.isClosed }, "histwin");
		
		^w;
	}
	*postDoc { |index, str| 
		var line = lines[index]; 
		if (line.isNil) { "no history yet!".inform; ^this };
		this.findDoc; 
		doc.string_(line[2]).front;
		w.front;
	}
	*findDoc {
		if (doc.isNil or: { Document.allDocuments.includes(doc).not } or: { pop.value == 1 }) { 
			doc = Document(title).bounds_(Rect(300, 500, 300, 100));
		};	
	} 
	
	*enter { |lineStr, id=\me|
		var dateNow, now; 
		if (lineStr.isEmpty) { ^this };	// nothing worth remembering
		
		if (lines.isEmpty) { 			// get startTime if first entry
			dateNow = Date.getDate;
			sortDate = dateNow.asSortableString;
			date = dateNow.asString;
			time0 = Main.elapsedTime;
		}; 
		hasMovedOn = true;
		now = (Main.elapsedTime - time0);
//		lineStr.postcs;
//		if (lineStr.first == $\n) { lineStr.drop(1) };
//		if (lineStr.first == $\n) { lineStr.drop(-1) };
//		lineStr.postcs;
		
		lines.addFirst([ now, id, lineStr ]);
		lineShorts.addFirst( "" + now.round(1) + "-" + id + "-" + lineStr.keep(50) );
	}
	
	*removeAt { |index| 
		if (index.isKindOf(Collection)) { index.do (this.removeAt(_)); ^this };
		
		if (index < lines.size) { 		// ignore out of range indices, keep lists synced.
			[lines, lineShorts].do(_.removeAt(index));		};
		hasMovedOn = true;
	}
	*keep { |num| [lines, lineShorts].do(_.keep(num)); hasMovedOn = true }
	*drop { |num| [lines, lineShorts].do(_.drop(num)); hasMovedOn = true }
	
	*saveCS { |path= "~/Desktop/history.sc"| 
		var file; 
		file = File(path.standardizePath,"w");
		file.write(this.lines.asCompileString);
		file.close;
	}
	*loadCS { |path= "~/Desktop/history.sc"| 
		
	}
}