/* Wrapper class for Node's to be played within client side group*/
ServerNode : ClientNode {
	var <>args;			// a flat array consisting of ControlValues and (Symbol, value) pairs
						// a ControlValue will map to its name unless preceded by a Symbol
	var <>node;

	var <>onEnd;			//function runs when node is freed.
	
	play {
		super.play;
		if (onEnd.notNil) {
			OSCpathResponder(node.server.addr, ['/n_end', node.nodeID],
				{ arg time, resp, msg; 
					defer({onEnd.value; nil })	
				}
			).add;
		}
	}
	replay {
		if(node.notNil, { 
			node.server.sendBundle(node.server.latency, ["/n_set", node.nodeID, \gate, 0]); 
			this.play
		});
	}

	stop { arg sendFlag = true;	
		node.free(sendFlag);
		node = nil;
		super.stop
	 }
	
	pause {  arg sendFlag = true;
	 	if (sendFlag, {node.server.sendBundle(node.server.latency,[12,node.nodeID,0]) });  
		super.pause;
	 }
	
	resume {  arg sendFlag = true;
	 	if (sendFlag, {node.server.sendBundle(node.server.latency,[12,node.nodeID,1]) });  
		super.resume;
	}
	
	asNodeID { ^node.nodeID }
	asTarget { ^node }
}	

ServerSynth : ServerNode {
	*new { arg name, args, target, addAction;
		^super.newCopyArgs( 0, nil, target, addAction ? \addToTail,name, args ? []);
	}	
	
	play { arg sendFlag = true;
		 var t, msg;
		t = target.asTarget;
			node = Synth.basicNew(name, t.server);
			msg = node.newMsg(	t, args.value.connectToNode(node.server,node.nodeID), addAction);
			super.play;
			if(sendFlag, {
				t.server.sendBundle(t.server.latency, msg);
			}, {
				^msg;
			});
	}
	
}


ServerGroup : ServerNode {
	var <>childDefinitions;			// TaskPlayer, PatternPlayer, SynthPlayer, GroupPlayer
	var <>children;				// playing attachments

	*new { arg args = #[], target, addAction;
		^super.newCopyArgs(0, nil, target, addAction ? \addToHead, "agroup", args);
	}

	play { arg sendFlag = true;
		var t, msg, argArray;
		t = target.asTarget;
			node = Group.basicNew(t.server);
			msg = msg.add(node.newMsg(t, addAction));

			childDefinitions.do({ arg c; msg = msg.add(c.play(false)) });
			argArray = args.value.connectToNode(node.server,node.nodeID);
			if (argArray.size != 0, { msg = msg.add(["/n_set"] ++ [argArray]) });

			node.isPlaying_(true).isRunning_(true);
			super.play;
			if (sendFlag, { t.server.listSendBundle(t.server.latency, msg) }, { ^msg});

	}

	stop { arg sendFlag = true;
		children.do({ arg c; c.stop(false) });	
		children = nil;
		super.stop(sendFlag);
	}
	
	pause { arg sendFlag = true;
		super.pause(sendFlag);
		children.do({ arg c; c.pause(false) })
	}
	
	resume { arg sendFlag = true;
		super.resume(sendFlag);
		children.do({ arg c; c.resume(false) })
	}
		
	asGroup { ^node }	
	
	add { arg obj, attach = true;  
		obj.group_(this); 
		obj.target_(this);
		if (attach, {childDefinitions = childDefinitions.add(obj) });
		if (value == 1, { obj.play });		
	}	 	
	pattern_ 	{ arg pattern, event, clock, quant, target, addAction; 
		this.add(PatPlayer(pattern, event, clock, quant, target, addAction)); 
	}
	task_ 	{ arg func, clock, quant, target, addAction; 
		this.add(TaskPlayer(func, clock, quant, target, addAction)); 
	}
	nodeProxy_ { arg func, clock, quant, target, addAction; 
		this.add(NodeProxyPlayer(func, clock, quant, target, addAction)); 
	}
	synth_ 	{ arg name, args, target, addAction; 
		this.add(ServerSynth(name, args, target, addAction)); 
	}
	synthDef_{ arg obj, args, target, addAction; 
		this.add(ServerSynthDef(obj, args, target, addAction)); 
	}
	
}


ServerSynthDef : ServerSynth {

	*new { arg obj, args, target, addAction;
		^super.new(this.makeSynthDefName(obj), args, target, addAction)
	}
	
	*makeSynthDefName { arg obj;
		var sd;
//		sd = obj.asSynthDef;
		sd = SynthDef(obj.hash.asString, obj);
		if (sd.notNil, { 
			sd.writeDefFile("synthdefs/");
			Server.set.do({ arg s; 
				if (s.serverRunning, { 
					s.sendSynthDef(sd.name, "synthdefs/") 
				});  
			});
			^sd.name;
		}, {
			^obj;
		});
	}
}

