/*
 ControlValue models a single floating point value.
 It consists of the value, a name, and a ControlSpec 
 (that determines the possible range of values).
 
 A ControlValue can be directly used within a Pattern or a Stream.
 Dependants (such as GUI objects or a Control of a Synth or Group
 running on a synthesis server) are connected to the ControlValue
 using a SimpleController linked to the symbol \synch. 
 
 ControlValue-input_(val) as val ranges from 0 to 1 the ControlValue will range
 from its minimum to maximum values. This allows external controllers, such as
 MIDI devices, to be connected to ControlValue's.
 
*/

ControlValue  : Stream {
	var <value;
	var <spec;
	var <>action;
	var <>name;

	*new {arg spec = \unipolar, default; 			
		^super.new.spec_(spec,default).name_("");
	}

	value_ { arg val;
	
		value = spec.constrain(val);
		action.value(this);
		this.changed(\synch);
	}
	
	input_	{ arg in; this.value_(spec.map(in)); }
	
	input 	{ ^spec.unmap(value) }

	asInput 	{arg val; ^spec.unmap(val) }
	
	spec_ 	{ arg s, v; 
				spec = s.asSpec; 
				this.value_(v ? spec.default); 
			}	
	sp	{ arg default= 0, lo = 0, hi=0, step = 0, warp = 'lin';
		this.spec = ControlSpec(lo,hi, warp, step, default);
	}
	db	{ arg default= 0;
		this.spec = ControlSpec(-100,20, 0,0, default);
	}
	
// Stream and Pattern support

	next { ^value }
	reset {}
	asStream { ^this }
	embedInStream { value.yield }

// connecting to a Server	
// msg = ['/n_set', nodeID, name] or ['/c_set', busIndex]
	connectToServer { arg server, msg, modifierOrCV;
		^SimpleController (this) .put(\synch, 
					{ server.sendBundle(server.latency, (msg ++ modifierOrCV.value(this)) ) });
	}

	connectToNodeProxy { arg np, name, modifierOrCV;
		^SimpleController (this) .put(\synch, { np.set(name, modifierOrCV.value(this)) });
	}
		

		
}


