// 2005 by till bovermann
//   bielefeld university

ScatterView {
	var <plot, <background;
	var <>highlightColor, <highlightItem, <>highlightSize, <>isHighlight;
	var <symbolSize, <>symbolColor, <>drawMethod = \lineTo;
	var adjustedData, specX, specY;

	*new {|parent, bounds, data, specX, specY|
		^super.new.initPlot(
			parent, bounds, data, specX, specY
		)
	}
	background_ {|color|
		background = color;
	}
	highlightItem_{|item|
		highlightItem = item.isKindOf(Collection).if({item}, {[item]});
		//this.plot.refresh;
	}
	highlightItemRel_{|val|
		this.highlightItem = (val * (adjustedData.size-1)).asInteger;
	}
	highlightRange{|start, end|
		this.highlightItem = (start .. end);
	}
	highlightRangeRel{|start, end|
		var startIndex, endIndex;
		startIndex = (start * (adjustedData.size-1)).asInteger;
		endIndex = (end * (adjustedData.size-1)).asInteger;
		this.highlightItem = (startIndex .. endIndex);
	}
	
	data_{|data|
		adjustedData = data.collect {|item|
			[specX.unmap(item[0]), specY.unmap(item[1])];
		};
		highlightItem = min(highlightItem, (adjustedData.size-1));
	}
	symbolSize_{|point|
		if(point.isKindOf(Point), {symbolSize = point}, {symbolSize = (point@point)});
	}
	resize{
		^plot.resize;
	}
	resize_{arg resize;
		plot.resize_(resize)
	}
	initPlot {
		arg parent, bounds, data, argSpecX, argSpecY;
		
		var widthSpec, heightSpec, cross;
		var possibleMethods;
		
		specX = argSpecX ? [0,1].asSpec;
		specY = argSpecY ? specX.copy;
		possibleMethods = [\fillRect, \fillOval, \strokeOval, \strokeRect];
		
		this.symbolSize = 1;
		symbolColor = symbolColor ? Color.black;
		
		background = Color.white;
		highlightColor = Color.red;
		highlightItem = 0;
		highlightSize = (2@2);
		isHighlight = false;

		this.data_(data);
		
		cross = {|rect, width, color|
			var dx, dy, extX, extY;
			dx = rect.left;
			dy = rect.top;
			extX = rect.width;
			extY = rect.height;
			Pen.use{
				color.set;
				Pen.translate(dx, dy);
				Pen.moveTo((0)@(extY*0.5));
				Pen.lineTo(extX@(extY*0.5));
				Pen.moveTo((extX*0.5)@(0));
				Pen.lineTo((extX*0.5)@(extY));
				Pen.stroke;
			};
		};
		
		plot = SCUserView.new(parent, bounds)
			.drawFunc_({|w|
				var width, height, rect, pad = 10;
				
				width =  w.bounds.width  - pad;
				height = w.bounds.height - pad;
				Pen.use{
					// clipping into the boundingbox
					Pen.moveTo((plot.bounds.left)@(plot.bounds.top));
					Pen.lineTo(((plot.bounds.left)@(plot.bounds.top)) 
							+ (plot.bounds.width@0));
					Pen.lineTo(((plot.bounds.left)@(plot.bounds.top)) 
							+ (plot.bounds.width@plot.bounds.height));
					Pen.lineTo(((plot.bounds.left)@(plot.bounds.top)) 
							+ (0@plot.bounds.height));
					Pen.lineTo((plot.bounds.left)@(plot.bounds.top));
					Pen.clip;
					
					// draw Background
					background.set;
					Pen.addRect(this.plot.bounds);
					Pen.fill;
					
					// draw data
					symbolColor.set;
					if (possibleMethods.includes(drawMethod), {
						Pen.beginPath;
						adjustedData.do{|item, i|
							rect = Rect(
								(   item[0]  * width)  - (symbolSize.x/2) + (pad/2),
								((1-item[1]) * height) - (symbolSize.y/2) + (pad/2), 
								symbolSize.x, symbolSize.y
							);
							Pen.perform(
								drawMethod,
								(Rect(plot.bounds.left, plot.bounds.top, 0, 0) + rect)
							);
						}
					},{ // else draw lines
						Pen.width = symbolSize.x;
						// move to first position;
						if (adjustedData.notNil) {
							Pen.moveTo(
								((adjustedData[0][0]  * width) + (pad/2))
								 @
								 (((1-adjustedData[0][1]) * height) + (pad/2))
							   	  + 
								  (plot.bounds.left@plot.bounds.top)
							);
						};
						// draw lines
						adjustedData.do{|item, i|
							Pen.lineTo(
								((item[0]  * width) + (pad/2))
								 @
								 (((1-item[1]) * height) + (pad/2)) 
							 	  + 
							 	  (plot.bounds.left@plot.bounds.top)
							)
						};
						if(drawMethod == \fill, {Pen.fill},{Pen.stroke});
					});
					
					// highlight datapoint
					if(isHighlight) {
						highlightItem.do{|item|
							cross.value((Rect(
								(adjustedData[item][0]  * width) 
								 - 
								 (highlightSize.x/2) + (pad/2),
								((1-adjustedData[item][1]) * height) 
								 - 
								 (highlightSize.y/2) 
								  + 
								  (pad/2), 
								highlightSize.x, highlightSize.y) 
								  + 
								  Rect(plot.bounds.left, plot.bounds.top, 0, 0)),
								symbolSize,
								highlightColor
							);
						}; // od
					};
				}; // end Pen.use
			});
	}
	canFocus_ { arg state = false;
		plot.canFocus_(state);
	}
	canFocus {
		^plot.canFocus;
	}
	visible_ { arg bool;
		plot.visible_(bool)
	}
	visible {
		^plot.visible
	}
}

ScatterView3d {
	var scatterView;
	var rotX, rotY, rotZ;
	var specX, specY, specZ;
	var data3d;
	
	*new {|parent, bounds, data, specX, specY, specZ, rotX = 0, rotY = 0, rotZ = 0|
		^super.new.init3d(parent, bounds, data, specX, specY, specZ, rotX, rotY, rotZ);
	}

	init3d {|parent, bounds, data, argspecX, argspecY, argspecZ, argrotX, argrotY, argrotZ|
		specX = argspecX ? [0, 1].asSpec;
		specY = argspecY ? specX.copy;
		specZ = argspecY ? specX.copy;

		rotX = argrotX;
		rotY = argrotY;
		rotZ = argrotZ;

		[rotX, rotY, rotZ].postln;


		// [-2.sqrt, 2.sqrt].asSpec approx. [-1.42, 1.42].asSpec
		scatterView = ScatterView(parent, bounds, [], [-1.42, 1.42].asSpec);
		this.data = data;		
	}
	data_{|data|
		data3d = data.collect {|item|
			Matrix.withFlatArray(3, 1, [
				specX.unmap(item[0]), 
				specY.unmap(item[1]), 
				specZ.unmap(item[2])
			] * 2 - 1);
		};
		scatterView.data = this.pr_project;

//		highlightItem = min(highlightItem, (adjustedData.size-1));
	}
	rotX_{|val|
		this.rot(val, rotY, rotZ);
	}
	rotY_{|val|
		this.rot(rotX, val, rotZ);
	}
	rotZ_{|val|
		this.rot(rotX, rotY, val);
	}
	rot {|rX, rY, rZ|
		rotX = rX;
		rotY = rY;
		rotZ = rZ;
		scatterView.data = this.pr_project;
	}
	drawMethod_{|method|
		scatterView.drawMethod = method;		
	}
	symbolSize_{|val|
		scatterView.symbolSize = val;		
	}
	symbolColor_{|val|
		scatterView.symbolColor = val;		
	}
	isHighlight_{|val|
		scatterView.isHighlight_(val);
	}
	highlightItem_{|item|
		scatterView.highlightItem_(item);
	}
	highlightRange{|start, end|
		scatterView.highlightRange(start, end);
	}
	highlightColor_{|color|
		scatterView.highlightColor_(color);
	}
	highlightSize_{|size|
		scatterView.highlightSize_(size);
	}
	background_{|val|
		scatterView.background = val;		
	}
	resize{
		^scatterView.resize
	}
	resize_{arg resize;
		scatterView.resize_(resize)
	}
	pr_project {
		var cx, cy, sx, sy, sz, cz, projectionMatrix;
		sx = sin(rotX);
		sy = sin(rotY);
		sz = sin(rotZ);

		cx = cos(rotX);
		cy = cos(rotY);
		cz = cos(rotZ);
			
		projectionMatrix = Matrix.with([
			[( (cy * cz) - (sx * sy * sz)), (sz.neg * cx), ((cz * sy) + (sz * sx * cy))],
			[( (cy * sz) + (cz * sx * sy)), ( cz * cx),    ((sz * sy) - (sx * cy * cz))]
		]);
	
	
		^data3d.collect{|row|
			(projectionMatrix * row).getCol(0)
		};
	}
}
