
#include "FFT_UGens.h"



//////////////////////////////////////////////////////////////////////////////////////////////////

extern "C"
{
	void load(InterfaceTable *inTable);

	void PV_AllMagsToDC_Ctor(PV_Unit *unit);
	void PV_AllMagsToDC_next(PV_Unit *unit, int inNumSamples);

	void PV_DiffAndToDC_Ctor(PV_Unit *unit);
	void PV_DiffAndToDC_next(PV_Unit *unit, int inNumSamples);



}

SCPolarBuf* ToPolarApx(SndBuf *buf)
{
	if (buf->coord == coord_Complex) {
		SCComplexBuf* p = (SCComplexBuf*)buf->data;
		int numbins = buf->samples - 2 >> 1;
		for (int i=0; i<numbins; ++i) {
			p->bin[i].ToPolarApxInPlace();
		}
		buf->coord = coord_Polar;
	}
	return (SCPolarBuf*)buf->data;
}

SCComplexBuf* ToComplexApx(SndBuf *buf)
{
	if (buf->coord == coord_Polar) {
		SCPolarBuf* p = (SCPolarBuf*)buf->data;
		int numbins = buf->samples - 2 >> 1;
		for (int i=0; i<numbins; ++i) {
			p->bin[i].ToComplexApxInPlace();
		}
		buf->coord = coord_Complex;
	}
	return (SCComplexBuf*)buf->data;
}

InterfaceTable *ft;

void init_SCComplex(InterfaceTable *inTable);

//////////////////////////////////////////////////////////////////////////////////////////////////



void PV_AllMagsToDC_next(PV_Unit *unit, int inNumSamples)
{
	PV_GET_BUF

	SCComplexBuf *p = ToComplexApx(buf);
	
	for (int i=0; i<numbins; ++i) {
		float rabs = sc_abs(p->bin[i].real);
		float iabs = sc_abs(p->bin[i].imag);
		float mag = sqrt((rabs*rabs) + (iabs*iabs));
		p->dc += mag;
		p->bin[i].real = 0.;
		p->bin[i].imag = 0.;
	}
}

void PV_AllMagsToDC_Ctor(PV_Unit *unit)
{
	SETCALC(PV_AllMagsToDC_next);
	ZOUT0(0) = ZIN0(0);
}


void PV_DiffAndToDC_next(PV_Unit *unit, int inNumSamples)
{
	PV_GET_BUF2
	
	SCComplexBuf *p = ToComplexApx(buf1);
	SCComplexBuf *q = ToComplexApx(buf2);
	
	// First the DC and nyquist. Load the differences into the DC component
	p->dc  = sc_abs(p->dc - q->dc) + sc_abs(p->nyq - q->nyq);
	p->nyq = 0.;

	for (int i=0; i<numbins; ++i) {
		float rdiff = p->bin[i].real - q->bin[i].real;
		float idiff = p->bin[i].imag - q->bin[i].imag;
		p->dc += sqrt(rdiff*rdiff + idiff*idiff);
		p->bin[i].real = 0.;
		p->bin[i].imag = 0.;
	}
}

void PV_DiffAndToDC_Ctor(PV_Unit *unit)
{
	SETCALC(PV_DiffAndToDC_next);
	ZOUT0(0) = ZIN0(0);
}

#define DefinePVUnit(name) \
	(*ft->fDefineUnit)(#name, sizeof(PV_Unit), (UnitCtorFunc)&name##_Ctor, 0, 0);



void load(InterfaceTable *inTable)
{
	ft= inTable;
	init_SCComplex(inTable);

	DefinePVUnit(PV_AllMagsToDC);
	DefinePVUnit(PV_DiffAndToDC);

}
