CtLoop { 
	classvar <>minTime = 0.002, <>post=false;
	
	var <id, <>ctlMap,
	<list, <task, <isRecording=false, then, 
	<>tempo=1, <>start=0, <>length=1, step=1, <>jitter=0.0,
	<>scaler=1, <>shift=0.0,	// assumes normalized values.
	<>rescaled=false, <nonRescalableCtls, inverter=1; 
	
	*new { arg id, ctlMap; 
		^super.newCopyArgs(id, ctlMap).init;		
	}
	
	init { 
		nonRescalableCtls = Set[];
		list = List[];
		task = Tdef(id, { 
			var dt, index, event, ctlID, val, listSize; 
			index = (start * list.size).asInteger.postln;
			
			loop({ 
				listSize = list.size;
				if (list.isEmpty)
				{ 0.1.wait; }
				{ 	
					index = (index + step + 
						(jitter.asFloat.squared.bilinrand * listSize).asInteger					)
						.wrap(
							(start * listSize).asInteger, 
							(start + length.abs * listSize).asInteger
						); 
					event = list.wrapAt(index); 
					if (post) { event.postln };
					#dt, ctlID, val = event;
							// jump over steps if speed is too high?
					max((dt / tempo), minTime).wait;
					if (rescaled and: { nonRescalableCtls.includes(ctlID).not })
						{ val = this.scaleVal(val); }; 
					 ctlMap[ctlID].value(val);
				};
			});
		});
	}
	
	recordEvent { arg ctlID, val; var now, delta; 
		if (isRecording) { 
			now = thisThread.seconds; 
			delta = now - (then ? now);
			then = now;
			list.add([ delta, ctlID, val ])
		}
	}
	
	clear { list.clear; then = nil; this.resetLoop.resetScaling; }
	startRec { isRecording = true; this.clear; this.stop; "startRec ctls".postcln; }
	stopRec {  isRecording = false; "stopRec ctls".postcln; }
	play { "playLoop".postln; this.stopRec; task.stop.play; }
	togglePlay { if (task.isPlaying, { this.stop }, { this.play }); }
	stop { "stopLoop".postln; task.stop; }
	pause { task.pause; }
	resume { task.resume; }
	isPlaying { ^task.isPlaying; }
	resetLoop { start = 0; length = 1; step = 1; tempo = 1; }

	isReversed { ^step == -1 }
	reverse { step = -1 }
	forward { step = 1 }
	flip { step = step.neg }

	scaleVal { arg val, clip=true;
		var scaledVal;
		scaledVal = val - 0.5 * scaler * inverter + 0.5 + shift;
		^if (clip, { scaledVal.clip(0, 1) }, scaledVal);
	}

	isInverse { ^inverter == -1 }
	invert { inverter = -1 }
	up { inverter = 1 }
	flipInv { inverter = inverter.neg }
	
	dontRescale { arg ... ids; nonRescalableCtls.addAll(ids.flat); }
	resetScaling { scaler = 1; shift = 0; inverter = 1; }
}