Data { 
	classvar <all; 
	var <>name, <chanNames, <>data, <>chanTypes; 
	var <numFrames, <numChans, <mins, <maxs, <means, <counts, <stdDevs, <filled; 
	var <sorted;
	var <>scaleSpecs;
	
		// support global storage by name:
	*initClass { all = IdentityDictionary[]; }
	
	*at { arg name; ^all[name]; }

	*remove { arg soni; all.remove(soni); }
	
	*removeAt { arg name; all.removeAt(name); } 
	
		// native data format is [ channel1, channel2, ... channelN ];
		// likely this is more efficient than [ vector1, vector2 etc ]
	*new { arg name, chanNames, data;
		^super.new
			.name_(name.asSymbol)
			.chanNames_(chanNames.collect({arg name; name.asSymbol }))
			.data_(data)
			.init;
	} 
	chanNames_ { arg names; 
		chanNames = names; 
		numChans = chanNames.size;
	}
		// convenience method: use vector data as input; store as channels.
	*fromVect { arg name, chanNames, data;
		^this.new(name, chanNames, data.flop);
	}
	init { if (this.checkData, { all.put(name, this) }); }
		
	at { arg name, index; var chan; 
		chan = this.chan(name);
		^if (index.isNil, chan, { chan.at(index) })
	}

	checkNames { |names| var validNames;
		if (names.isNil) { ^this.chanNames };
		validNames = names.asArray.sect(chanNames);
		if (names != validNames) { 
			warn(this.class.asString + this.name + 
			"filtered bad chanNames:" + (names.asArray.difference(validNames)));
			names = validNames;
		};
		^names;
	}
	
	indicesForNames { |names| // names should be checked first!
		^names.asArray.collect { |name| chanNames.indexOf(name) }
	}
		
	vector { arg index, names; var chanIndices;
		names = this.checkNames(names);
		chanIndices = this.indicesForNames(names);
		^chanIndices.collect({ arg chan; data.at(chan).at(index) }); 
	}

	eventAt { arg index, names; var ev, indices; ev = ();
		names = this.checkNames(names);

		names.do { |name| ev.put(name, data[chanNames.indexOf(name)][index]) };
		^ev
	}
	
	chanAtIndex { arg index; ^data.at(index); }
	chans { |names| ^names.collect(this.chan(_)) }

	chan { arg name; var index;
		if (name.isNumber, { index = name; }, { index = chanNames.indexOf(name) });
		if (index.isNil, { ("" ++ this.class + name + ": chan" + name + "not found.").postcln; ^nil; });
		^data.at(index);
	}
	
	checkData { var namesChansOK, allSizes, lastSize, allSizesOK; 
	
		if (data.isNil, { ("" ++ this.class + name +  "no data yet.").postcln; ^true; });
		
		namesChansOK = chanNames.size == data.size; 
		if (namesChansOK.not, { 
			warn("" ++ this.class + name + ":" + chanNames.size + "names and" + data.size + "channels?!?");
		});
		
		allSizes = data.collect({ arg chan; chan.size; });
		lastSize = allSizes.pop;
		allSizesOK = allSizes.every({ arg el; el == lastSize }); 
		
		if (allSizesOK.not, { 
			warn("" ++ this.class + name + ": channel lengths are unequal: " + allSizes + "?!?");
		});
		numFrames = lastSize;
		^namesChansOK && allSizesOK
	}
	
	analyseChan { arg index; 	// this is safe for notnumbers. 
		var chan, min, max, sum=0.0, count=0, avg, stdDev=0.0, isFull; 
		
		chan = this[index]; 	// name or index should work.
		min = max = chan.detect({ |el| el.isNumber }); 
		
		chan.do({ |el| 
			if (el.isNumber, { 
				min = min(min, el); 
				max = max(max, el); 
				sum = sum + el;
				count = count + 1;
			});
		}); 
		avg = if (count > 0, { sum / count }); 
	
		chan.do({ |el| 
			if (el.isNumber, { stdDev = (el - avg).squared + stdDev; });
		}); 
		stdDev = if (count > 0, { (stdDev / count).sqrt });
		isFull = (count == numFrames);
		^[ min, max, avg, count, stdDev, isFull ];
	}
	
	analyse { 				// simple statistics. 
		var stats; 
		stats = data.collect({ arg ch, i; this.analyseChan(i); }).flop; 
		#mins, maxs, means, counts, stdDevs, filled = stats;
		^stats
	}
	
	storeArgs { arg stream; 	// support storage as SC code.
		^[name, chanNames, data]
	}
	currentGlobalSpecs {
		Spec.specs.keysValuesDo({ |key, spec| 
			if (chanNames.includes(key)) {
				Post <<< (key -> spec); Post.nl;
			}
		});
	}
	suggestSpecs { 
		var ranges; ranges = [ mins, maxs ].flop;
		"Spec.specs.addAll([".postln;
		"	// name	->		min, max, warp, step, step=0.0, default, units".postln;
		ranges.do({ arg range, i; 
			Post.tab <<< 
			(chanNames.at(i) -> ControlSpec(range.at(0), range.at(1), 'lin', 0.0, means.at(i))); 
			if (i < (ranges.size - 1)) { ",".postln; };
		});
		"\n])".postln;
	}
		// loading stored Soni objects : 
		// saved as SC3 code:
	*fromPath { arg path;
		^thisProcess.interpreter.executeFile(path);
	}
		// data from SPSS tab delimited, space separated text file:
		// this should go away and be replaced with TabFileReader!
	*fromSPSSTabDel { arg path, soniName;
		var file, allText, lines, names, vectors, vector, word;
		file = File(path, "r");
		if (file.isOpen) { 
		//	"reading file.".inform;
			allText = file.readAllString;
			file.close;
		//	"splitting lines...".inform; 
			
			// make an empty array, make hopefully large enough guess
			lines = List(1024); 
			vector = List(512); 
			word = "";
							// could likely be more efficient;
							// could convert immediately...
			allText.do({ arg char, i; var nl;
				if ( (char == $\t) or: (nl = char == $\n)) {
					vector.add(word);
					word = ""; 
					if (nl) {
						lines.add(vector);
						vector = List(512);
					}
				} {
					word = word ++ char;
				}	
			});
				// hack: drop stupid last char before newline. 
			lines.do({ arg line; 
				line[line.size-1] = line[line.size-1].drop(-1);
			});
			names = lines.first; 
			vectors = lines[1..].collect({ arg list; 
				list.collect({ arg str; str.interpret });
			});
		};
		^Soni.vect(soniName, names, vectors);
	}
	addChan {|key, chData| 
		this.chanNames_(chanNames.add(key));
		this.data_(data.add(chData));
	}
}