TimeSeriesData : Data { 
	classvar <>classDir = "testData/", <>classSFDir = "tempSF/";
	var <buffers, <bigBuf, <>dir="testData/", <>sfDir="tempSF/"; 
	
	/*** assumptions: 
		data are only numerical, 
		usually equal time steps, 
		makes sense to load them as buffers.
	***/ 
			// we know it is all numbers, so speed it up...
//	analyseChan { arg index; 	// this is safe for notnumbers. 
//		var chan, min, max, sum=0.0, count=0, avg, stdDev=0.0, isFull; 
//		
//		chan = this[index]; 	// name or index should work.
//		min = max = chan.detect({ |el| el.isNumber }); 
//		
//		chan.do({ |el| 
//			if (el.isNumber, { 
//				min = min(min, el); 
//				max = max(max, el); 
//				sum = sum + el;
//				count = count + 1;
//			});
//		}); 
//		avg = if (count > 0, { sum / count }); 
//	
//		chan.do({ |el| 
//			if (el.isNumber, { stdDev = (el - avg).squared + stdDev; });
//		}); 
//		stdDev = if (count > 0, { (stdDev / count).sqrt });
//		isFull = (count == numFrames);
//		^[ min, max, avg, count, stdDev, isFull ];
//	}

//	analyse { 	// simple statistics. 
//		// data : not safe, because it assumes all data to be numbers. 
//		// filter nils first for usable data channels!
//		// for large integers, means, means2 and stdDevs break!!! 
//		
//		ranges = 		data.collect({ |chan| 	[ chan.minItem, chan.maxItem ] });
//		means = 		// .sum / chan.size });  
//					data.collect({ |chan| 	chan.inject(0.0, { arg sum, el; sum + el }) / chan.size }); 
//					
//		means2 = 		data.collect({ |chan| 	(chan.sum({ |x| x.squared }) / chan.size).sqrt }); 
//		stdDevs = 	data.collect({ |chan,i|	(chan - means[i]).squared.sum / chan.size });
//	}

	sfName { |folder="", ending="", ext=".aif"| 
		^folder ++ this.name.asString.splitext[0] ++ ending ++ ext;
	}
	
	writeMonoFiles { |folder, names, headerFormat, sampleFormat |
		var ext = ".aif";
		names = this.checkNames(names);
		folder = folder ? sfDir ? classSFDir; 
		if (headerFormat == "WAV", { ext = ".wav" }); 
		chanNames.do { |chname, i| 
			this.writeSF( 
				this.sfName(folder, "_" ++ i ++ "_" ++ chname, ext), 
				[chname], 
				headerFormat, 
				sampleFormat
			);
		};
	}

	writeOneSF { |folder|
			// write all data, one channel, or a group of channels as soundfile.
			// should write out a txt file with same name to recover data object! 
			// also useful to transfer large data to server. 
		folder = folder ? sfDir ? classSFDir;
		^this.writeSF(this.sfName(folder));		
	}
	
	writeSF {  arg sfpath, names, headerFormat, sampleFormat; 
		var sfdata, sf, res = false, sr; 

		names = this.checkNames(names);
		"writeSF: ".post; [name, sfpath, names].postcs;
		sfdata = this.chans(names); 
						// convert if multichannel file; 
						// should have tests!
		if (names.size > 1, { 
			sfdata = sfdata.collect(_.as(Array));
			sfdata = sfdata.lace(sfdata.collect(_.size).sum);
		}, { 
			sfdata = sfdata[0] 
		});
		sfdata = sfdata.as(Signal);
		
		sf = SoundFile.new; 
		sf.numChannels_(names.size);
		if (headerFormat.notNil, { sf.headerFormat_(headerFormat) });
		if (sampleFormat.notNil, { sf.sampleFormat_(sampleFormat) });
		
		try { sr = this.sampleRate; sf.sampleRate = sr; } { "using default sr".postln };

		if (sf.openWrite(sfpath), { 
				res = sf.writeData(sfdata);
				sf.close;
		});
		^res
	}
			// decide how to put data into this by hand for now.
	readSF {  arg sfpath;
		var sf, datArray;
		sf = SoundFile.new; 
		^if (sf.openRead(sfpath), { 
			datArray = Signal.newClear(sf.numFrames * sf.numChannels);
			sf.readData(datArray);
			sf.close;
			if (sf.numChannels > 1, { 
				datArray = datArray
					.unlace(sf.numChannels)
					.collect(_.as(Signal));
			});
			^datArray
		});
	}
	makeBigBuf { arg server, size;
		size = size ? this.numFrames;
		server = server ? Server.default;
		^bigBuf = bigBuf ?? { Buffer.alloc(server, size, data.size) };
	}
	
	readBigBuf { arg folder = "testData/", fileStartFrame = 0, numFrames, bufStartFrame = 0;
		this.makeBigBuf; 
		folder = folder ? sfDir ? classSFDir;
		numFrames = numFrames ? this.numFrames;
		[ this.sfName(folder), fileStartFrame, numFrames, bufStartFrame ].postln;
		^bigBuf.read(this.sfName(folder), fileStartFrame, numFrames, bufStartFrame);
	}
	
	makeBuffers { arg server, nFrames;	// ready for partial buffers.
		server = server ? Server.default;
		nFrames = nFrames ? this.numFrames;
		buffers = buffers ?? {
			buffers = data.collect { 
				 |chan, i| Buffer.new(server, nFrames, 1);
			}
		}
	}
	bufnum { |index| 
		if (index.isNumber.not) { index = chanNames.indexOf(index) }; 
		^buffers[index].bufnum;
	}

	allocReadBuffers { arg folder, fileStartFrame = 0, numFrames;
		folder = folder ? sfDir;
		this.makeBuffers; 
		numFrames = numFrames ? this.numFrames;
		buffers.do { |buf, i|
			buf.waitForBufInfo
			   .allocRead(
				this.sfName(folder, "_"  ++ i ++ "_" ++ chanNames[i]), 
				fileStartFrame, 
				numFrames, 
				{|buf|["/b_query",buf.bufnum]}
			);
		};
	}

	readBuffers { arg folder = "testData/", fileStartFrame = 0, numFrames, bufStartFrame = 0;

		this.makeBuffers; 
		numFrames = numFrames ? this.numFrames;
		buffers.do { |buf, i|
			buf.read(
				this.sfName(folder, "_"  ++ i ++ "_" ++ chanNames[i]), 
				fileStartFrame, 
				numFrames,
				bufStartFrame
			);
		};
	}
	
	removeDC { 
		data = data.collect { |ch| ch - ch.meanF }; 
	}
	
	removeDrift { arg zone = 256; // average over <zone> many samples
		// ...
	}
	
	normAll { 
		var peak; 
		peak = data.collect { |ch| ch.peak }.maxItem(_.abs);
		data = data / peak;
	}
			// quick utility, should be improved later! 
	quickPlot { |firstChan=0, lastChan| // , firstSec=0, lastSec

		var plotChans;
		lastChan = lastChan ? (this.numSigs - 1); 
	//	lastSec = lastSec ? (this.numRecords - 1); 
		
		plotChans = data.copyRange(firstChan, lastChan); 
		
		plotChans.collect { |chan| 
			chan		// .copyRange(firstSec * sampleRate, lastSec * sampleRate)
			.resamp1(512).as(Array)
		}	.lace(512 * this.numSigs)
			.plot(this.name, Rect(0, 0, 520, 600), numChannels: plotChans.size);
	}
}

EEGData : TimeSeriesData { 
	
	var <path, <file, <dataFormat, <patientID, <localRecID, <recDateDMY, <recDataHMS, 
	<headerSize,			// headerSize 
	<numRecords, 			// numBlocks
	<durOneRecord, 		// dur of one block (one record), typically one sec.
	<>numSigs, 			
	<>bytesPerSample = 2, 	// 16bit = 2, 24bit = 3;
	<>sampleRate,
	<>duration,
	<startBlock, <loadedFrames; 	// which part of the file is currently loaded?
	var <nameMap; 
	
	var <fileChanNames, <transducers, <dimensions, 
		<physMins, <physMaxs, <digMins, <digMaxs, 
		<filterInfos, <numSamplesPerRec, <frameSize;

	*fromFile { arg path, name, startTime = 0, durToRead; 
		var newSoni;

		name = (name ?? { path.basename.splitext.first }).asSymbol;
		newSoni = this.new(name).readEDFHeader(path);
		if (newSoni.isNil, { ^nil });

		^newSoni.readData(startTime, durToRead);
	}
	
	nameMap_ { |name| 
		var newMap; 
		newMap = EEGMap.eegMaps[name];
		if (newMap.notNil) { 
			if (newMap.size != chanNames.size) { 
				("wrong sizes: new" + newMap.size ++", old" + chanNames.size + "!").postln;
				^this
			};
			nameMap = name;
			"old chanNames: ".post; [ "size:", chanNames.size, chanNames ].postln; 
			"new chanNames: ".post; [ "size:", newMap.size, newMap ].postln; 
			chanNames = newMap;
		}
	}
	
	readEDFHeader { arg argPath;
		var headStart, dummy, newNum; 
		
		file = File(argPath, "r");
		if (file.isOpen.not) { ("EEGSoni: file" + argPath + "not found.").warn; ^nil };
	//	file.dump;
		path = argPath;
		if (path.splitext.last == "bdf") { "bdf == 24bit file.".postln; bytesPerSample = 3 };
		
		headStart = file.nextN(256);
		#dataFormat, patientID, localRecID, recDateDMY, recDataHMS, headerSize, 
		dummy, numRecords, durOneRecord, numSigs = 
		headStart.clumps([ 8, 80, 80, 8, 8, 8, 44, 8, 8, 4 ]);
		
		patientID = "";		// anonymity. 
		
		headerSize = headerSize.interpret;
		numRecords = numRecords.interpret;
		durOneRecord = durOneRecord.interpret;
		numSigs = numSigs.interpret;
		
		if (headerSize != (1 + numSigs * 256)) { 
			warn("Wrong header size???" + [ headerSize, (1 + numSigs * 256) ])
		};

			// check that info is consistent, e.g. that headerSize is correct,
			// numRecords, durOneRecord, numSigs, etc. ...
		
			// read headers for all data channels:
		#fileChanNames, transducers, dimensions, physMins, physMaxs, digMins, digMaxs, 		filterInfos, numSamplesPerRec = 
		[ 16, 80, 8, 8, 8, 8, 8, 80, 8, 32 ].collect({ arg size, i; 
			var result, chunk; 
			chunk = file.nextN(size * numSigs); 
			
			result = chunk.clump(size); 
			if ([ 3, 4, 5, 6, 8 ].includes(i), { 
				result = result.collect(_.interpret);
			});
		});		
			// check consistency of channel info... 
		if (numRecords < 0) { 
			"calculating numRecords: ".post;
			newNum = file.length - headerSize / (numSamplesPerRec.sum * bytesPerSample); 
			[ "numRecords old/new:", numRecords, newNum ].postln;
			numRecords = newNum.asInteger;
			
		};
		file.close;
		frameSize = numSamplesPerRec.sum; 
		sampleRate = numSamplesPerRec[0] / durOneRecord;
		if (numSamplesPerRec.any(_ != sampleRate))
			{ "unequal sampleRates! using first chan for sampleRate...".warn };
		numFrames = numRecords * sampleRate;
		duration = numFrames/sampleRate;
		
		this.fixChanNames; 
		^this;
	}
	
	fixChanNames { 
		if (chanNames.isNil) { 
			if (fileChanNames.notNil) { 
				chanNames = fileChanNames.collect(_.asSymbol);
			} { 
				chanNames = { |i| ("ch" ++ i).asSymbol } ! numSigs;
			}
		};
		("chanNames:" + chanNames).postln;
	} 
	
	readData { arg startTime=0, durToRead, littleEndian = true; 
		if (bytesPerSample == 2) { 
			this.read16(startTime, durToRead, littleEndian);
		} { 
			this.read24(startTime, durToRead, littleEndian);
		};
	}
	
	read16 { arg startTime=0, durToRead, littleEndian = true; 
		var 	file, newData, rawArray, frameSize, numLoadFrames; 

		var startOffset = 0, intArraySize; 
		("EEGData" + name + "reading 16bit ").post;		
		frameSize = numSamplesPerRec.sum; 

			// determine where to read: 
		durToRead = durToRead ? duration;
		numLoadFrames = numRecords;
		
		startBlock = (startTime / durOneRecord).clip(0, numRecords).asInteger; 
		startOffset = startBlock * frameSize; 
		numLoadFrames = (durToRead / durOneRecord).clip(0, (numRecords - startBlock)).asInteger;
		intArraySize = numLoadFrames * frameSize;
			
			// read segment from file.		
		file = File(path, "rb");
		file.pos_(headerSize + (startOffset * 2));
		rawArray = Int16Array.newClear(intArraySize);
		if (littleEndian) { 
			"wintel (little endian).".postln;
			file.readLE(rawArray); 
		} { 
			"mac (big endian).".postln;
			file.read(rawArray); 
		};
		file.close;
		
		newData = rawArray.clumps(numSamplesPerRec).clump(numSigs).flop
			.collect{ |sigs| Signal.newFrom(sigs.flat) };
		
		this.data_(newData);
	}
	
	read24 { arg startTime = 0, durToRead;		// nil means entire file. 
	
			// reads data as Int8Array, because 24bits are not supported yet. 
			// possible speedup by reading native 16bit arrays? 
					
		var 	file, newData, byteArray, lastByte, byteSum=0, 
			whichChan, whichFrame, whichRecord, whichBlock; 
		var byteSize, blockSize, frameSize; 
		
		var startOffset, intArraySize; 
		
		"EEGData.read24 is very likely broken! 
		at least it does not know about little endian/big endian files yet.".warn;
		byteSize = bytesPerSample;
		blockSize = numSamplesPerRec[0];
		if (numSamplesPerRec.any { |num| num != blockSize }) { 
			"EEGSOni: unequal blockSizes! not supported yet, sorry.".warn;
			^false
		};
		frameSize = numSamplesPerRec.sum * bytesPerSample; 
		numFrames = numRecords * blockSize;

			// determine where to read: 
		durToRead = durToRead ? duration;
		startBlock = (startTime / durOneRecord).clip(0, numRecords).asInteger; 
		startOffset = startBlock * frameSize; 
		loadedFrames = (durToRead / durOneRecord).clip(0, (numRecords - startBlock)).asInteger;
		intArraySize = loadedFrames * frameSize;
			
			// read segment from file.		
		file = File(path, "r");
		file.pos_(headerSize + startOffset);
		byteArray = Int8Array.newClear(intArraySize);
		file.read(byteArray);
		file.close;
		
		newData = numSamplesPerRec.collect { |size| Signal.newClear(size * loadedFrames) }; 
		lastByte = bytesPerSample - 1; 
		
		byteArray.do { |byte, index| 
			var byteIndex; 
			byteIndex = index % byteSize; 	
			byteSum = byteSum + (byte.wrap(0, 255) <<  (byteIndex * 8));
		
			if (byteIndex == lastByte) {
										// where to put it?
				whichRecord = index div: (byteSize * blockSize);
				whichChan = (whichRecord % numSigs); 
				whichFrame = index div: bytesPerSample % blockSize +
					(whichRecord div: numSigs * blockSize);

				 						// here.
				newData[whichChan][whichFrame] = byteSum; 
				byteSum = 0;
			};
		}; 
		this.data_(newData);
	}
	
	openFile { 
		if (file.isOpen.not, { file = File(path, "r"); });
		^if (file.isOpen, file, nil);
	}
	closeFile { ^file.close; }
	
	readBlock16 { arg blockNum = 0; 
				// useful later for on-the fly data...
				// read one 'record' or data block, i.e. one second worth of data,
				// n channels, 256 samples per frame.
			//	headerSize
	}
	
	highpass { arg cutoff = 1; // cutoff freq in Hz
		var coef = 1 - (cutoff / (sampleRate * pi));
		data.do { arg channel;
			var lastIn, lastOut, offset;
			offset = channel[0]; // treat first sample as 0 to avoid discontinuity (simple hack)
			lastIn = 0;
			lastOut = 0;
			channel.do { arg sample, index;
				var in = sample - offset;
				channel[index] = lastOut = in - lastIn + (coef * lastOut);
				lastIn = in;
			};
		};
	}
	
}
