/*
	Filename: EEGSpace.sc 
	created: 2005

	Copyright (C) IEM 2005 
		Alberto deCampo [decampo@iem.at]
		Christopher Frauenberger [frauenberger@iem.at] 

	This program is free software; you can redistribute it and/or 
	modify it under the terms of the GNU General Public License 
	as published by the Free Software Foundation; either version 2 
	of the License, or (at your option) any later version. 

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
	GNU General Public License for more details. 

	You should have received a copy of the GNU General Public License 
	along with this program; if not, write to the Free Software 
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

	IEM - Institute of Electronic Music and Acoustics, Graz 
	Inffeldgasse 10/3, 8010 Graz, Austria 
	http://iem.at

	SonEnvir project
	http://sonenvir.at
	
	$LastChangedDate$
	$Rev$
*/


/*****************
	// contains all spatial and frequency constants needed for eeg work.
*/
		
EEGSpace {
	classvar <posDict, <freqDict; 
	*initClass { 
				// reference positions
		posDict = Event[ 	
					// left side, lower ring
			\FP1 -> 	[0.1, 	0], 
			\F7	-> 	[0.3, 	0],
			\T3	->	[0.5, 	0],
			\T5	->	[0.7,	0],
			\O1	->	[0.9, 	0],
					// right side, lower ring
			\O2	->	[-0.9, 	0],
			\T6	->	[-0.7, 	0],
			\T4	->	[-0.5, 	0],
			\F8	->	[-0.3,	0],
			\FP2	->	[-0.1,	0],
					// interpolated, middle ring
			\F3	->	[0.2, 	0.4],
			\P3	->	[0.8, 	0.4],
			\P4	->	[-0.8, 	0.4],
			\F4	->	[-0.2, 	0.4],
					// top ring
			\FZ	->	[0, 		0.6],
			\C3	->	[0.5, 	0.6],
			\PZ	->	[1, 		0.6],
			\C4	->	[-0.5, 	0.6],
					// top centre
			\CZ	->	[0, 		1]
		];
		this.addBipolarLengthMap; 
		
		freqDict = Event[ 	
			\delta -> [1, 3.5],
			\theta -> [4, 7.5],
			\alpha -> [8, 13],
			\beta -> [14, 30],
		];
	}
	*keys { 	^posDict.keys }

	*at { 	|key, factors=#[1,1]| 
		^key.asArray.collect { |k| posDict.at(k) * factors }.unbubble 
	}

	*atDeg { 	|key| ^this.at(key, [180, 90]) }

	*atRad { 	|key| ^this.at(key, [pi, 0.5pi]) }
	
	*atFreq { |key| 
		^key.asArray.collect { |k| freqDict.at(k) }.unbubble 
	}
	*addBipolarLengthMap { 
		(				// all eeg-positions
			FP2_F8: 	[ -0.2, 0.0],
			F8_T4: 	[ -0.4, 0.0],
			T4_T6: 	[ -0.6, 0.0],
			T6_O2: 	[ -0.8, 0.0], 
			
			FP1_F7: 	[ 0.2, 0.0],
			F7_T3: [ 0.4, 0.0],
			T3_T5: [ 0.6, 0.0],
			T5_O1: [ 0.8, 0.0],
			
			FP2_F4: [ -0.15, 0.2],
			F4_C4: [ -0.35, 0.5],
			C4_P4: [ -0.65, 0.5],
			P4_O2: [ -0.85, 0.2], 
			
			FP1_F3: [ 0.15, 0.2],
			F3_C3: [ 0.35, 0.5],
			C3_P3: [ 0.65, 0.5],
			P3_O1: [ 0.85, 0.2],
			
			FZ_CZ: [ 0.0, 0.8],
			CZ_PZ: [ 1.0, 0.8]
		).keysValuesDo { |k, v| posDict.put(k, v) }	
	}
}