/*
	Filename: SEParser.sc 
	created: 12.11.2005 

	Copyright (C) IEM 2005 
		Alberto deCampo [decampo@iem.at]
		Christopher Frauenberger [frauenberger@iem.at] 

	This program is free software; you can redistribute it and/or 
	modify it under the terms of the GNU General Public License 
	as published by the Free Software Foundation; either version 2 
	of the License, or (at your option) any later version. 

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
	GNU General Public License for more details. 

	You should have received a copy of the GNU General Public License 
	along with this program; if not, write to the Free Software 
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

	IEM - Institute of Electronic Music and Acoustics, Graz 
	Inffeldgasse 10/3, 8010 Graz, Austria 
	http://iem.at

	SonEnvir project
	http://sonenvir.at
	
	$LastChangedDate$
	$Rev$
*/

/*****************
	Super Class for parsing files
*/
SEParser {

	// virtual method, to be overwritten by sub-classes
	parse { arg sourceUrl, targetUrl;
	
	}
}

SEParserEDF : SEParser { 
	classvar defaultFolderExt = "_dir"; 

			// read an edf, convert it to a folder with header file and soundfiles.
	*parse { arg sourceUrl, targetUrl, folderExt = "_dir"; 
//		var urlExt, urlStart, folderName, info; 
//		#urlStart, urlExt = sourceUrl.splitext; 
//		
//		if (File.exists(sourceUrl).not) { ("file" + sourceUrl + "does not exist.").warn; ^nil };
//		if (urlExt != "edf") { ("file"  + sourceUrl + "is not an edf file.").postln; ^nil };
//
//		targetUrl = targetUrl ?? { urlStart ++ folderExt };
//		
//		info = this.readHeader(sourceUrl); 
//				
//		if (File.exists(targetUrl)) { 
//		//	var soundfilesOK; 
//			"target folder exists, checking info and soundfiles..." 
//			// check whether the folder and the files look OK.
//			// if yes, load them from there, if not, write them again.
//			
//		} {	
//			"dir not there yet, do convert...".postln;
//			this.convertToFolder(info, targetUrl);
//		};
	}

	*readHeader { arg argPath;		// did checks already
		var headStart, newNum; 
		var file, d, chanInfos;
		var dataFormat, patientID, localRecID, recDateDMY, recDataHMS, 
			headerSize, dummy, numRecords, durOneRecord, numSigs;
		
		d = ();
		file = File(argPath, "r");
		if (file.isOpen.not) { ("ParserEDF: file" + argPath + "not found.").warn; ^false };
		
		"file opened ok.".postln; 
		
		d.path = argPath;
		if (d.path.splitext.last == "edf") { 
			"edf == 16bit file.".postln; 
			d.bytesPerSample = 2; 
		};
		if (d.path.splitext.last == "bdf") { 
			"bdf == 24bit file.".postln; 
			d.bytesPerSample = 3; 
		};
		
		headStart = file.nextN(256);
		
		#dataFormat, patientID, localRecID, recDateDMY, recDataHMS, headerSize, 
		dummy, numRecords, durOneRecord, numSigs = 
			headStart.clumps([ 8, 80, 80, 8, 8, 8, 44, 8, 8, 4 ]);
		
		
		d.headerSize = headerSize.interpret;
		d.numRecords = numRecords.interpret;
		d.durOneRecord = durOneRecord.interpret;
		d.numSigs = numSigs.interpret;
		
//			// check that info is consistent, e.g. that headerSize is correct,
//			// numRecords, durOneRecord, numSigs, etc. ...

		if (d.headerSize != (1 + d.numSigs * 256)) { 
			warn("Wrong header size???" + [ d.headerSize, (1 + d.numSigs * 256) ]); 
			file.close; ^false
		};

			// read headers info for all data channels:
		chanInfos = [ 16, 80, 8, 8, 8, 8, 8, 80, 8, 32 ].collect({ arg size, i; 
			var result, chunk; 
			chunk = file.nextN(size * d.numSigs); 		
			result = chunk.clump(size); 
			if ([ 3, 4, 5, 6, 8 ].includes(i), { 
				result = result.collect(_.interpret);
			});
		});	
		
		[\fileChanNames, \transducers, \dimensions, \physMins, \physMaxs, 
			\digMins, \digMaxs, \filterInfos, \numSamplesPerRec
		].do { |name, i| d.put(name, chanInfos[i]) };
		
		
			// check consistency of channel info... 
		if (d.numRecords < 0) { 
			"calculating numRecords: ".post;
			newNum = file.length - d.headerSize / (d.numSamplesPerRec.sum * d.bytesPerSample); 
			[ "numRecords old/new:", numRecords, newNum ].postln;
			d.numRecords = newNum.asInteger;
		};
		
		file.close;
		
		d.frameSize = d.numSamplesPerRec.sum; 
		d.sampleRate = d.numSamplesPerRec[0] / d.durOneRecord;
		if (d.numSamplesPerRec.any(_ != d.sampleRate))
			{ "unequal sampleRates! using first chan for sampleRate...".warn };
		d.numFrames = d.numRecords * d.sampleRate;
		d.duration = d.numFrames/d.sampleRate;
		
		^d;

	//	this.fixChanNames; 
	}

	*writeSCInfoFile { |path, info| 
			// simplest possible info file:
			// write name/value pairs as a dict. 
			// maybe this should better be one pair per line, in orig. order?
			
		var res; 
		var f = File(path ++ "__info.sc", "w"); 
		res = f.write(info.asCompileString);	
		f.close;
		^res;
	}

	*readSCInfoFile { |path, info| 
		^(path ++ "__info.sc").load;
	}

	*checkDirSfsExist { |infodict, folderName|
		^pathMatch(folderName).notEmpty and: { 
			
		};
	}
	
	*convertToFolder { |infodict, folderExt = "_dir", folderPath|
		// by default, put the folder next to the edf file with '_edf_sf' instead of .edf.
		
		var file, dirname, basename, soundfiles, dataArray, numSamplesPerRec; 
		var dirEmpty;
		
		numSamplesPerRec = infodict.numSamplesPerRec.first; // assuming equal samplerates
		file = File(infodict.path, "rb"); 
		if (file.isOpen.not, { warn("file" + infodict.path + "not found!"); ^nil });
		
			// make directory, write soundfiles
		basename = infodict.path.basename.splitext.first; 
		dirname = folderPath ?? { infodict.path.dirname } ++ "/" 
			++ basename ++ folderExt;
		
			// assume we checked that dir may not be there 
			// and soundfiles are definitely not there yet ...
			
			// have to do this in a task because there is no 
			// easy way to wait for a unixcmd to return.
		fork { 
			var count = 0;
			while { 
				dirEmpty = dirname.pathMatch.postln.isEmpty;
				dirEmpty and: (count < 50); 
			} { 
				unixCmd("mkdir" + dirname);
				0.1.wait;
			}; 
			if (dirEmpty) { 
				warn("could not write directory" + dirname + ". edf not converted!");
			} {
					// make sure to wait for directory to be there!!! (Task { while }?)
					// soundfiles not there, so make them:
				soundfiles = { |i|
					var f = SoundFile.new
						.sampleFormat_("int16")
						.sampleRate_(infodict.sampleRate);
						
					f.openWrite( 
						dirname ++ "/" 
						++ basename 
					//	++ "_" ++ infodict.chanNames[i]
						++ "_" ++ (i + 101).asString.drop(1)
						++ ".aif" );
					f;
				} ! infodict.numSigs;
				
				dataArray = Int16Array.newClear(infodict.frameSize);
		
				file.pos_(infodict.headerSize);		
				infodict.numRecords.do { |rec| 
					file.readLE(dataArray);		// read one record (1sec block)
					dataArray.clump(numSamplesPerRec).do { |chanBlock, i| 
						soundfiles[i].writeData(chanBlock.as(Int16Array));
					};
				};
				soundfiles.do (_.close);
				"conversion done.".postln;
			}
		};
	}
}