JPxPreset { 
	classvar <>scheme; 
	
	var <proxy, <settings, <currSet, <targSet, count=0; 
//	var <>ignoreKeys; 
	
	var <win, <zone, <setLPop, <storeBtn, <delBtn, <setRPop, <setLBox, <xfader, 
	<setRBox, <skipjack;
	
	*initClass { 	// same as ProxyMixer, NodeProxyEditor, ProxyMonitorGui.
		scheme = ();
		scheme.font = JFont("Monaco", 10);
		scheme.fontColor = Color.black;
		
		scheme.background = Color(0.8, 0.85, 0.7, 0.5);
		scheme.foreground = Color.grey(0.95);
		scheme.onColor = Color(0.5, 1, 0.5); 
		scheme.offColor = Color.clear; 
		
		scheme.buttonHeight = 16;
		scheme.gap = 0 @ 0;
	}
	
	*new { |px, win, zone| 
		^super.new.init(win, zone).proxy_(px);
	}
	
	proxy_ { |px| 
		proxy = px; 
		if (proxy.isKindOf(NodeProxy)) { 
						// proper init state
			this.currFromPx; 
			currSet = targSet = this.getSet(\curr);
		};
	}
	
	init { |win, zone| 		// a list of assocs
		settings =  List[];
		this.show(win, zone);
	} 
	
	keys { ^settings.collect(_.key) }
	
	getIndex { |name| ^settings.detectIndex({ |assoc| assoc.key == name }) }
	
	getSet { |name| 
		var index = this.getIndex(name);
		^if (index.notNil) { settings[index] } { nil };
	}
	
	setPx { |name| proxy.set(*this.getSet(name).value.flat) }
	
	addSet { |name| 
		var index;
		if (name.isNil, { count = count + 1; name = "set" ++ count; });
		name = name.asSymbol;
		
		index = this.getIndex(name);
		if (index.notNil)
			{ 	settings.put(index, name -> proxy.getKeysValues) }
			{ 	settings.add(name -> proxy.getKeysValues) }
	}
	
	removeSet { |name|
		var index = this.getIndex(name);
		if (index.notNil, { settings.removeAt(index) });
	}
	
	randSet { |rand=0.25, startSet, except|
		var randKeysVals, set, randRange; 
				// vary any given set too?
		set = this.getSet(startSet).value ?? { proxy.getKeysValues(except) };
	
		randKeysVals = set.collect { |pair| 
			var key, val, normVal, randVal, spec; 
			#key, val = pair;
			spec = key.asSpec; 
			if (spec.notNil, {
				normVal =  spec.unmap(val); 
				randVal = rrand( 
					(normVal - rand).max(0), 
					(normVal + rand).min(1)
				); 
			//	[key, val, normVal].postcs;
				[key, spec.map(randVal)]
			});
		};
		^randKeysVals;
	}
	setRand { |rand=0.25, startSet, except|
		proxy.set(*this.randSet(rand, startSet, except).flat);
	}

	morph { |val, name1, name2, mapped=true|
		proxy.set(*(this.blend(val, name1, name2, mapped).flat));
	}	
	
	blend { |blend = 0.5, name1, name2, mapped=true|
		var set1, set2;
		set1 = if (name1.isNil, currSet, { this.getSet(name1) }).value;
		set2 = if (name2.isNil, targSet, { this.getSet(name2) }).value;
		if (mapped) { 
			set1 = this.unmapSet(set1);
			set2 = this.unmapSet(set2);
			^this.mapSet(this.blendSets(blend, set1, set2))
		} { 
			^this.blendSets(blend, set1, set2)
		}
	}
		// expects just list of [key, val]s
	unmapSet { |set| 
		var key, val; 
		^set.collect { |pair| 
			#key, val = pair;
			[key, key.asSpec.unmap(val)]
		}
	}
	
		// expects just list of [key, val]s
	mapSet { |set| 
		var key, val;
		^set.collect { |pair| 
			#key, val = pair;
			[key, key.asSpec.map(val)]
		}
	}
	
	blendSets { |blend = 0.5, set1, set2| 
		^set1.blend(set2, blend); 
	}

	setCurr { |name| 
		var foundSet; 
		foundSet = this.getSet(name);
		if (foundSet.notNil) { 
			currSet = foundSet; 
			proxy.set(*currSet.value.flat.postln);
		//	xfader.value_(0);
		};
	 }
	setTarg { |name| 
		var foundSet; 
		foundSet = this.getSet(name);
		if (foundSet.notNil) { targSet = foundSet; };
	}
	
	currFromPx { this.addSet(\curr, proxy.getKeysValues) }


		// gui stuff; move into its own class? 
	
	show { |w, z| 
		var flow;

		win = w ?? { JSCWindow("PxPreset", Rect(128, 64, 350, 50)).front };
		zone = z ?? { JSCCompositeView(win, Rect(0, 0, 350, scheme.buttonHeight * 2)) };
		zone.decorator = flow = zone.decorator ?? { FlowLayout(zone.bounds, 0@0, 0@0) };
		
		JSCStaticText(zone, Rect(0,0,30,16)).string_("curr")
			.background_(scheme.foreground)
			.font_(scheme.font).align_(0);

		setLPop = JSCPopUpMenu(zone, Rect(0,0,50,16))
			.items_([]).font_(scheme.font)
			.background_(scheme.foreground)
			.action_({ |pop| var name; 
				name = pop.items[pop.value].asSymbol;
				this.setPx(name); 		// sets proxy too.
				this.setCurr(name);
			});
			
		storeBtn = JSCButton(zone, Rect(0,0,40, scheme.buttonHeight))
			.states_([["sto", scheme.fontColor, scheme.foreground]])
			.font_(scheme.font)
			
			.action_({ this.storeDialog });
		delBtn = JSCButton(zone, Rect(0,0,40, scheme.buttonHeight))
			.states_([["del", scheme.fontColor, scheme.foreground]]).font_(scheme.font)
			.action_({ this.deleteDialog });
		
		JSCButton(zone, Rect(0,0,40, scheme.buttonHeight))
			.states_([["rand", scheme.fontColor, scheme.foreground]]).font_(scheme.font);
		JSCButton(zone, Rect(0,0,40, scheme.buttonHeight))
			.states_([["doc", scheme.fontColor, scheme.foreground]]).font_(scheme.font);
		
		setRPop = JSCPopUpMenu(zone, Rect(0,0,50, scheme.buttonHeight))
			.items_([]).font_(scheme.font)
			.background_(scheme.foreground)
			.action_({ |pop| this.setTarg(pop.items[pop.value].asSymbol); });

		JSCStaticText(zone, Rect(0,0,30, scheme.buttonHeight))
			.background_(scheme.foreground)
			.string_("targ").font_(scheme.font).align_(0);
		
		flow.nextLine;
		setLBox = JSCNumberBox(zone, Rect(0,0,20, scheme.buttonHeight))
			.background_(scheme.foreground)
			.font_(scheme.font).align_(0).value_(-1);
		xfader = JSCSlider(zone, Rect(0,0,280, scheme.buttonHeight))
			.action_({ |sl| 
				this.morph(sl.value, currSet.key, targSet.key); 
			});
			
		setRBox = JSCNumberBox(zone, Rect(0,0,20, scheme.buttonHeight))
			.background_(scheme.foreground)
			.font_(scheme.font).align_(0).value_(-1);
		
		skipjack = SkipJack({ this.updateGui }, 0.5, { win.isClosed }, "PxPreset");
	}
	updateGui { 
		if (win.isNil or: { win.isClosed }) { win = nil; zone = nil; ^this };
		if (proxy.isNil) { ^this };
		
		setLPop
			.items_(this.keys)
			.value_(this.getIndex(this.currSet.key) ? 0);

		setRPop
			.items_(this.keys)		// .reject(_ == \curr) ??
			.value_(this.getIndex(this.targSet.key) ? 0);
	}
	
	storeDialog { |name| 		// check before overwriting a setting?
		var w;
		if (name.isNil, { count = count + 1; name = "set" ++ count; });
		w = JSCWindow("", Rect(0,600,150, 24), false);
		JSCStaticText(w, Rect(0,0,70,20)).align_(0).string_("name set:");
		JSCTextField(w, Rect(70,0,70,20)).align_(0)
			.string_(name)
			.action_({ arg field; 
				this.addSet(field.value.asSymbol.postln); 
				w.close; 
			})
			.focus(true);
		w.front;
	}
			// good enough for now
	deleteDialog { |name| 
		var w, names, size, buttons, redraw; 
		names = this.settings.collect(_.key);
		names.remove(\curr); 
		size = names.size; 
		[size, names].postcs;
		
			// up arr, down arr, delete, listview; 
		w = JSCWindow("delete which?", Rect(0,600, 200, 400));
		w.view.decorator = FlowLayout(Rect(0,0, 200, 400));
		
		buttons = names.collect { |setname|
			JSCButton(w, Rect(0,0,96, 20))
				.states_([[setname.asString]])
				.action_({ |btn| 
					this.removeSet(btn.states.first.first.asSymbol.postcs);
					redraw.value;
				});
		};
		
		redraw = { 
			names = this.settings.collect(_.key);
			names.remove(\curr); 
			buttons.do { |btn, i| 
				btn.states_([[names[i] ? '_']])
			};
			w.refresh;
		};
//		SCListView(w, Rect(0,0,200, 400))
//			.background_(Color.white)
//			.items_(names)
//			.action_({ 
//				"not done yet.".postln; 
//			});
		w.front;
	}

}