DomoScope { 

	var <server,  proxy, resp, skipjack;
	var <window, ampViews, startBtn, slider, <>redLevel=0.95, magnify=2;
	
	*new { arg server, numRings=#[16,12,8,4], angleOffsets;
		^super.newCopyArgs(server ? Server.default)
	//	.initData(posData)
		.makeWindow(numRings, angleOffsets)
		.start
	}
	posFor { |ringNums = #[16,12,8,4], angleOffsets |
		angleOffsets = angleOffsets ? [2pi / 32, 2pi / 12, 2pi / 16, 0]; // Animax Offsets
		ampViews = ringNums.collect { |perRing, ring| { |num|  
			var center = 200, size = 25; 
			var radius = [200, 150, 100, 50][ring];
			var angle = (2pi * (num / perRing)) + angleOffsets[ring];
			var left = angle.sin * radius + center; 
			var top =  center - (angle.cos * radius); 
			
			SCStaticText(window, Rect(left, top, size, size))
			.string_(num + [0, 16, 28, 36][ring]).align_(0).stringColor_(Color.white)
			.background_(Color.black)
		} ! perRing }.flat;
	}
	
	offset {  ^0 }
	numChannels { ^40 }
	makeWindow { |numRings, angleOffsets|
	
		window = SCWindow("Domoskop", Rect( 600, 500, 430, 430)).front;
		window.view.background_(Color(0,0,0.15));
		#startBtn = [ \start, \stop ].collect { |name, i| 
			SCButton(window, Rect(i * 50 + 2, 2, 48, 20))
				.states_([[name, Color.white, Color(0,0,0.15)]])
				.action_({ this.perform(name); });
		};
		slider = SCSlider(window, Rect(300, 0, 100, 20))
			.action_({ |sl| magnify = 10 ** sl.value });
		this.posFor(numRings, angleOffsets);	
	}
	
	start {
		if(server.serverRunning.not) {  "domoscope: server not running".warn; ^this };

		skipjack.stop;

		proxy = proxy ?? {  NodeProxy.control(server, this.numChannels) };
		proxy.source = {
			Amplitude.kr(InFeedback.ar(this.offset, this.numChannels), 0, 0.5)
		};
		resp.remove;
		resp = OSCresponder(server.addr, '/c_set', { arg time, r, msg; 
			var amps = msg.copyToEnd(1).clump(2).flop[1];
			{  this.amps_(amps * (magnify ? 1)); }.defer;
		} ).add; 

		startBtn.states_([[\start, Color.white, Color(0,0,0.5)]]).refresh; 
		
		skipjack = SkipJack( 
			{ proxy.wakeUp; this.updateViews; }, 
			0.2, 
			{ window.isClosed; },
			"domoscope"
		);
	}

	updateViews { 
		server.listSendMsg(["/c_get"] ++ ((_ + proxy.index) ! proxy.numChannels));
	}
	
	stop { 
		startBtn.states_([[\start, Color.white, Color(0,0,0.15)]]).refresh; 
		this.amps_([]);
		skipjack.stop;
		proxy.free;
		resp.remove;
	}
	
	amps_ { arg vals;
		var amp, col;
		ampViews.do { |el, i|  
			amp = (vals[i] ? 0).sqrt; 
			col = if (amp > redLevel, { Color.red }, { Color.yellow( amp ) });
			el.background_(col) 
		}
	}
	
	color_ { arg background;
		ampViews.do(_.background_(background));
	}
}

CUBEScope : DomoScope {		// for IEM Cube; geometry can be more precise.

	posFor { |ringNums = #[12,8,4], angleOffsets |
		angleOffsets = angleOffsets ? [0, pi/8, pi/4, 0]; // ca. IEM cube offsets
		
		ampViews = ringNums.collect { |perRing, ring| { |num|  
			var center = 200, size = 25; 
			var radius = [200, 130, 50][ring];
			var angle = (2pi * (num / perRing)) + angleOffsets[ring];
			var left = angle.sin * radius + center; 
			var top =  center - (angle.cos * radius); 
			
			SCStaticText(window, Rect(left, top, size, size))
			.string_(num + 1 + [0, 12, 20][ring]).align_(0).stringColor_(Color.white)
			.background_(Color.black)
		} ! perRing }.flat;
	}
}


	