/*
	Filename: EEGViewer.sc
	created: 12.10.2005 

	Copyright (C) IEM 2005 
		Alberto deCampo [decampo@iem.at]
		Christopher Frauenberger [frauenberger@iem.at] 

	This program is free software; you can redistribute it and/or 
	modify it under the terms of the GNU General Public License 
	as published by the Free Software Foundation; either version 2 
	of the License, or (at your option) any later version. 

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
	GNU General Public License for more details. 

	You should have received a copy of the GNU General Public License 
	along with this program; if not, write to the Free Software 
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

	IEM - Institute of Electronic Music and Acoustics, Graz 
	Inffeldgasse 10/3, 8010 Graz, Austria 
	http://iem.at

	SonEnvir project
	http://sonenvir.at
	
	$LastChangedDate$
	$Rev$
*/


EEGViewer {

	// dimensions and number of channels
	var width, height, numCh;

	// loaded EEG data 
	var <>eegData, fileChooser, chLabels, chView, start=0, end=10, len=10, scale=1, duration=10;
	
	// main window and other gui elements
	var win, layout, stbox, lenbox, endbox, magbox, rslider;
	
	*new { | width = 1000, height = 640, numCh = 30 |
		^super.newCopyArgs(width, height, numCh)
			.eegData_(IdentityDictionary.new)
			.makeWindow;
	}

	makeWindow {
		var chHeight;
	
		win = SCWindow("EEGDataViewer", Rect(0, 0, width, height)).front;
		win.view.decorator = FlowLayout(win.bounds.copy.left_(0).top_(0), 2@0, 1@0);
		layout = win.view.decorator;
		
		SCButton(win, Rect(0,0,50,16)).states_([["file"]]);
		fileChooser = SCPopUpMenu(win, Rect(0,0,100,16)).background_(Color.white)
			.action_ { |pop|
				chLabels.do { |lb, i| lb.string_(eegData[fileChooser.items[pop.value]].chanNames[i] ? "_") };
				this.viewRange;
			};
		layout.shift(5); 
		SCStaticText(win, Rect(0,0,40,16)).string_("start").align_(0).background_(Color.white);
		stbox = SCNumberBox(win, Rect(0,0,40,16)).value_(start)
			.action_({ |box| 
			start = box.value; 
			end = len + start; 
			this.viewRange;
			}
		);
		SCStaticText(win, Rect(0,0,40,16)).string_("length").align_(0).background_(Color.white);
		lenbox = SCNumberBox(win, Rect(0,0,40,16)).value_(len)
			.action_({ |box| 
				len = box.value; 
				end = len + start; 
				this.viewRange;
			}
		);
		rslider = SCRangeSlider(win, Rect(0,0,300,16)).lo_(0).hi_(1)
			.action_({ |slider|
				start = slider.lo * duration;
				end = slider.hi * duration;
				len = end -start;
				this.viewRange;				
			});
		SCStaticText(win, Rect(0,0,40,16)).string_("end").align_(0).background_(Color.white);
		endbox = SCNumberBox(win, Rect(0,0,40,16)).value_(end)
			.action_({ |box| 
				end = box.value; 
				len = end - start; 
				this.viewRange;
			}
		);

		layout.shift(2); 
		SCStaticText(win, Rect(0,0,50,16)).string_("magnify").align_(0).background_(Color.white);
		magbox = SCNumberBox(win, Rect(0,0,40,16)).value_(1)
			.action_({ |box| 
				scale = box.value; 
				this.viewRange;
			}
		);
		layout.nextLine.shift(0, 3); 
		chHeight = height - 40 div: numCh; 
		#chLabels, chView = ({ |i| 				
			[ SCStaticText(win, Rect(0,0,40,16)).string_("ch" + i)
				.align_(0).background_(Color.white)
				.font_(Font("Monaco", 10))
				.visible_(false),
			
			SCMultiSliderView(win, Rect(0,0, width - 50, chHeight))
				.valueThumbSize_(1)
				.indexThumbSize_(1).drawLines_(true)
				.background_(Color.white)
				.visible_(false);
			]	
		} ! numCh).flop;
	}

	add { | name, data|
		eegData.put(name, data);
		fileChooser.items_(eegData.keys.asArray.sort);
		fileChooser.value_(fileChooser.items.indexOf(name));
		chLabels.do { |lb, i| lb.string_(eegData[name].chanNames[i] ? "_") };
		this.viewRange;
	}

	remove { |name|
		eegData.removeAt(name);
		fileChooser.items_(eegData.keys.asArray.sort);
	}

	viewRange { 

		var curData;
		var data2View, view, srate; 
		var startFrame, endFrame, chans, chHeight; 
	
		if (fileChooser.items.size !=0, {
			curData = eegData[fileChooser.items[fileChooser.value]];

			srate = curData.sampleRate; 
			startFrame = (start * srate).asInteger; 
				// no idea why I need to multiply with 2, 
				// but it is necessary!
			endFrame = (start + (len * 2) * srate - 1).asInteger; 
			
			stbox.value_(start); 
			lenbox.value_(len = end - start); 
			endbox.value_(end); 
			duration = curData.duration;
			rslider.lo_(stbox.value/duration).hi_(endbox.value/duration);
			
			if (curData.data.size > chView.size, { 
				("not showing chans" + chView.size + "-" + (curData.data.size - 1)).postln; 
			});
		
			chans = curData.data.size;
			chHeight = height - 40 div: chans; 
			30.do({ |i| 
				if (i < chans, {
					chView[i].visible_(true); 
					chView[i].bounds_(Rect(50,i*chHeight + 20, width - 50, chHeight-2));
					chLabels[i].visible_(true);
					chLabels[i].bounds_(Rect(5,i*chHeight + 20, 40, chHeight-2));
				},{
					chView[i].visible_(false);
					chLabels[i].visible_(false);  	
				});
			});
	
			curData.data.do { |chan, i|
				view = chView[i]; 
				if (view.notNil, { 
					data2View = chan.copyRange(startFrame, endFrame)
						.resamp1(width - 50).as(Array)
						* scale + 0.5;
					view.value_(data2View);
				}, { 
				//	("no space to view data chan" + i ++".").postln;
				})
			};
		});
	}
}