
ProxyMonitorGui { 
	classvar <>scheme; 
	
	var <>proxy, <win, <zone, <flow, <skipjack, <usesPlayN; 
	var <ampSl, <ampSpec, <playBut, <nameBut, <setOutBox, <playNDialogBut, <pauseBut, <sendBut; 
	var <monHasSeriesOut = true, <allowSetOutBox = true; 
	
	*new { |px, w, height=16, usePlayN=true, makeWatcher=true| 
		^super.new.proxy_(px).init(w, height, usePlayN, makeWatcher); 
	}

	*initClass { 	// scheme defines the class look and feel.
		scheme = ();
		scheme.font = Font("Monaco", 10);
		scheme.fontColor = Color.black;
		
		scheme.background = Color(0.8, 0.85, 0.7, 0.5);
		scheme.foreground = Color.grey(0.95);
		scheme.onColor = Color(0.5, 1, 0.5); 
		scheme.offColor = Color.clear; 
		
		scheme.gap = 0 @ 0;
	}

	init { |w, height, usePlayN, makeWatcher| 
	
		usesPlayN = usePlayN; 
		ampSpec = ControlSpec(0, 1, 'amp', 0, 0.25);
		
		win = w ?? { SCWindow("pxmon", Rect(0, 0, 330, 50)) };
		zone = SCCompositeView(win, Rect(0, 0, 310, height)); 
		zone.background_(scheme.foreground);

		flow = FlowLayout(zone.bounds, 0@0, scheme.gap);
		zone.decorator = flow; 
		
		ampSl = SCSlider(zone, Rect(0,0,110, height))
			.action_({ arg slid; if(proxy.notNil) { proxy.vol_(ampSpec.map(slid.value)); } });
		
		nameBut = SCDragSource(zone, Rect(0,0,60,height))
			.font_(scheme.font).align_(0);
			
		nameBut.object_(try { proxy.key });	
		
		playBut = SCButton(zone, Rect(0,0,34,height))
			.font_(scheme.font)
			.states_([
				["play", scheme.fontColor, scheme.offColor], 
				["stop", scheme.fontColor, scheme.onColor ]
			]);
		playBut.action_({ arg btn, modif; 
			// 524576 is alt, 262401 is ctl, 8388864 is fn
			if(proxy.notNil) {
				[ 	{ 	if (modif == 524576) { proxy.end } { proxy.stop } }, 
					{ 	proxy.playN; } 
				].at(btn.value).value
			}
		});
		
		setOutBox = SCNumberBox(zone, Rect(0,0,16,height))
			.font_(scheme.font).align_(0).value_('-');
			
		setOutBox.action_({ |box, mod| 
			var val = box.value;
			mod.postln;
			monHasSeriesOut = proxy.monitor.hasSeriesOuts;
			if (monHasSeriesOut or: allowSetOutBox) { 
				val = val.max(0).asInteger;
				proxy.monitor.out_(val);
			} { 
				"proxy has multiple outs, use playN dialog button!".postln;
				box.value_("-");
			}
		}); 
		if (usesPlayN) { 
			playNDialogBut = SCButton(zone, Rect(0,0,20,height))
				.states_([
					["-<", scheme.fontColor, scheme.onColor], 
					["-=", scheme.fontColor, scheme.offColor]
				]).font_(scheme.font); 
				
			playNDialogBut.action_({ |box, mod|
				// option should make make numberbox valid again
				if (proxy.notNil) { 
					if (mod == 524576) 
						{ allowSetOutBox = true } 
						{ proxy.playNDialog }
				}
			});
		}; 
		
		pauseBut = SCButton(zone, Rect(0,0,34,height)).font_(scheme.font)
			.states_([
				["paus", scheme.fontColor, scheme.onColor], 
				["rsum", scheme.fontColor, scheme.offColor]
			]);
		pauseBut.action_({ arg btn; 
			if(proxy.notNil, {
					[ 	{ proxy.resume; }, { proxy.pause; }  ].at(btn.value).value;
				})
			});
		
		sendBut = SCButton(zone, Rect(0,0,34,height)).font_(scheme.font)
			.states_([ 
				["send", scheme.fontColor, scheme.offColor], 
				["send", scheme.fontColor, scheme.onColor] 
			])
			.action_({ arg btn, mod; 
				if(proxy.notNil and: (btn.value == 0)) { 
					if (mod == 524576) { proxy.rebuild } { proxy.send }
				}
			});
		
		if (win.isKindOf(SCWindow)) { win.front };
		if (makeWatcher) { this.makeWatcher };
	}

	makeWatcher { 
		skipjack.stop;
		skipjack = SkipJack({ this.updateAll }, 
			0.5, 
			{ win.isClosed },
			"PxMon" + try { proxy.key }
		);
		skipjack.start;
	}
	
	updateAll { 
		var monitor, outs, amps, newHasSeriesOut;

		if (win.isClosed) {skipjack.stop; ^this };
		if (proxy.isNil) {
			ampSl.value_(0);
			playBut.value_(0);
			nameBut.object_(nil);
			pauseBut.value_(0);
			sendBut.value_(0);
			^this;
		}; 
		
	//	"updating ... ".postln;
		nameBut.object_(proxy.key);	
		pauseBut.value_(proxy.paused.binaryValue);
		monitor = proxy.monitor; 
		
		playBut.value = monitor.isPlaying.binaryValue; 
		sendBut.value = proxy.objects.notEmpty.binaryValue; 

		if (monitor.isNil, { 
			setOutBox.value = "-";
		}, { 
			ampSl.value = ampSpec.unmap(proxy.monitor.vol);
			
				// switching to multi locks setOutBox
			newHasSeriesOut = proxy.monitor.hasSeriesOuts; 
			if (newHasSeriesOut.not and: monHasSeriesOut) { allowSetOutBox = false };
			
			if (usesPlayN) { 
				monHasSeriesOut = newHasSeriesOut;
				playNDialogBut.value_(monHasSeriesOut.binaryValue);
	
				outs = monitor.outs; 
				if (setOutBox.hasFocus.not) { 
						// if consecutive, it was play, set outbox to legal value.
					if (monHasSeriesOut, {
						setOutBox.value = try { outs[0] } ? 0;
					}, { 
						setOutBox.value = "-";
					});
				};
			} {		// outs assumed to be in series
				outs = monitor.outs; 
				if (setOutBox.hasFocus.not) { setOutBox.value = try { outs[0] } ? 0; }
			
			};
		});
	}
	clear { proxy = nil }
}