Spectralyzer { 
	var <nBands, <minFreq, <maxFreq, <server, <rq, <freqs; 
	var <win, <ampsV, <dbVals, <midDb = -30, <rangeDb = 15, <dt=0.1;
	var <inPx, <ampsPx, task, lastNDbs, <>avgN = 5; 
	
	*new { |nBands=60, minFreq=20, maxFreq=20000, server| 
		server = server ? Server.default;
		^super.newCopyArgs(nBands, minFreq, maxFreq, server).init;
	}
	init { 
		freqs = (maxFreq / minFreq) ** ((0.5..nBands) / nBands) * minFreq;
		rq = freqs[1] / freqs[0] - 1; 
		inPx = NodeProxy(server, \audio, 1);
		ampsPx = NodeProxy(server, \control, nBands + 1);
		ampsPx.source = this.ampsFunc; 
		task = Task { 
			loop { max(0.1, dt).wait; 
				this.poll({ |mAmps| { this.setAmps(mAmps) }.defer; });
			};
		};
	}
	midDb_ { |val| midDb = val.clip(-100, 0) }
	rangeDb_ { |val| rangeDb = val.clip(0, 30) }
	
	ampsFunc { 
		^{ Amplitude.kr([inPx.ar(1)] ++ BPF.ar(inPx.ar(1), freqs, rq), 0.0, 1) }
	}
	listenTo { |src| inPx.source = src; }
	
	poll { |func| ^ampsPx.bus.getn(nBands, func) }

	start { 
		inPx.wakeUp; 
		ampsPx.wakeUp; 
		task.stop.play; 
	}
	stop { 
		inPx.end; 
		ampsPx.end; 
		task.stop; 
	}
	setAmps { |amps|
		var viewVals; 
		dbVals = amps.drop(1).ampdb; 
		lastNDbs = lastNDbs.add(dbVals).keep(avgN.neg.asInteger); 
		
		viewVals = (lastNDbs.mean - midDb) / (rangeDb * 2) + 0.5; 
		this.setViewVals(viewVals);
	}
			// osx specific	
	setViewVals { |vals| ampsV.value_(vals) }

	makeWindow { 
		var flo, font;
		font = Font("Monaco", 8);
		win = SCWindow(this.class.name, Rect(0,0, 790, 260));
		win.view.decorator = flo= FlowLayout(win.bounds, 1@1, 2@2);
		
		SCButton(win, Rect(0,0,60,20))
			.states_([["start"]])
			.action_({ this.start });
		SCButton(win, Rect(0,0,60,20))
			.states_([["stop"]])
			.action_({ this.stop });
		[	[\midDb, [-80,0,\lin,1,-30].asSpec],
			[\rangeDb, [2,40,\lin,1,15].asSpec],
			[\avgN, [1, 50, \lin,1, 5].asSpec],
			[\dt, [0.1, 10, \exp, 0.1, 0.2].asSpec]
		].do { |list| var name, spec; 
			#name, spec = list;
			EZSlider(win, 155@20, name, spec, 
				{|sl| this.perform((name ++ "_").asSymbol, sl.value); }, 
				labelWidth: 50, numberWidth: 25
			);
		};
		
		flo.nextLine;
		
			// only works with 60 for now.
		20.do { |i| 
			var freq = freqs[i * 3 + 1].round(0.1);
			if (freq >= 100, { freq = freq.round(1) }); 
			
			SCStaticText(win, Rect(0,0, 37, 16))
				.background_(Color(0.9,0.9,0.9))
				.string_(freq).align_(0);
		};
		flo.nextLine;
		ampsV = SCMultiSliderView(win, Rect(0,0,13*3*20, 200))
		.indexThumbSize_(13.0).valueThumbSize_(2.0)
		.value_(0.5 ! nBands);
		
		flo.shift(0, -204);
		3.do { |i|
			flo.shift(0, 48); 
			SCStaticText(win, Rect(0,0,13*3*20, 1))
				.background_(Color.white.alpha_([0.3, 0.6, 0.3][i]));
		};
		win.front;
	}
}