/*
 *  AquaFocusBorder.java
 *  de.sciss.gui package
 *
 *  Copyright (c) 2004-2006 Hanns Holger Rutz. All rights reserved.
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either
 *	version 2, june 1991 of the License, or (at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public
 *	License (gpl.txt) along with this software; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *	For further information, please contact Hanns Holger Rutz at
 *	contact@sciss.de
 *
 *
 *  Changelog:
 *		16-Apr-06	created
 */

package de.sciss.gui;

import java.awt.*;
import javax.swing.border.*;

/**
 *  @author		Hanns Holger Rutz
 *  @version	0.70, 15-Apr-06
 */
public class AquaFocusBorder
extends AbstractBorder
{
	private final Insets	insets		= new Insets( 3, 3, 3, 3 );

	private static final Color[] colrFocus		= {
		new Color( 0x96, 0xA5, 0xB7, 0x40 ), new Color( 0x96, 0xA5, 0xB7, 0x80 ),
		new Color( 0x96, 0xA5, 0xB7, 0xC0 ), new Color( 0x96, 0xA5, 0xB7, 0x90 ),
		new Color( 0x96, 0xA5, 0xB7, 0x40 )
	};

	public AquaFocusBorder()
	{
		super();
	}
	
	public Insets getBorderInsets( Component c )
	{
		return new Insets( insets.top, insets.left, insets.bottom, insets.right );
	}
	
	public Insets getBorderInsets( Component c, Insets insets )
	{
		insets.top		= this.insets.top;
		insets.left		= this.insets.left;
		insets.bottom	= this.insets.bottom;
		insets.right	= this.insets.right;
		return insets;
	}
	
	public void paintBorder( Component c, Graphics g, int x, int y, int width, int height )
	{
		if( !c.hasFocus() ) return;

		final Graphics2D g2	= (Graphics2D) g;
							
		g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
		for( int j = 0, k = 1; j < 5; j++, k += 2 ) {
			g2.setColor( colrFocus[ j ]);
			g2.drawRoundRect( x + j, y + j, width - k, height - k, 10 - k, 10 - k );
		}
	}
}