/*
 *  RangeSlider.java
 *  SwingOSC
 *
 *  Copyright (c) 2005-2006 Hanns Holger Rutz. All rights reserved.
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either
 *	version 2, june 1991 of the License, or (at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public
 *	License (gpl.txt) along with this software; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *	For further information, please contact Hanns Holger Rutz at
 *	contact@sciss.de
 *
 *
 *  Changelog:
 *		15-Dec-05	created
 */

package de.sciss.swingosc;

import java.awt.*;
import java.awt.image.*;
import java.net.URL;

/**
 *	A slider that can have variable width.
 *
 *	@author		Hanns Holger Rutz
 *	@version		0.40, 16-Apr-06
 */
public class RangeSlider
extends SliderBase
{
	private static Image progImg 	= null;
	private static Image progImgD = null;

	private Image progImgColr	= null;

	private BufferedImage 	progBufImg 	= null;
	private BufferedImage 	progBufImgD 	= null;
	private Paint			pntProg		= null;
	private Paint			pntProgD		= null;
	private int				recentHeight	= -1;
	
	private static final Insets valInsets = new Insets( 1, 1, 1, 1 );

	public RangeSlider()
	{
		super( true, false, true, false );

		setKnobWidth( 1.0f );
		
		if( progImg == null ) {
			try {
				URL url = getClass().getResource( "aquaprog.png" );
				progImg = Toolkit.getDefaultToolkit().getImage( url );
				MediaTracker mt = new MediaTracker( this );
				mt.addImage( progImg, 0 );
				mt.waitForAll();
			}
			catch( InterruptedException e1 ) {}
			catch( NullPointerException e1 ) { System.err.println( e1 );}
		}
	}

	public void dispose()
	{
		if( progImgColr != null ) {
			progImgColr.flush();
			progImgColr = null;
		}
		if( progBufImg != null ) {
			progBufImg.flush();
			progBufImg = null;
		}
		if( progBufImgD != null ) {
			progBufImgD.flush();
			progBufImgD = null;
		}
		pntProg		= null;
		pntProgD		= null;
		
		super.dispose();
	}

	public void setKnobColor( Color c )
	{
		if( progImgColr != null ) {
			progImgColr.flush();
			progImgColr = null;
		}

		recentHeight = -1;	// triggers re-creation
		
		super.setKnobColor( c );
	}
	
	protected Insets getValueInsets()
	{
		return valInsets;
	}

	private int getKnobPixelWidth( int w )
	{
		final int cw = w - 2;
		return( Math.max( 1, (int) ((knobX + knobWidth) * cw) - (int) (knobX * cw) ));
	}
	
	protected void paintKnob( Graphics2D g2, int w, int h )
	{
		final int kw = getKnobPixelWidth( w );
		final int kh = h - 2;
		final int kx = (int) (knobX * (w - 2)) + 1;
		final int ky = 1;

		if( kh != recentHeight ) {
			if( progBufImg != null ) {
				pntProg = null;
				progBufImg.flush();
				progBufImg = null;
			}
			if( progBufImgD != null ) {
				pntProgD = null;
				progBufImgD.flush();
				progBufImgD = null;
			}
			
			recentHeight = kh;
		}
		
		if( progImg != null ) {
			if( isEnabled() ) {
				if( pntProg == null ) {
					final Graphics2D gImg;
					progBufImg = new BufferedImage( 16, kh, BufferedImage.TYPE_INT_ARGB );
					gImg = progBufImg.createGraphics();
					if( colrKnob != null ) {
						if( progImgColr == null ) {
							progImgColr = createImage( new FilteredImageSource(
									progImg.getSource(), fltKnob ));
							// please don't ask me why
							// we need a media tracker here
							// ; if we omit mit, the image
							// will sometimes not be painted
							waitForImage( progImgColr );
						}
						gImg.drawImage( progImgColr, 0, 0, 16, kh, this );
//						gImg.drawImage( progImgColr, 0, 0, this );
					} else {
						gImg.drawImage( progImg, 0, 0, 16, kh, this );
					}
					gImg.dispose();
					pntProg = new TexturePaint( progBufImg, new Rectangle( 0, 0, 16, kh ));
				}
				g2.setPaint( pntProg );
			} else {
				if( pntProgD == null ) {
					final Graphics2D gImg;
					if( progImgD == null ) {
						progImgD = createImage( new FilteredImageSource(
							progImg.getSource(), fltKnobD ));
						// please don't ask me why
						// we need a media tracker here
						// ; if we omit mit, the image
						// will sometimes not be painted
						waitForImage( progImgD );
					}
					progBufImgD = new BufferedImage( 16, kh, BufferedImage.TYPE_INT_ARGB );
					gImg = progBufImgD.createGraphics();
					gImg.drawImage( progImgD, 0, 0, 16, kh, this );
					gImg.dispose();
					pntProgD = new TexturePaint( progBufImgD, new Rectangle( 0, 0, 16, kh ));
				}
				g2.setPaint( pntProgD );
			}
		}
		g2.fillRect( kx, ky, kw, kh );
		g2.setPaint( colrBdLRSh );
		g2.drawLine( kx, ky + 1, kx, kh - 2 );
		g2.fillRect( kx + kw - 1, ky + 1, 2, kh - 1 );
	}
}
