/*
 
 A CV can be directly used within a Pattern or a Stream.
 Dependents (such as GUI objects or a Control of a Synth or Group
 running on a synthesis server) are connected to the ControlValue
 using a SimpleController linked to the symbol \synch. 
 
 CV-input_(val) as val ranges from 0 to 1 the CV will range
 from its minimum to maximum values. This allows external controllers, such as
 MIDI devices, to be connected to ControlValue's.
 
*/

CV : Stream {
	classvar <>synthEvent, <>groupEvent, <>controlBusEvent, <>actionEvent, <>bufferEvent;
	var <value;
	var <spec;
	

	*initClass {
	// These events create synths, groups, and buses with CVs linked to their controls
		CV.bufferEvent = (
			msg: \wave, 				// \cheby, \wave, \signal
			size: 512,
			play: {
				var func, cv,server, ev;
				ev = currentEnvironment;
				cv = ~cv;
				~server = ~server ?? {Server.default};
				~buffer = Buffer.alloc(~server, ~size, 1);
				~bufnum = ~buffer.bufnum;
				func = (
					sine1: {	 ev[\buffer].sine1(cv.value); },
					cheby: { ev[\buffer].cheby(cv.value);  },
					wave: { ev[\buffer].setn(0, cv.value.clump(2)
						.collect { |a| [a[0], a[1] - a[0] ]}.flatten); },
					signal: {ev[\buffer].setn(0, cv.value);  }
				).at(~msg);
				~controller = SimpleController (cv) .put(\synch, { func.value; ev.get; });
				func.value;
				Task({  0.2.wait; defer { ev.get }; }).play;
			},
				
			stop: {
				~controller.remove;
				~buffer.free;
			},
			get: { | ev |
				if (ev[\display].notNil) {
					Task({ 
						ev[\server].sync;
						ev[\buffer].getn(0, ev[\display].value.size, { | vals|
							if (ev[\msg] != \signal) {
								vals = vals.clump(2).collect { |a| [ a[0], a[0] + a[1] ] }.flatten
							};
							ev[\display].value_(vals)
						});
					}).play;
				}
			}
		);
		CV.synthEvent = (
			lag: 0,
			play: #{ 
				var server, latency, group, addAction;
				var instrumentName, synthLib, desc, msgFunc;
				var msgs, cvs;
				var bndl, ids;
				~server = server = ~server ?? { Server.default };
				group = ~group;
				addAction = ~addAction;
				synthLib = ~synthLib ?? { SynthDescLib.global };
				instrumentName = ~instrument.asSymbol;
				desc = synthLib.synthDescs[instrumentName];					if (desc.notNil) { 
					msgFunc = desc.msgFunc;
					~hasGate = desc.hasGate;
				}{
					msgFunc = ~defaultMsgFunc;
				};
			
				msgs = msgFunc.valueEnvir.flop;
				cvs = ~cvs.value;
				ids = Event.checkIDs(~id);
				if (ids.isNil ) { ids = msgs.collect { server.nextNodeID } };
				bndl = ids.collect { |id, i|
					[\s_new, instrumentName, id, addAction, group]
					 ++ msgs[i]
					 ++ cvs.connectToNode( server, id);
				};
			
				if ((addAction == 0) || (addAction == 3)) {
					bndl = bndl.reverse;
				};
				server.sendBundle(server.latency + ~lag, *bndl);
				~id = ids;
				~isPlaying = true;	 
				~isRunning = true;	 
				NodeWatcher.register(currentEnvironment);
			},
			defaultMsgFunc: #{|freq = 440, amp = 0.1, pan = 0, out = 0| 
				[\freq, freq, \amp, amp, \pan, pan, \out, out] }

		).putAll(Event.partialEvents.nodeEvent);
		
		CV.groupEvent = ( 
			lag: 0,
			play: #{
				var server, group, addAction, ids, bundle, cvs;
				group = ~group;
				addAction = ~addAction;
				~server = server= ~server ?? {Server.default};
				ids = Event.checkIDs(~id);
				if (ids.isNil) { ids = ~id = server.nextNodeID };
				cvs = ~cvs.value;
				if ((addAction == 0) || (addAction == 3) ) {
					ids = ids.asArray.reverse;
				};
				bundle = ids.collect {|id, i|
					cvs.connectToNode(server, id);
					[\g_new, id, addAction, group];
 
				};
				server.sendBundle(server.latency + ~lag, *bundle);
				~isPlaying = true;
				~isRunning = true;
				NodeWatcher.register(currentEnvironment);
			}
		).putAll(Event.partialEvents.nodeEvent)[\hasGate] = false;
		CV.controlBusEvent = (
			play: #{ var cvs, index, server, bundle, size;
				~server = server = ~server ?? {Server.default};
				cvs = ~cvs.asArray;
				index = ~index;
				if (index.isNil) {
					~bus = Bus.control(server,cvs.size);
					index =  ~index =  ~bus.index;
				} {
					~bus = Bus(\control, index,cvs.size, server)
				};
				~bus.setControls(cvs, cvs.size);

			},
			
			stop: #{
				~bus.free;
			}
		);
		CV.actionEvent = ( asEventStreamPlayer: #{|ev| ev } );
					 
	}

	*new {arg spec = \unipolar, default; 			
		^super.new.spec_(spec,default);
	}


	action_ { | function |
		^SimpleController(this) .put(\synch, function);
	}

	value_ { arg val;
	
		value = spec.constrain(val);
//		action.value(this);
		this.changed(\synch);
	}
	
	input_	{ arg in; this.value_(spec.map(in)); }
	
	input 	{ ^spec.unmap(value) }

	asInput 	{arg val; ^spec.unmap(val) }
	
	spec_ 	{ arg s, v; 
				spec = s.asSpec; 
				this.value_(v ? spec.default); 
			}	
	sp	{ arg default= 0, lo = 0, hi=0, step = 0, warp = 'lin';
		this.spec = ControlSpec(lo,hi, warp, step, default);
	}
	db	{ arg default= 0;
		this.spec = ControlSpec(-100,20, 'lin',1, default);
	}
	
	split {
		^value.collect { |v| CV(spec, v) }
	}
// Stream and Pattern support

	next { ^value }
	reset {}
	asStream { ^this }
	embedInStream { value.yield }

// connecting to a Server	
// msg = ['/n_set', nodeID, name] or ['/c_set', busIndex]
	connectToServer { arg server, msg, modifierOrCV;
		^SimpleController (this) .put(\synch, 
					{ server.sendBundle(server.latency, (msg ++ modifierOrCV.value(this)) ) });
	}

	connectToNodeProxy { arg np, name, modifierOrCV;
		^SimpleController (this) .put(\synch, { np.set(name, modifierOrCV.value(this)) });
	}
		

	draw { |win, name =">"|
		var labelView, slider, ctl;
		if (value.isKindOf(Array) ) {
			~multislider.value(win, name, this);
		} {
			~nslider.value(win, name, this);
		};
	}

		
}

BufferCV : CV {
	var <>buffer;
	var <>server;
	
	*new { | spec, default|
		var cv;
		cv = super.new(spec, default);
		cv.value = cv.value.asArray;
		^cv
	}
	
	draw { |win, name =" a buffer"|
	var slider, w, rect;
		w = SCVLayoutView(win, Rect(0,0,80, 200));
		SCStaticText(w, Rect(0,0,80, 20))
			.string_(name)
			.align_(\right);
		rect = Rect(0,0, ~labelW, 30);
		SCButton(w, rect )
			.states_([ ["send wave\n", Color.black,Color.gray(1,0)] ])
			.action_({ this.input_(slider.value); 
				buffer.setn(0, this.value.clump(2).collect { |v | [v[0], v[1] - v[0] ] }.flatten(1)); });
		SCButton(w, rect)
			.states_([ ["get wave\n", Color.black,Color.gray(1,0)] ])
			.action_({ buffer.getn(0, this.value.size, 
					{ |vals| this.value = vals.clump(2).collect { |v | [v[0], v[0] + v[1] ] }.flatten(1) }
				) 
			});
		SCButton(w, rect )
			.states_([ ["send signal\n", Color.black,Color.gray(1,0)] ])
			.action_({ this.input_(slider.value); buffer.setn(0, this.value); });
		SCButton(w, rect )
			.states_([ ["get signal\n", Color.black,Color.gray(1,0)] ])
			.action_({ buffer.getn(0, this.value.size, { |vals| this.value = vals}) 
			});
		SCButton(w, rect)
			.states_([ ["send spect\n", Color.black,Color.gray(1,0)] ])
			.action_({ this.input_(slider.value); 
				buffer.sine1((slider.value.copyRange(0, (slider.size/2).asInteger) * 96 - 96).dbamp )
			});
		SCButton(w, rect )
			.states_([ ["get spect\n", Color.black,Color.gray(1,0)] ])
			.action_({ buffer.getn(0, this.value.size, { |vals| this.value = vals}) 
			});
		rect = ~sliderRect.value;
		rect = rect.resizeBy(~labelW * -1,~vh);
		slider = SCMultiSliderView(win, rect).connect(this);
		slider.thumbSize = 520/this.value.size;
		slider.valueThumbSize = 2;
	}
	
	pause {}
	resume {}
	play { 
		server = server ?? {Server.default };
		buffer = Buffer.alloc(server, value.size, 1);
	}
	stop { buffer.free; buffer = nil }
	
	bufnum { ^buffer.bufnum }
}