GUIEvent {
	classvar <>osx;
	*initClass {
		osx = 
		(
			h:		20,				// default height
			w: 		800,				// default width
			labelW:	80,				// width of a label and numberbox

			vh:		200,				// height of a vertical slider
			vw: 		20, 				// width of a vertical slider
			
			// functions that specify default dimensions of various views
			
			numericalRect: 		{ Rect(0,0, ~labelW, ~h) },
			smallNumericalRect: 	{ Rect(0,0, ~labelW/2, ~h) },
			sliderRect:			{ Rect(0,0, ~w, ~h) },
			vsliderRect:			{ Rect(0,0, ~vw, ~vh) },
			msliderRect:			{ Rect(0,0, ~w, ~vh) },
			tdsliderRect:			{ Rect(0,0, ~vh, ~vh) },
			movieRect:			{ Rect(0,0, 240, 320) },
			labelRect:			{ Rect(0,0, ~labelW, ~h) },
			radioButtonsRect: 		{ Rect(0,0,10,~h) },

			label: { |win, name, cv, rect |
				rect = rect ?? ~labelRect; 
				SCStaticText(win, rect).align_(\right).string_(name)
			},
			
			popup: { |win, name, cv, rect |
				rect = rect ?? ~labelRect; 
				SCPopUpMenu(win, rect).items_(name)
			},
			
			list: { |win, name, cv, rect |
				rect = rect ?? ~labelRect; 
				SCListView(win, rect).items_(name)
			},
			
			text: { |win, name, cv, rect |
				rect = rect ?? ~sliderRect; 
				SCTextView(win, rect).string_(name.asString)
			},
		
			movie: { |win, name, cv, rect |
				rect = rect ?? ~movieRect; 
				SCMovieView(win, rect)
			},
		
			numberBox: { |win, name, cv, rect |
				rect = rect ?? ~numericalRect; 
				SCNumberBox(win,rect).connect(cv)
			},

			slider: { |win, name, cv, rect|
				rect = rect ?? ~sliderRect;
				SCSlider(win,rect).connect(cv)
			},

			rangeSlider: { |win, name, cv, rect|
				rect = rect ?? ~sliderRect;
				SCRangeSlider(win,rect).connect(cv);
			},

			twodSlider: { |win, name, cv, rect|
				rect = rect ?? ~tdsliderRect;
				SC2DSlider(win,rect).connect(cv);
			},

			smallNumberBox: { |win, name, cv, rect|
				rect = rect ?? ~smallNumericalRect;
				cv.asArray.do{ |cv|
					SCNumberBox(win,rect).connect(cv)
				}
			},

			numerical: { |win, name, cv, rect |
				rect = rect ?? ~numericalRect; 
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				cv.asArray.do{ |cv|
					SCNumberBox(win,rect).connect(cv)
				}
			},
			
			smallNumerical: { |win, name, cv, rect|
				rect = rect ?? ~smallNumericalRect;
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				cv.asArray.do{ |cv|
					SCNumberBox(win,rect).connect(cv)
				}
			},
			
			vslider: { |win, name, cv, rect|
				rect = rect ?? ~vsliderRect;
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				cv.asArray.do{ |cv|
					SCSlider(win,rect).connect(cv)
				}
			},

			vrslider: { |win, name, cv, rect|
				rect = rect ?? ~vsliderRect;
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				cv.asArray.do{ |cv|
					SCRangeSlider(win,rect).connect(cv)
				}
			},

			nslider: { |win, name, cv, rect|
				rect = rect ?? ~sliderRect;
				rect = rect.resizeBy(~labelW * -2,0);
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				SCSlider(win,rect).connect(cv);
				SCNumberBox(win,~numericalRect.value).connect(cv);
			},

			tdslider: { |win, name, cv, rect|
				var slider;
				rect = rect ?? ~tdsliderRect;
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				slider = SC2DSlider(win,rect);
				if (cv.notNil) { slider.connect(cv) };
				slider;
			},
			
			rslider: { |win, name, cv, rect|
				var slider;
				rect = rect ?? ~sliderRect;
				rect = rect.resizeBy(~labelW * -1,0);
				SCStaticText(win, ~labelRect.value).string_(name).align_(\right);
				slider = SCRangeSlider(win,rect);
				if (cv.notNil) { slider.connect(cv) };
				slider;
			},

			multislider: { |win, name, cv, rect|
				var size, slider;
				rect = rect ?? ~msliderRect;
				rect = rect.resizeBy(~labelW * -1,0);
				~label.value(win, name); 
				slider = SCMultiSliderView(win, rect);
				if (cv.notNil) { slider.connect(cv) };
				slider;
			},

			radiobuttons:	{ |win, name, cv|
				var buttons, link,  size, preVal;
				
				size = cv.spec.maxval;
				preVal = cv.value;
				buttons = Array.fill(size).collect { | n, i  |  
					SCButton(win, ~radioButtonsRect.value)
						.states_([
							["", Color.red, Color.grey(0.4)],
							["", Color.red, Color.blue(0.3)]
						])
				};
				buttons.do { |bt, i|
					bt.action_({ | ...x|
						if (preVal > 0) { buttons[preVal - 1].value_(0) }; 
						if (bt.value == 0) 
							{ preVal = 0; cv.value = 0 } 
							{ cv.value = preVal = i + 1};  
					
					})
				};
				
				link = SimpleController(cv).put(\synch, {
					if (preVal > 0) { buttons[preVal - 1].value = 0 };
					preVal = cv.value;
					if (preVal > 0) { buttons[preVal - 1].value = 1 }
				});
				buttons[0].onClose_({link.remove});
				
			}
		)	
	}
}