Preset : Display {

	var <>presetControl;
	var <>goalControl;
	var <>interpolateControl;
	
	*initClass {
			Class.initClassTree(ControlEnvironment);
			Class.initClassTree(Display);
	}
	init { arg args;
		clientgroup = ClientGroup.new;
		presetControl = ControlValue([-1,-1,'lin',1,-1]).name_("preset");
		presetControl.action_({
			this.setPreset(presetControl.value.asInteger);
		});
		
		goalControl = ControlValue([-1,-1,'lin',1,-1]).name_("     target");
		goalControl.action_({
			this.setInterpolationTarget(goalControl.value.asInteger);
		});
			
		interpolateControl = ControlValue(\unipolar).name_("interp");
		interpolateControl.action_({
			this.interpolate(interpolateControl.value);
		});

		args.do({ | c | this.put(c.name, c) });
		valueItems = args;
		guiItems = this.presetGUI ++ this.interpolateGUI ++ args;
		stopOnClose = true
	}
		
	
	presetGUI {
		^[
					clientgroup,
				[	SimpleButton({this.addPreset; }).states_([["+", Color.black,Color.grey] ]),
					SimpleButton({this.removePreset(presetControl.value); }).states_([["-", Color.black,Color.grey] ]),
					SimpleButton({this.loadDialog}).states_([["ld", Color.black,Color.grey] ]),
					SimpleButton({this.saveDialog}).states_([["sv", Color.black,Color.grey] ]),
				
					Numerical(presetControl, 36),
					SimpleButton({this.nextPreset(-1)}).states_([["<", Color.black,Color.grey] ]),
					SimpleButton({this.nextPreset(1)}).states_([[">", Color.black,Color.grey] ]),
				
				]
				
			];
	}


	interpolateGUI {
		^ [		[	Numerical(goalControl, 46),
					SimpleButton({this.nextGoal(-1)}).states_([["<", Color.black,Color.grey] ]),
					SimpleButton({this.nextGoal(1)}).states_([[">", Color.black,Color.grey] ])
				],
				 NSlider(interpolateControl)
		]
	}				

	draw { arg win;
		SimpleController(win)
			.put(\close, { 
				views.do({arg v;  v.remove; }); 	// before stopping
				views = nil;
				if (stopOnClose == true, { 		// so deferred synch messages
					this.stop; 				// don't update deleted
				});							// dependants
			});

		guiItems
		.do({arg v;
			v.asArray.do({ arg v; views = views.add(v.draw(win)); }); 
			win.view.decorator.nextLine;
			v.update(this,\synch); 
		});
	}	
				
	loadDialog { 		
		File.openDialog(nil, {arg path;  this.load(path) })
	}
	
	saveDialog {
		File.saveDialog(nil, nil, { arg path; this.save(path)})
	}
	
	updateControlRanges {
			presetControl.spec_([0,presets.size-1,'lin',1,presets.size-1]);
			goalControl.spec_([0,presets.size-1,'lin',1]);
	}

	addPreset { 	
		super.addPreset; 
		this.updateControlRanges;
	}
	
	removePreset { arg preset;	
		super.removePreset(preset);
		this.updateControlRanges;			
	}
	
	load { arg path;
		super.load(path);
		this.updateControlRanges;
	}
	
	setInterpolationTarget { arg index;
		super.setInterpolationTarget(index);
		interpolateControl.value_(0);
	}

	nextPreset { arg inc = 1;
		presetControl.value_(presetControl.value + inc % presets.size)
	}
	
	nextGoal { arg inc = 1;
		goalControl.value_(goalControl.value + inc % presets.size)
	}
	

}