/*
	SuperCollider real time audio synthesis system
 Copyright (c) 2002 James McCartney. All rights reserved.
	http://www.audiosynth.com
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//some UGens by Nick Collins 
//SLUGens released under the GNU GPL as extensions for SuperCollider 3, by Nick Collins, http://www.cus.cam.ac.uk/~nc272/
//ideas: SortBuf2- online quicksort saving tree recursion state? 

#include "SC_PlugIn.h"

static InterfaceTable *ft; 

struct SortBuf : public Unit   //Karplus Strong with sorting algorithm on a buffer 
{
	uint32 mBufNum,mBufSize;
	int mOutStep;
	int mSorti, mSortj, mSortdone;
	
	float* mBufCopy;
};

struct GravityGrid : public Unit    
{
	float x[9],y[9],velx,vely,posx,posy;
	float * m_weights;
};


//bufnum, 0 capturetrigger 1, duration of repeat 2, ampdropout 3
struct Breakcore : public Unit    
{
	uint32 mBufNum,mBufSize;	
	float * mBuf;
	int captureon,captureneeds,repeatpos,capturepos;
	float m_prevtrig;
	//int duration;
};


//find maximum value within last x blocks
struct Max : public Unit {
	int m_blocks, m_counter,m_last;
	float m_max;     
	float* m_blockmaxes;
};

//print in lang a value every x blocks
struct PrintVal : public Unit {
	int m_count,m_block,m_id;
};

//from http://www.musicdsp.org/showone.php?id=97
struct EnvDetect : public Unit {
	float m_env; //m_atk,m_rel,
};

struct FitzHughNagumo : public Unit    
{
	float u,w;
};

struct DoubleWell : public Unit    
{
	float x,y,t;
};

struct WeaklyNonlinear : public Unit    
{
	float x,y,t;
};

extern "C" {  
	
	void SortBuf_next_k(SortBuf *unit, int inNumSamples);
	void SortBuf_Ctor(SortBuf* unit);
	//void SortBuf_Dtor(SortBuf* unit);
	
	void GravityGrid_next_k(GravityGrid *unit, int inNumSamples);
	void GravityGrid_Ctor(GravityGrid* unit);
	
	void Breakcore_next_k(Breakcore *unit, int inNumSamples);
	void Breakcore_Ctor(Breakcore* unit);
	
	void Max_next(Max *unit, int inNumSamples);
	void Max_Ctor(Max* unit);
	void Max_Dtor(Max* unit);
	
	void PrintVal_next(PrintVal *unit, int inNumSamples);
	void PrintVal_Ctor(PrintVal* unit);
	
	void EnvDetect_next(EnvDetect *unit, int inNumSamples);
	void EnvDetect_Ctor(EnvDetect* unit);
	
	void FitzHughNagumo_next_k(FitzHughNagumo *unit, int inNumSamples);
	void FitzHughNagumo_Ctor(FitzHughNagumo* unit);
	
	void DoubleWell_next_k(DoubleWell *unit, int inNumSamples);
	void DoubleWell_Ctor(DoubleWell* unit);
	
	void WeaklyNonlinear_next_k(WeaklyNonlinear *unit, int inNumSamples);
	void WeaklyNonlinear_Ctor(WeaklyNonlinear* unit);
	
}


void SortBuf_Ctor( SortBuf* unit ) {
	
	SETCALC(SortBuf_next_k);
	
	World *world = unit->mWorld;
    
	uint32 bufnum = (uint32)ZIN0(0);
	if (bufnum >= world->mNumSndBufs) bufnum = 0;
	unit->mBufNum=bufnum;
	
	//printf("%d \n",bufnum);			
	
	SndBuf *buf = world->mSndBufs + bufnum; 
	
	unit->mBufSize = buf->samples;
	
	unit->mBufCopy= buf->data; //(float*)RTAlloc(unit->mWorld, unit->mBufSize * sizeof(float));
	
	//initialise to copy
	//int i=0;
	//	for(i=0; i<unit->mBufSize;++i) 
	//		unit->mBufCopy[i]=buf->data[i];
	//	
	unit->mSorti=unit->mBufSize - 1;
	unit->mSortj=1;
	unit->mSortdone=0;
	unit->mOutStep=1; //allows one run through before it sorts
}

//void SortBuf_Dtor(SortBuf *unit)
//{
//	//RTFree(unit->mWorld, unit->mBufCopy);
//}
	
void SortBuf_next_k( SortBuf *unit, int inNumSamples ) {
	
	float *out = ZOUT(0);
	
	int sortrate= (int)ZIN0(1);
	float reset= (float)ZIN0(2);
	//float freq= (float)ZIN0(3);
	
	//printf("%d \n",unit->mWorld->mBufCounter);			
	
	int outstep=unit->mOutStep;
	
	//output, doing sorts if requested
	int bufsize=unit->mBufSize;
	
	float * data= unit->mBufCopy;
	
	int sorti, sortj, sortdone; 
	float temp;
	sorti=unit->mSorti;
	sortj=unit->mSortj;
	sortdone=unit->mSortdone;
	
	//printf("sortparams %d %d %d \n",sorti,sortj,sortdone);
	
	if(reset>0.5 && (unit->mWorld->mBufCounter>10)) {
		//		//printf("reset!\n");
		reset=0.0;
		//SndBuf *buf = unit->mWorld->mSndBufs + unit->mBufNum; 
		//float *source=buf->data;
		
		//for(int i=0; i<bufsize;++i)
		//	data[i]=source[i];
		
		sorti=(bufsize - 1);
		sortj=1;
		sortdone=0;
	}
	
	
	for (int j=0; j<inNumSamples;++j)
	{
		
		if (outstep ==0) { //SORT
			
			int sorttodo=sortrate;
			
			//printf("sort! %d \n", sorttodo);
			
			while((sorttodo>0) && (sortdone==0)){
				
				if (data[sortj-1] > data[sortj]) {
					temp = data[sortj-1];
					data[sortj-1] = data[sortj];
					data[sortj] = temp;
				}
				
				++sortj;
				
				if(sortj>sorti) {
					--sorti;
					sortj=1;
					if(sorti<0)
					{
						sortdone=1;
						//printf("sort finished %d \n",sortdone);			
						
					}
					
				}
				
				sorttodo--;
			}
			
		}
		
		//printf("outstep %d bufsize %d \n",outstep,bufsize);
		
		//else output
		ZXP(out) = data[outstep];
		
		outstep= (outstep+1)%bufsize;
	}
	
	unit->mSorti=sorti;
	unit->mSortj=sortj;
	unit->mSortdone=sortdone;
	
	unit->mOutStep=outstep;
}


void GravityGrid_Ctor(GravityGrid* unit) {
	int i;
	
	World *world = unit->mWorld;
    
	for(i=0;i<9;++i) {
		
		unit->x[i]=(i%3)-1;
		
		unit->y[i]=1-(i/3);
		
		//printf("i %d x %f y %f \n",i,unit->x[i],unit->y[i]);
		
	}
	
	unit->velx=0.0;
	unit->vely=0.0;
	unit->posx=0.0;
	unit->posy=0.0;
	
	//for (i=0;i<9;++i)
	//	unit->m_weights[i]= g_weights[i];
	unit->m_weights=NULL; 
	
	uint32 bufnum = (uint32)ZIN0(4);
	
	if (!(bufnum > world->mNumSndBufs || bufnum<0)) {
		SndBuf *buf = world->mSndBufs + bufnum; 
		
		if(buf->samples==9) {
			unit->m_weights= buf->data;
			
		}
	}
	
	SETCALC(GravityGrid_next_k);
	
}


//can't have independent x and y, causes trouble- zero has a pusher at it with infinite size! 
//must use euclidean distances
void GravityGrid_next_k(GravityGrid *unit, int inNumSamples) {
	
	float *out = ZOUT(0);
	float accelx,accely, xdiff,ydiff, rdiff; //hyp;
	
	int reset= (int)ZIN0(0);
	float rate= (float)ZIN0(1);
	
	float velx, vely, posx, posy; 
	
	velx= unit->velx;
	vely=unit->vely;
	posx=unit->posx;
	posy=unit->posy;
	
	if(reset) {
		//RGen& rgen = *unit->mParent->mRGen;
		
		//printf("reset! \n");
		
		posx= (float)ZIN0(2); //0.8*(2*rgen.frand() - 1.0); //never right near edge
		posy= (float)ZIN0(3); //0.8*(2*rgen.frand() - 1.0);
		
		velx=0.0;
		vely=0.0;
		
		if (posx>0.99) posx=0.99;
		if (posx<-0.99) posx=-0.99;
		if (posy<-0.99) posy=-0.99;
		if (posy>0.99) posy=0.99;
		
		
	}
	
	float *x,*y, *w;
	
	x=unit->x;
	y=unit->y;
	w=unit->m_weights;
	
	if (w) {
		for (int j=0; j<inNumSamples;++j) {
			
			accelx=0.0;
			accely=0.0;
			
			//could use this to decide on run function but can't be bothered, negligible on loop performance and 
			//keeps code in one place for edits
			
			for(int i=0;i<9;++i) {
				
				if(i!=4) {
					
					xdiff= ((posx)-(x[i]));
					ydiff= ((y[i])-(posy));
					
					//hyp= 1.0/sqrt(xdiff*xdiff+ydiff*ydiff);
					
					rdiff= w[i]*0.0001*sqrt(xdiff*xdiff+ydiff*ydiff); //(0.0001/(xdiff*xdiff+ydiff*ydiff));
					
					accelx+= rdiff*xdiff; //rdiff*(xdiff*hyp);
					
					accely+= rdiff*ydiff; //rdiff*(ydiff*hyp);
					
				}
				
			}
			
			
			velx=velx+accelx;
			vely=vely+accely;
			
			posx=posx+(rate*(velx));
			posy=posy+(rate*(vely));
			
			//assumes fmod works correctly for negative values
			if ((posx>1.0) || (posx<-1.0)) posx=fabs(fmod((posx-1.0),4.0)-2.0)-1.0;
			if ((posy>1.0) || (posy-1.0)) posy=fabs(fmod((posy-1.0),4.0)-2.0)-1.0;
			
			//printf("%f %f %f %f %f %f %f \n",accelx, accely, unit->velx,unit->vely,unit->posx,unit->posy, unit->posy*unit->posy+ unit->posx*unit->posx);
			
			float sign; 
			if (fabs(posx) < 0.0000001) sign= 1;
			else
				sign= (posx/fabs(posx));
			
			ZXP(out) = sign*0.5*(posy*posy+ posx*posx);
			
			//printf("%f \n",(unit->posx/fabs(unit->posx))*0.5*(unit->posy*unit->posy+ unit->posx*unit->posx));
		}
	}
	else {
		
		for (int j=0; j<inNumSamples;++j) {
			
			accelx=0.0;
			accely=0.0;
			
			//could use this to decide on run function but can't be bothered, negligible on loop performance and 
			//keeps code in one place for edits
			
			for(int i=0;i<9;++i) {
				
				if(i!=4) {
					
					xdiff= ((posx)-(x[i]));
					ydiff= ((y[i])-(posy));
					
					rdiff= 0.0001*sqrt(xdiff*xdiff+ydiff*ydiff);
					accelx+= rdiff*xdiff; 
					accely+= rdiff*ydiff; 
					
				}
				
			}
			
			velx=velx+accelx;
			vely=vely+accely;
			
			posx=posx+(rate*(velx));
			posy=posy+(rate*(vely));
			
			//assumes fmod works correctly for negative values
			if ((posx>1.0) || (posx<-1.0)) posx=fabs(fmod((posx-1.0),4.0)-2.0)-1.0;
			if ((posy>1.0) || (posy-1.0)) posy=fabs(fmod((posy-1.0),4.0)-2.0)-1.0;
			
			//printf("%f %f %f %f %f %f %f \n",accelx, accely, unit->velx,unit->vely,unit->posx,unit->posy, unit->posy*unit->posy+ unit->posx*unit->posx);
			
			float sign; 
			if (fabs(posx) < 0.0000001) sign= 1;
			else
				sign= (posx/fabs(posx));
			
			ZXP(out) = sign*0.5*(posy*posy+ posx*posx);
			
			//printf("%f \n",(unit->posx/fabs(unit->posx))*0.5*(unit->posy*unit->posy+ unit->posx*unit->posx));
		}
		
	}
	
	
	unit->velx=velx;
	unit->vely=vely;
	unit->posx=posx;
	unit->posy=posy;
	
	
}



void Breakcore_Ctor( Breakcore* unit ) {
	
	SETCALC(Breakcore_next_k);
	
	World *world = unit->mWorld;
    
	uint32 bufnum = (uint32)ZIN0(0);
	if (bufnum >= world->mNumSndBufs) bufnum = 0;
	unit->mBufNum=bufnum;
	
	//printf("Breakcore activated with bufnum %d \n",bufnum);			
	
	SndBuf *buf = world->mSndBufs + bufnum; 
	
	unit->mBufSize = buf->samples;
	
	unit->mBuf = buf->data;
	
	unit->captureon= 0;
	unit->captureneeds= 1000;
	unit->capturepos= 0;
	unit->repeatpos=0;
	unit->m_prevtrig=0;
	
	//unit->duration=1000;
}


//bufnum, 0 capturein 1, capturetrigger 2, duration of repeat 3, ampdropout 4

void  Breakcore_next_k(  Breakcore *unit, int inNumSamples ) {
	
	float *out = ZOUT(0);
	
	float curtrig= ZIN0(2);
	//int duration= unit->duration; //(int)ZIN0(3);
	//float ampdropout= (float)ZIN0(3);
	float * in = ZIN(1);
	
	//printf("duration %d curtrig %f\n",duration, curtrig);
	
	
	//int bufsize=unit->mBufSize;
	
	float * data= unit->mBuf;
	   
	
	
	if (unit->m_prevtrig <= 0.f && curtrig > 0.f){
		
		//printf("capture \n");
		
		unit->captureon=1;
		unit->captureneeds=(int)ZIN0(3);
		
		//unit->duration=(int)ZIN0(3);
		//duration= unit->duration; 
		//unit->captureneeds=duration; //can't be less than 64  
		unit->capturepos=0;
		unit->repeatpos=0;
	} 
	
	int captureon,captureneeds,repeatpos,capturepos;
	
	captureon= unit->captureon;
	captureneeds= unit->captureneeds;
	capturepos= unit->capturepos;
	repeatpos= unit->repeatpos;
	
	for (int j=0; j<inNumSamples;++j)
	{
		
		if(captureon==1) {
			
			data[capturepos]=in[j];
			
			capturepos++;
			
			if(capturepos==captureneeds) {
				unit->capturepos=0;
				unit->captureon=0;
				captureon=0;
			}
		};
		
		
		ZXP(out)=data[repeatpos];
		
		repeatpos= (repeatpos+1)%captureneeds;
		
	}
	
	unit->capturepos= capturepos;
	unit->repeatpos= repeatpos;
	
	unit->m_prevtrig = curtrig;
	
}



void Max_Ctor( Max* unit ) {
	
	SETCALC(Max_next);
	
	int msamp= (int) ZIN0(1);
	//integer division
	unit->m_blocks= sc_max(msamp/(unit->mWorld->mFullRate.mBufLength),1);
	
	//printf("%d \n",unit->m_blocks);
	
	unit->m_blockmaxes= (float*)RTAlloc(unit->mWorld, unit->m_blocks * sizeof(float));
	//initialise to zeroes
	for(int i=0; i<unit->m_blocks; ++i)
		unit->m_blockmaxes[i]=0.f;
	
	unit->m_counter=0;
	unit->m_last=unit->m_blocks-1;   
	unit->m_max=0.0;
	
}

void Max_Dtor(Max *unit) {
	RTFree(unit->mWorld, unit->m_blockmaxes);
}

void Max_next( Max *unit, int inNumSamples ) {
	
	int j;
	float *in = ZIN(0);
	//float *out = ZOUT(0);
	float max=0;
	
	//only to be used at .kr
	//printf("samp to calc %d", inNumSamples);
	
	int ksamps = unit->mWorld->mFullRate.mBufLength;
	
	//find max this block
	for(j=0; j<ksamps; ++j) {
		float next= fabs(ZXP(in)); 
		//printf("next %f\t",next);
		if(next>max) max=next;
	}
	
	//printf("max %f  unit->m_max  %f \t",max, unit->m_max);
	
	//now check max over last m_blocks
	
	if(max>(unit->m_max)) {
		
		unit->m_max=max;
		unit->m_last=unit->m_counter;
		unit->m_blockmaxes[unit->m_counter]=max;
	}
	else {
		
		unit->m_blockmaxes[unit->m_counter]=max;
		
		//if max being lost
		if(unit->m_counter==unit->m_last) {//calc new max position
			max=0;
			int last=0;
			
			for(j=0; j<unit->m_blocks; ++j){
				float test= unit->m_blockmaxes[j]; 
				
				if(test>max) {max=test; last=j;}
			}
			
			unit->m_max=max;
			unit->m_last=last;
			
		}
		
	}
	
	max=unit->m_max;
	
	unit->m_counter=(unit->m_counter+1)%(unit->m_blocks);
	
	//control rate, just one out value
	//output max
	//for(j=0; j<inNumSamples; ++j) {
	ZOUT0(0)=max; 
	//}
	
}

void PrintVal_Ctor(PrintVal* unit) {
	SETCALC(PrintVal_next);
	
	unit->m_block= (int) ZIN0(1);
	
	unit->m_id= (int) ZIN0(2);
	
	unit->m_count=0;
}


void PrintVal_next(PrintVal *unit, int inNumSamples) {
	
	if(unit->m_count==0) {printf("%d  %f\n",unit->m_id,ZIN0(0));}
	
	unit->m_count=(unit->m_count+1)%(unit->m_block);
}


void EnvDetect_Ctor(EnvDetect* unit) {
	SETCALC(EnvDetect_next);
	
	//	unit->m_atk= (float) exp(-1/(SAMPLERATE*ZIN0(1)));
	//	
	//	unit->m_rel= (float) exp(-1/(SAMPLERATE*ZIN0(2)));
	//	
	//printf("atk %f rel %f\t",unit->m_atk,unit->m_rel);
	
	unit->m_env=0.0;
}


void EnvDetect_next(EnvDetect *unit, int inNumSamples) {
	
	float *in = ZIN(0);
	float *out = ZOUT(0);
	
	//int ksamps = unit->mWorld->mFullRate.mBufLength;
	
	float envelope= unit->m_env;
	//float ga= unit->m_atk;
	//	float gr= unit->m_rel;
	//	
	
	float ga= (float) exp(-1/(SAMPLERATE*ZIN0(1)));
	float gr= (float) exp(-1/(SAMPLERATE*ZIN0(2)));
	
	//find max this block
	for(int j=0; j<inNumSamples; ++j) {
		float next= fabs(ZXP(in)); 
		//printf("next %f\t",next);
		
		if(envelope<next){
			envelope*=ga;
			envelope+= (1-ga)*next;
		}
		else {
			envelope*=gr;
			envelope+= (1-gr)*next;
		}
		
		ZXP(out)= envelope;
		
	}
	
	unit->m_env=envelope;
	
}




void FitzHughNagumo_Ctor(FitzHughNagumo* unit) {
	
	unit->u=0.0;
	unit->w=0.0;
	
	SETCALC(FitzHughNagumo_next_k);
	
}


void FitzHughNagumo_next_k(FitzHughNagumo *unit, int inNumSamples) {
	
	float *out = ZOUT(0);
	
	int reset= (int)ZIN0(0);
	float urate= (float)ZIN0(1);
	float wrate= (float)ZIN0(2);
	float b0= (float)ZIN0(3);
	float b1= (float)ZIN0(4);
	
	float u,w; 
	
	u= unit->u;
	w=unit->w;
	
	if(reset) {
		
		u= (float)ZIN0(5);
		w= (float)ZIN0(6);
	}
	
	for (int j=0; j<inNumSamples;++j) {
		
		//the naive Euler update, could compare Runge-Kutta later
		float dudt= urate*(u-(0.33333*u*u*u)-w);
		float dwdt= wrate*(b0+b1*u-w);
		
		u+=dudt;
		w+=dwdt;
		
		//assumes fmod works correctly for negative values
		if ((u>1.0) || (u<-1.0)) u=fabs(fmod((u-1.0),4.0)-2.0)-1.0;
		//if ((posy>1.0) || (posy-1.0)) posy=fabs(fmod((posy-1.0),4.0)-2.0)-1.0;
		
		ZXP(out) = u;
		
		//printf("%f \n",(unit->posx/fabs(unit->posx))*0.5*(unit->posy*unit->posy+ unit->posx*unit->posx));
	}
	
	unit->u=u;
	unit->w=w;
	
}







void DoubleWell_Ctor(DoubleWell* unit) {
	
	unit->x=0.0;
	unit->y=0.0;
	unit->t=0.0;
	
	SETCALC(DoubleWell_next_k);
	
}


void DoubleWell_next_k(DoubleWell *unit, int inNumSamples) {
	
	float *out = ZOUT(0);
	
	int reset= (int)ZIN0(0);
	float xrate= (float)ZIN0(1);
	float yrate= (float)ZIN0(2);
	float F= (float)ZIN0(3);
	float w= (float)ZIN0(4);
	float delta= (float)ZIN0(5);
	
	float x,y,t; 
	
	x= unit->x;
	y=unit->y;
	t=unit->t;
	
	if(reset) {
		
		x= (float)ZIN0(6);
		y= (float)ZIN0(7);
		t=0;
	}
	
	//Runge Kutta? would require four cos calls
	for (int j=0; j<inNumSamples;++j) {
		
		//the naive Euler update, could compare Runge-Kutta later
		float dxdt= xrate*y;
		
		//could make Fcos term another ar or kr input itself? 
		//float dydt= yrate*(-delta*y+ x - x*x*x + F*cos(w*t));
		
		//Runge-Kutta
		float temp,z,k1,k2,k3,k4;
		temp= x - x*x*x + F*cos(w*t);
		z= y;
		k1= yrate*(-delta*z+ temp);
		z=y+0.5*k1;
		k2= yrate*(-delta*z+ temp);
		z=y+0.5*k2;
		k3= yrate*(-delta*z+ temp);
		z=y+k3;
		k4= yrate*(-delta*z+ temp);
		
		float dydt= 0.166667*(k1+2*k2+2*k3+k4);
		
		t=t+1;
		x+=dxdt;
		y+=dydt;
		
		//assumes fmod works correctly for negative values
		if ((x>1.0) || (x<-1.0)) x=fabs(fmod((x-1.0),4.0)-2.0)-1.0;
		//if ((posy>1.0) || (posy-1.0)) posy=fabs(fmod((posy-1.0),4.0)-2.0)-1.0;
		
		ZXP(out) = x;
		
		//printf("%f \n",(unit->posx/fabs(unit->posx))*0.5*(unit->posy*unit->posy+ unit->posx*unit->posx));
	}
	
	unit->x=x;
	unit->y=y;
	unit->t=t;
}





void WeaklyNonlinear_Ctor(WeaklyNonlinear* unit) {
	
	unit->x=0.0;
	unit->y=0.0;
	unit->t=0.0;
	
	SETCALC(WeaklyNonlinear_next_k);
	
}

#define TWOPI 6.28318530717952646f

void WeaklyNonlinear_next_k(WeaklyNonlinear *unit, int inNumSamples) {
	
	float *out = ZOUT(0);
	
	float *in = ZIN(0);
	
	int reset= (int)ZIN0(1);
	float xrate= (float)ZIN0(2);
	float yrate= (float)ZIN0(3);
	float w0= (float)ZIN0(4);
	w0=w0*TWOPI/(SAMPLERATE); //convert frequency in Hertz to angular frequency
	w0=w0*w0; //constant needed for equation
			  //float eta= (float)ZIN0(5);
	
	float x,y,t; 
	
	x= unit->x;
	y=unit->y;
	t=unit->t;
	
	if(reset) {
		
		x= (float)ZIN0(5);
		y= (float)ZIN0(6);
		t=0;
	}
	
	//alpha*(x^xexponent*-beta)*y^yexponent 
	float alpha=(float)ZIN0(7);
	float xexponent=(float)ZIN0(8);
	float beta=(float)ZIN0(9);
	float yexponent=(float)ZIN0(10);
	
	//Runge Kutta? would require four cos calls
	for (int j=0; j<inNumSamples;++j) {
		
		//the naive Euler update
		float dxdt= xrate*y;
		
		float nonlinear= 0.0;
		
		if(alpha>0.000001 || alpha<(-0.000001)) {
			nonlinear= alpha * (pow(x,xexponent)+beta)*(pow(y,yexponent));
		}
		
		float dydt= yrate*(ZXP(in) - w0*x + nonlinear);
		
		//t=t+1;
		x+=dxdt;
		y+=dydt;
		
		//assumes fmod works correctly for negative values
		if ((x>1.0) || (x<-1.0)) x=fabs(fmod((x-1.0),4.0)-2.0)-1.0;
		//if ((posy>1.0) || (posy-1.0)) posy=fabs(fmod((posy-1.0),4.0)-2.0)-1.0;
		
		ZXP(out) = x;
		
		//printf("%f \n",(unit->posx/fabs(unit->posx))*0.5*(unit->posy*unit->posy+ unit->posx*unit->posx));
	}
	
	unit->x=x;
	unit->y=y;
	unit->t=t;
	
}







extern "C" void load(InterfaceTable *inTable) {
	
	ft = inTable;
	
	DefineSimpleUnit(SortBuf);
	DefineSimpleUnit(GravityGrid);
	DefineSimpleCantAliasUnit(Breakcore);
	DefineDtorUnit(Max);
	DefineSimpleUnit(PrintVal);
	DefineSimpleUnit(EnvDetect);
	DefineSimpleUnit(FitzHughNagumo);
	DefineSimpleUnit(DoubleWell);
	DefineSimpleCantAliasUnit(WeaklyNonlinear);
}


