
ChuckableBrowser {
	classvar	<chuckClasses, <>default;
	var	<classMenu, <subTypes, <subTypeMenu,
		<instanceList, <instanceListView,
		<keyCommandView, <>keyController;
	
	var	<stateByClass;	// holds last-touched subtype and instance name for each chuck class
	
	var	<layout, <masterLayout, <iMadeMasterLayout = false;

	*initClass {
		chuckClasses = AbstractChuckArray.allSubclasses
			.reject({ |item| item.name.asString[0..7] == "Abstract" })
			.sort({ |a, b| a.name < b.name });
	}

	*new { |masterLayout, bounds|
		^super.new.init(masterLayout, bounds)
	}
	
	init { |master, bounds|
		default.isNil.if({ default = this });
		masterLayout = master ?? {
			iMadeMasterLayout = true;
			FlowView.new
		};
			// note, if you give bounds that are too small, you will be unhappy
		layout = FlowView.new(masterLayout, bounds ?? { Rect(0, 0, 290, 310) });
			// since this should generally stay open, no garbage collection yet
//		layout.onClose = { this.free };
		this.initStates;
		this.makeViews;
		this.changeClass;
		instanceListView.focus;
	}
	
	initStates {
		stateByClass = IdentityDictionary.new;
		chuckClasses.do({ |class|
			stateByClass[class] = [nil, nil];
		});
	}
	
	changeClass { |class|
		var	thisClass, thisType;
		thisClass = class ?? { this.currentClass };
//		(thisClass != class).if({
			classMenu.value = chuckClasses.indexOf(thisClass);
//		});
//		thisType = subTypes.tryPerform(\at, subTypeMenu.value - 1);
		thisType = this.subTypeForClass(thisClass);
			// this use of IdentitySet is like "select distinct" in SQL
		subTypes = IdentitySet.new.addAll(thisClass.collection
			.select(_.respondsTo(\subType))	// the collect breaks if there is nil in the coll.
			.collect(_.subType))
			.asArray.sort;
		{ subTypeMenu.items_(["all"] ++ subTypes)
			.value_(1 + (subTypes.indexOf(thisType) ? -1));
		}.defer;
		this.getInstances(thisClass);
	}
	
	getInstances { |thisClass, currentSubType|
//		var	thisType;
		thisClass = thisClass ?? { this.currentClass };
		instanceList = thisClass.collection.keys;
		currentSubType = currentSubType ?? { this.currentSubtype };
		(currentSubType.notNil).if({
//			thisType = subTypes[subTypeMenu.value - 1];
			instanceList = instanceList.select({ |key|
				thisClass.new(key).respondsTo(\subType) and: {
					thisClass.new(key).subType == currentSubType
				}
			});
		});
		instanceList = instanceList.asArray.sort;
		{	instanceListView.items = instanceList.collect(_.asString);
			instanceListView.value = instanceList.indexOf(this.instanceForClass(thisClass)) ? 0;
		}.defer;
	}
	
	makeViews {
		SCDragSource(layout, 100@20)
			.string_("drag selection").align_(\center)
			.background_(Color.white).stringColor_(Color.new255(70, 130, 200))
			.beginDragAction_({ |drag|
				this.currentClass.new(instanceList[instanceListView.value])
			});
		SCDragSource(layout, 100@20)
			.string_("drag name").align_(\center)
			.background_(Color.white).stringColor_(Color.new255(70, 130, 200))
			.beginDragAction_({ |drag|
				instanceList[instanceListView.value].asSymbol
			});
		SCButton(layout, 70@20)
			.states_([["refresh", Color.new255(70, 130, 200), Color.new255(255, 218, 237)]])
			.action_({ this.changeClass });
		layout.startRow;
		SCButton(layout, 100@20)
			.states_([["insert selection", Color.new255(70, 130, 200),
				Color.new255(255, 218, 237)]])
			.action_({
				Document.current.selectedString_(
					this.currentClass.new(instanceList[instanceListView.value])
					.asCompileString
				)
			});
		SCButton(layout, 90@20)
			.states_([["insert name", Color.new255(70, 130, 200), Color.new255(255, 218, 237)]])
			.action_({
				Document.current
				.selectedString_(instanceList[instanceListView.value].asSymbol.asCompileString)
			});
		SCButton(layout, 80@20)
			.states_([["insert proto", Color.new255(70, 130, 200), Color.new255(255, 218, 237)]])
			.action_({
				this.currentClass.new(instanceList[instanceListView.value])
					.tryPerform(\proto);
			});
		layout.startRow;
		classMenu = SCPopUpMenu(layout, 160@20)
			.items_(chuckClasses.collect(_.name)).value_(0)
			.action_({ |menu|
				this.changeClass;
			});
		subTypeMenu = SCPopUpMenu(layout, 120@20)
			.action_({ |menu|
				this.setSubtypeForClass(this.currentClass, this.currentSubtype);
				this.getInstances(this.currentClass, this.currentSubtype)
			});
		layout.startRow;
		instanceListView = SCListView(layout, 284@220)
			.action_({ |list|
				this.setInstanceForClass(this.currentClass,
					instanceList[list.value])
			});
		keyCommandView = SCTextField(layout, 284@20);
	}
	
	focus {	// no flag, this always takes focus
		instanceListView.focus(true);
	}
	
	classMatch { |string|
		chuckClasses.do({ |class|
			(class.name.asString[0..string.size-1] == string).if({
				this.changeClass(class);
				^class
			});
		});
		^nil
	}
	
	subtypeMatch { |string|
		subTypes.do({ |type, i|
			(type.asString[0..string.size-1] == string).if({
				subTypeMenu.value = i + 1;
				this.setSubtypeForClass(this.currentClass, type);
				this.getInstances(nil, type);
				^type
			});
		});
		^nil
	}
	
	instanceMatch { |string|
		instanceList.do({ |inst, i|
			(inst.asString[0..string.size-1] == string).if({
				instanceListView.value = i;
				this.setInstanceForClass(this.currentClass, inst);
				^inst
			});
		})
		^nil
	}
	
	currentClass {
		^chuckClasses[classMenu.value]
	}
	
	currentSubtype {
		^subTypes[subTypeMenu.value - 1]
	}
	
	currentObject {
		^this.currentClass.new(instanceList[instanceListView.value])
	}
	
	currentObject_ { |obj|
		var index;
		this.changeClass(obj.class);
		(index = instanceList.detectIndex({ |inst| inst === obj.collIndex })
		).notNil.if({
			instanceListView.value = index;
			this.setInstanceForClass(obj.class, obj.collIndex)
		});
	}
	
	subTypeForClass { |class|
		^stateByClass[class][0]
	}
	
	instanceForClass { |class|
		^stateByClass[class][1]
	}

	setSubtypeForClass { |class, type|
		stateByClass[class][0] = type;
	}

	setInstanceForClass { |class, inst|
		stateByClass[class][1] = inst;
	}
}
