
// to do: VC(0) => VP(0); aVoicer => VC(0) -- should this update VP's value?
// would need to use dependants for this

// passive mode recorder - how to specify wrapper for BP?
// - how to start/stop? Use MBM trigger button?
// - on stop, need to unwrap and go

// openly stealing from ChucK for high-level control of a performance interface

	// this is the top-level abstract class
	// methods that should be inherited by all chuckables should be defined here
AbstractChuckArray {
	classvar	collection,	// each chuckable is a collection of related objects
			<lastIndex,	// indexed by name or symbol
			<>directory = "~/scwork/chucklib/",
			<>defaultSubType = \basic;	// set this when loading a piece to differentiate
									// in the browser
	
	var	<collIndex,	// the instance is stored at this.class(collIndex)
		<>value,		// what I'm pointing to
		<>subType;

		// collection is shared among all subclasses, so must index by classname
	*initClass {
		var traverseSubclasses;
		collection = IdentityDictionary.new;
			// recursively go thru subclasses to add a collection for each subclass to the coll
		(traverseSubclasses = { |class|
				// do not add a collection for abstract classes
			class.name.asString.contains("Abstract").not.if({
				collection.put(class.name, class.collectionType.new);
			});
			class.subclasses.do({ |subcl|
				traverseSubclasses.value(subcl);
			});
		}).value(this);
// startup.rtf should include
// (AbstractChuckArray.directory ++ "startup*.rtf").loadPaths;
	}
	
	*collectionType { ^Array }

	*collection { ^collection[this.name] }
	
		// to be removed later
	*debugCollection { ^collection }
	
		// new is misleading: I want to index the collection by CLASS(index)
		// the default method for AClass() is new, so...

		// this is for collections that cannot create new members
		// if index is an array, the output is an array. sweet!
	*new { |index|
		^index.notNil.if({ collection[this.name][index] });
	}

	*prNew { |index|
		var	temp;
		index.respondsTo(\at).not.if({
			this.put(index, temp = super.new.prInit(index));
		}, {
			Error("Index must not be a collection when creating a new object").throw;
		});
		^temp
	}
	
	*put { |index, member|
			// might need to extend the array so .put will succeed
		var	collTemp = collection[this.name];
		(index > (collTemp.size-1)).if({
			collection.put(this.name, collTemp.extend((index + 1).max(collTemp.size)));
		});
			// can't use collTemp here because it may be replaced with a new collection
		collection[this.name].put(index, member);
	}
	
	removeFromCollection {
			// do it this way b/c removeAt will change indices of other objects, not good
		this.class.collection[collIndex] = nil;
	}
	
		// to garbage collect everything for a piece
		// e.g. Fact.freeType(\techno1)
	*freeType { |type|
		type.notNil.if({
			this.collection.values.do({ |item|
				(item.subType == type).if({ item.free });
			});
		});
	}
	
	exists { ^value.notNil }
	
	init {}

	prInit { |index|
		collIndex = index;
		subType = defaultSubType;
		this.init(index);
	}
	
	v { ^value }	// shortcut to get value - keystroke efficiency is paramount!
	
	storeArgs { ^[collIndex] }  // compilestring will be ClassName.new('index')
	printOn { |stream|
		if (stream.atLimit, { ^this });
		stream << this.class.name << "(" <<< collIndex << ")";
	}

	
	*loadGui { (directory ++ "devEnvironment.rtf").loadPath }
	*openCodeDoc { |path|
		var doc;
		doc = Document.open(path);
		{ doc.bounds_(Library.at(\codeBounds)) }.defer(1);
	}
	*open {
		GetFileDialog.new({ |ok, path|
			ok.if({ this.openCodeDoc(path) })
		});
	}
	*runFile {
		GetFileDialog.new({ |ok, path|
			ok.if({ path.loadPath });
		});
	}

}

AbstractChuckNewArray : AbstractChuckArray {
		// this subclass is for collections that make a new member automatically
	*new { |index|
		var temp;
		^index.notNil.if({
				// if an object exists at index, give it back
			(collection[this.name][index] ??
					// else, make a new object and give it back
				{ this.prNew(index) })
		});
	}
}

AbstractChuckDict : AbstractChuckArray {
		// allow array indexing by passing an array of keys
		// BP([\mel, \chord, \drums]).play(4)
	*new { |index|
		var	collTemp;
		^index.notNil.if({
			collTemp = collection[this.name];
			index.respondsTo(\at).if({
				index.collect({ |i| collTemp[i] });
			}, {
				collTemp[index] ? collTemp[\default]
			});
		});
	}

	*put { |index, member|
			// should not call extend on a dictionary
		collection[this.name].put(index, member);
	}
	
	removeFromCollection {
		this.class.collection.removeAt(collIndex)
	}
	
	*collectionType { ^IdentityDictionary }
	
	*listKeys { |type|
		var collTemp = collection[this.name];
		collTemp.keys.asArray.reject({ |k| collTemp[k].isNil }).sort.do({ |k|
			(type.isNil or: { collTemp[k].tryPerform(\subType) == type }).if({
				k.postln
			});
		});
	}
}

AbstractChuckNewDict : AbstractChuckDict {
	*new { |index|
		var	collTemp;
		^index.notNil.if({
			collTemp = collection[this.name];
			index.respondsTo(\at).if({
				^index.collect({ |i| collTemp[i] });
			}, {
				^(collTemp[index] ?? { this.prNew(index) })
			});
		});
	}
}

/// actual collections

VC : AbstractChuckNewDict {
		// environment may be created by VoicerFactory (Fact)
	var	<env;

	init {
		env = Environment.new.know_(true);
	}
	
	use { |function|
		^env.use(function)
	}
	
	bindVoicer { |voicer|
		value = voicer;
	}
	
	bindFact { |fact, adverb, parms|
			// can be a voicer or an fx factory
//"Entered bindFact".debug;
		(fact.isVoicer.not).if({ "wrong type".warn },
				// default, assume voicer
			{	value.notNil.if({ this.free });  // free system resources before replacing
				env = fact.make(parms).know_(true);
				// make func must return voicer and place support objects (mixer, buffers, etc)
				// in the environment
//				value = env.use({ ~make.value });
				value = env[\value];
			});
//"Exited bindFact".debug;
	}
	
	free {
			// environment's free func must free all the support objects
		env.tryPerform(\use, {
			var mixer;
			(mixer = ~target.tryPerform(\asMixer)).notNil.if({
				mixer.mcgui.notNil.if({ mixer.mcgui.mixer = nil });
			});
			~free.value(value);	// maybe free func needs to know the voicer object
		});
		value.free;
		this.removeFromCollection;
//		value = env = nil;
	}
	
	bindBP { |bp|
		bp.voicer_(value)
	}
	
	draggedIntoVoicerGUI { |dest|
		value.tryPerform(\draggedIntoVoicerGUI, dest);
	}
	
	draggedIntoMixerGUI { |gui|
		var	mixer;
			// must have an environment, a voicer, and the target must be a mixer
		(env.notNil and: { value.notNil and:
			{ (mixer = env[\target].tryPerform(\asMixer)).notNil } }).if({
				gui.mixer = mixer;
			});
	}
	
	asVC { ^this }
}

// synth/target descriptor to use with synthNote event
// value is synthdef name
// env contains target and out keys
// populate using a factory
SY : VC {
	// inherits bindFact and free	
	bindBP { |bp|
		bp.v[\event] = bp.v[\event].copy.putAll((
			instrument: value,
			target: env[\target],
			out: env[\out]
		));
	}
}

	// factory for any kind of object that requires support objects
Fact : AbstractChuckNewDict {
	classvar	<>voicerTypes = #[\voicer],
			<>bpTypes = #[\bp];
	bindEnvironment { |env|
		value = env.know_(true);
	}
	
	bindEvent { |env|
		value = env.know_(true);
	}
	
	draggedIntoVoicerGUI { |dest|
		var	vp;
		(value.notNil and: { this.isVoicer }).if({
			this => VC(collIndex);
			(vp = VP.locateVPbyGui(dest)).notNil.if({
					// some ops are done in bind; need to do update thru a bind op
				VC(collIndex) => vp;
			}, {
				dest.model.voicer_(VC(collIndex).v);
			});
		});
	}
	
	draggedIntoMTGui { |gui, index|
		this.isBP.if({
			gui.model.add(this.makev, index);
			^true	// success flag
		}, {
			"Fact(%)'s type must be 'bp', but is %. Not instantiated."
				.format(collIndex.asCompileString, value[\type].asCompileString).warn;
			^false
		});
	}
	
	make { |argEnv ... args|
		var	env;
//"Entered Fact-make".debug;
		env = this.prepareEnv(argEnv);
		^env.make({
//"Running make function".debug;
			~value = ~make.value(*args);
		});
//"Exited Fact-make".debug;
//^env
	}
		// makes the value, but throws out the environment and returns only value
	makev { |argEnv ... args|
		var	env;
		env = this.prepareEnv(argEnv);
		^env.use({
			~make.value(*args);
		})
	}
	
	prepareEnv { |argEnv|
		var	env;
		env = value.copy;
		value[\keys].do({ |key|
			value[key].isNil.if({ env[key] = currentEnvironment[key] });
		});
		(argEnv.respondsTo(\keysValuesDo)).if({
			env.putAll(argEnv);
		});
		^env
	}
	
	proto {
		var	str, func;
		(value.notNil and: { value[\make].notNil }).if({
			str = CollStream(String(256));
			str << "Fact(" <<< collIndex << ").make(argEnv";
			((func = value[\make]).def.argNames > 1).if({
				str << ", ";
			});
			value[\make].streamArgs(str);
			str << ")";
			Document.current.selectedString_(str.collection);
		}, { "\nWARNING:\nFact(%) is empty.\n".postf(collIndex.asCompileString) });
	}
	
	isVoicer { ^voicerTypes.includes(value[\type]) }
	isBP { ^bpTypes.includes(value[\type]) }
}

VP : AbstractChuckArray {
	bindVoicerProxy { |proxy|
		value = proxy;
	}
	
	bindVC { |vc|
		this.bindVoicer(vc.value);
		vc.use({ ~onVPchuck.value(this) });	// allow other user defined ops
	}

		// Fact() => VP() is a shortcut for Fact() => VC() => VP()
	bindFact { |factory|
		factory.isVoicer.if({
			this.bindVC(factory => VC(factory.collIndex));
		});
	}
	
	bindVoicer { |voicer|
		value.tryPerform(\clearControlProxies);
		value.voicer_(voicer);
	}
	
	bindNil { value.voicer_(NullVoicer.new) }
	
	bindBP { |bp|
			// assign the process to play on this voicerproxy
			// or call useGui if we have it
		bp.v[\useGui].notNil.if({
			bp.v.useGui(collIndex)
		}, {
			bp.voicer_(value);
		});
	}
	
	bindGenericGlobalControl { |gc, adverb|
		var gcproxy;
		(adverb.asString[0].inclusivelyBetween($0, $9)).if({
			gcproxy = value.controlProxies[adverb.asInteger];
		}, {
			gcproxy = value.getFreeControlProxy(gc);
		});
		
		gcproxy !? { gcproxy.gc_(gc) };
	}
		
	draggedIntoVoicerGUI { |dest|
		value.tryPerform(\draggedIntoVoicerGUI, dest);
	}
	
	*doUseGuiOnBP { |bp, gui|
		var vpInColl;
			// useGui in a BP expects the index of the VP object
		(vpInColl = this.locateVPbyGui(gui)).notNil.if({
			bp.v.useGui(vpInColl.collIndex)
		});
	}
	
	*locateVPbyGui { |gui|
		^block { |break|
			this.collection.do({ |vp|
				(vp.v.editor === gui).if({ break.(vp) })
			});
			break.(nil)
		}
	}
	
	asVC {
		^VC.collection.values.detect({ |vc|
			vc.value === this.value.voicer
		});
	}

// this is a nice idea but I don't have the VC object to free
// that would result in memory leaks
//	freeByGui {
//		value.active.if({
//			^value.voicer.free
//		});
//	}
}

/// midi recording and buffer management

// stand-in for MIDIRecSocket
// does not contain or create the recsocket; MBM does that when it receives MRS as a chuck
// to use, this should be chucked into a MBM (MIDIBufManager)
// then use the MIDI control for the stop button, or click on the gui
// supply the name of the new buf as a symbol at instantiation

MRS {
	var <name, <>mbm;
	*new { |name, mbm|
		(mbm.notNil and: mbm.isKindOf(MBM).not).if({
				// allow index into collection
			(mbm = MBM(mbm)).isKindOf(MBM).not.if({
				Error("Invalid bufmgr. Must be an MBM or index thereto.").throw;
			});
		});
		^super.newCopyArgs(name,
			mbm ? MBM(0)
// if you don't create your MBM first, you lose control over clock, channel, and controller
				?? { MIDIBufManager(TempoClock.default, 0, nil) => MBM.prNew(0) }
		)
	}
//	init { |n|
//		name = n;
//	}
}

MBM : AbstractChuckArray {
	bindMIDIBufManager { |obj|
		value = obj;
	}
	
	bindMIDIRecBuf { |obj|
		value.add(obj);
	}
	
	bindMRS { |obj, index|
		value.recorder.notNil.if({
			"Already recording. Can't start a new recording.".warn;
		}, {
				// was index given as adverb? make sure buf goes in that slot
				// can also say MRS(\newbufname) =>.new MBM(0)
			(index.notNil and: (index != \new)).if({
				try { index = index.asInteger }
						// throw a more meaningful error than "asInteger not understood"
					{ Error("Invalid index supplied as adverb. Must be an integer.").throw };
				(index >= value.bufs.size).if({
					value.bufs.grow(index - value.bufs.size + 1);
				});
				value.value_(index);
			}, {
				value.value_(value.bufs.size);	// so that new buf will be created
			});
// don't like this... I should do function composition here
			value.postRecFunc_({ |buf|
				buf.name = obj.name;	// apply name specified in MRS(\name)
			});
			value.initRecord;
		});
	}
	
		// so I can say MBM(0)[1] => BP(\x)
	at { |index| ^value[index] }
	current { ^value.current }
}

// eventually I will want a passive-mode player that will be chuckable into MBM

PR : AbstractChuckNewDict {
	classvar	<defaultEnv;

//	var	// <adhoc,		// AdhocClass to store data and fill in specific functions
//		<>requiredKeys; // keys that must be non-nil before play or streaming

	*initClass {
			// sometime I might put default pseudo-methods into parent environment
		defaultEnv = Environment.new(8, nil,
				// parent environment - contains methods which all processes must implement
			Environment.make({
				~canStream = { |self|
					var out;
					out = true;
					~requiredKeys.do({ |key|
						key.envirGet.isNil.if({ out = false });
					});
						// if this is a wrapper process,
					~canWrap.if({
							// result depends on whether child exists and can stream
						out and: #{ ~child.tryPerform(\canStream, ~child) ? false }
					}, {
						out		// if not, result is whether the required keys are there
					});
				};
			});
		);
	}

	bindAdhocClass { |ad|
		value = ad;
			// merge the default environment parent entries into this environment's parent
		value.env.parent.isNil.if({
			value.env.parent_(Environment.new);
		});
		value.env.parent.putAll(defaultEnv.parent);
		value[\canWrap].isNil.if({ value.env.parent.put(\canWrap, false) });
	}
	
	bindEvent { |event|
		value.put(\event, event)
	}	

	bindProtoEvent { |proto|
		value.put(\event, proto.value)
	}

		// midi trigger support
	draggedIntoMTGui { |gui, index|
		gui.model.add(this => BP(collIndex), index);
		^true	// success flag
	}
}

BP : AbstractChuckNewDict {
	classvar	<>defaultQuant, <>defaultClock, <>defaultLatency = 0.5,
				// true = BP plays on voicerproxy (can switch vp's voicer during play)
				// false = BP maintains a hard link to a specific voicer (change by code only)
			<>useVoicerProxy = true;

	var	<>latency;	// process manages its own latency
					// latency is given in beats:
					// EVENT PROTOTYPES MUST CONVERT TO SECONDS

	*initClass {
		Class.initClassTree(Clock);
		defaultClock = TempoClock.default;
		defaultQuant = defaultClock.beatsPerBar;
	}
	
	init {
		latency = defaultLatency;
	}
	
	free {
		this.exists.if({
				// some processes need to do different cleanups if freed while playing
			value[\wasPlayingWhenFreed] = this.isPlaying;
			this.stopNow;
			value.freeCleanup;
			this.changed(\free);
			BP.removeDependant(value);
			value = nil;
		});
		this.removeFromCollection;
	}

	bindPR { |process, adverb, parms|
		this.bindAdhocClass(process.value.copy, adverb, parms);
	}
	
		// BPs can store state of wrapper processes, which can be reused by
		// BP(\wrapper) => BP(\child)
// should I clone it? Maybe I want to use the same wrapper on multiple processes
	bindBP { |process, adverb, parms|
			// if the Adhoc implements bindBP, use the Adhoc's method
			// this is for support of driver processes
		value[\bindBP].isFunction.if({
			value.bindBP(process, adverb, parms)
		}, {
				// otherwise wrap/replace
			this.bindAdhocClass(process.value, adverb, parms)
		});
	}
	
	bindAdhocClass { |process, adverb, parms|
			// midi input needs to be able to access this object from within the AdhocClass
		process.put(\collIndex, this.collIndex);
		this.exists.not.if({
			value = process;
		}, {
				// make sure outermost process is a dependant if the respondsToBass flag is set
				// first remove current outermost as dependant
			BP.removeDependant(value);
			case { adverb == \nest }
					// typically doReplay will be true, but MIDI input processes
					// want to stop the child process on receipt of the first midi msg
					// so they should not replay immediately
					{ this.wrap(process, process.doReplay ? true); }
				{ adverb.isNil or: { adverb == \wrap } }
					{ this.rewrap(process, process.doReplay ? true); }
				{ adverb == \replace }
					{ this.replace(process, process.doReplay ? true); }
				{ adverb == \overwrite }
					{ this.overwrite(process) }	// no doreplay for this guy
				{ adverb == \relate }	// for bp's that modify other bps' behavior
					{ value.put(\subordinate, process) }
				{ ("Unrecognized adverb " ++ adverb).warn; };
		});
		value[\isPlaying].isNil.if({
			value.put(\isPlaying, false).put(\isWaiting, false);
		});
		(value.notNil and: { value[\respondsToBass] == true }).if({
			BP.addDependant(value);
		});
			// allow some instance vars to be set at instantiation, before prep
		parms.respondsTo(\keysValuesDo).if({ value.putAll(parms) });
		this.promoteChildEventObjects;
		value.prep;	// process can do some initialization on instantiation
					// if unwrapping, child process should be aware that it's already inited
	}
	
	bindFact { |fact, adverb, parms|
		(fact.isBP).if({
			^fact.makev(parms, collIndex);
		}, {
			"%'s subtype is %; must be 'bp' to chuck into BP.".format(fact, fact.v[\type]).warn;
		});
	}
	
		// collapse multiple binds into one operation
		// order of binding is not guaranteed (limitation of Dictionaries)
		// event is (adverb: thing, adverb: thing)
		// translates to thing =>.adverb BP; thing =>.adverb BP
	bindEvent { |evt|
		evt.keysValuesDo({ |adverb, thing|
			thing.perform('=>', this, adverb);
		})
	}
	
		// 4 => BP(0) sets quant to BasicTimeSpec(4)
// handling of latency?
	bindNilTimeSpec { |spec|
		this.exists.if({ value.put(\quant, spec); });
	}
	
	bindSimpleNumber { |num, adverb|
		adverb.isNil.if({
			this.bindNilTimeSpec(num.asTimeSpec);
		}, {
			value.bindSimpleNumber(num, adverb);
		});
	}
	
	bindArray { |ar|
		this.bindNilTimeSpec(ar.asTimeSpec);
	}

		// 4 => BP sets default quant for all processes to BasicTimeSpec(4)
	*bindNilTimeSpec { |spec|
		defaultQuant = spec;
	}
	
	*bindSimpleNumber { |num|
		defaultQuant = num.asTimeSpec;
	}
	
	*bindArray { |ar|
		defaultQuant = ar.asTimeSpec;
	}
	
	bindSymbol { |symbol, adverb|
		this.exists.if({ 
			adverb.isNil.if({
				"Must supply an adverb when chucking a symbol into a BP. No action taken.".warn;
			}, {
				value.bindSymbol(symbol, adverb)
			});
		});
	}

		// adverb is the environment key into which the midi buf will go
		// MBM(0)[1] =>.melody BP(0)
	bindMIDIRecBuf { |buf, adverb, parms|
		this.exists.if({ value.acceptMIDIBuf(buf, adverb, parms); });
	}
	
	bindMIDIBufManager { |bufmgr, adverb, parms|
		this.bindMIDIRecBuf(bufmgr.current, adverb, parms)
	}

	bindMBM { |bufmgr, adverb, parms|
		this.bindMIDIBufManager(bufmgr.v, adverb, parms)
	}
	
		// this needs to wrap the current process in a MIDI input process
	bindMRS { |mrs, adverb|	// adverb is type of material
		this.exists.if({
			adverb = adverb ? value.defaultMIDIType;
				// process may need streams to be populated before midi input can commence
				// must make them before wrapping in the midi process
			this.prepareForPlay;
			this.bindPR(PR((adverb ++ "MIDI").asSymbol), \nest);  // PR(wrapper) => this
				// this chuck operation also makes the MIDIRecSocket
			mrs.mbm.bindMRS(mrs, nil);	// always create a new buf here (nil); mrs => mbm
			value.getMIDIParser(mrs.mbm.v.recorder);	// getMIDIParser should assign to ~parser
			value.eventStreamPlayer.isNil.if({
					// populate streams
				this.asEventStreamPlayer;
			});
			value.preparePlay;
// anything else?
		});
	}
	
	bindModalSpec { |mode, adverb|
		this.exists.if({ value.mode_(mode, adverb); });
	}
	
	bindMode { |mode, adverb|
		this.exists.if({ value.mode_(mode.collIndex, adverb); });
	}		
	
		// remove all adaptation sources and results (original material only will be left)
		// will also traverse children
	clearAdapt {
		var	child;
		this.exists.if({ 
			child = value;
			{ child.notNil }.while({
				child.clearAdapt;
				child = child.child;
			});
		});
	}
	
		// patterns for micro/macro rhythm keys
	bindPattern { |pat, adverb|
		this.exists.if({ value.bindPattern(pat, adverb); });
	}
	
	bindMacRh { |pat, adverb|
		this.exists.if({ value.bindPattern(pat, adverb) });
	}
	
	bindSA { |sa, adverb|
		this.exists.if({ value.bindSA(sa, adverb) });
	}
	
		// must not do this while playing
	bindTempoClock { |cl|
		this.isPlaying.not.if({		// also handles situation of empty bp
			value.put(\clock, cl);
		}, {
			Error("Cannot change the clock while playing.").throw;
		});
	}

	*bindTempoClock { |cl|
		defaultClock = cl;
	}
	
		// will try to change voicer while playing
		// if this is a wrapper, voicer will not change until the next event for outermost wrapper
	voicer_ { |vc|
		this.exists.if({ value.event.put(\voicer, vc); });
	}

		// play support
		// will do nothing if not playing and doReset is false
		// reversing argument order because quant is more important in performance
	play { |argQuant, argClock, doReset, notify = true|
		var	goTime;
		this.isPlaying.if({
			"already playing".warn;
		}, {
			this.exists.if({
				case { value[\asPattern].isNil }		// support for single-action BP's
					{		// must set clock variable so event can be retriggered later
						value[\clock] = (argClock ? value.clock ? defaultClock);
						value.put(\isPlaying, true);
						notify.if({ this.changed(\play); });
						value.clock.schedAbs(this.eventSchedTime(argQuant), {
							value.doAction;  // doAction is a pseudomethod in the AdhocClass
							value.put(\isPlaying, false);
							notify.if({ this.changed(\stop); });
						});
					}
				{ value.canStream }
					{
//"BP-play".postln;
						this.populateAdhocVariables(argClock);
						goTime = this.eventSchedTime(argQuant)/*.debug("BP-play-gotime")*/;
						value[\eventSchedTime] = goTime;
						this.prepareForPlay(argQuant, argClock, doReset);
						value.clock.schedAbs(goTime, {
								// nextBeat.isNil means that the stream is stopped
							(value[\isWaiting]
								and: { value[\eventStreamPlayer].tryPerform(\nextBeat).isNil })
							.if({
								value.clock.sched(0, value.eventStreamPlayer.refresh);
							});
							value[\isWaiting] = false;
							nil
						});
						value.put(\isPlaying, true).put(\isWaiting, true);
						notify.if({ this.changed(\play, goTime); });  // update MTGui
					};
			});
		});
	}
	prepareForPlay { |argQuant, argClock, doReset|
		var oldEventStreamPlayer;
		(this.exists and: { value.canStream }).if({
			this.populateAdhocVariables(argClock);
			(value.eventStreamPlayer.isNil or: { doReset ? false
				or: { value[\alwaysReset] ? false } })
				.if({
					(oldEventStreamPlayer = value.eventStreamPlayer).notNil.if({
						(argClock ? value.clock ? defaultClock)
							.schedAbs(this.eventSchedTime(argQuant), {
								oldEventStreamPlayer.stop;
							})
					});
					this.asEventStreamPlayer;
					value.eventStreamPlayer.refresh;  // needed for play to succeed
				});
		});
	}
	populateAdhocVariables { |argClock|
		value.put(\clock, argClock ? value.clock ? defaultClock);
		value.event.put(\clock, value.clock)
			.put(\latency, latency)
			.put(\child, value.child)
			.put(\propagateDownward, this.propagateDownFunc);
	}
		// dereference allows you to force play to start exactly now on the clock with `nil
	eventSchedTime { |argQuant|
//"BP-eventSchedTime ".post;
		this.exists.if({ 
			^(argQuant ? value.quant ? defaultQuant).dereference.asTimeSpec
				.applyLatency(this.latency ? 0)
				.schedTime(value.clock)
		}, { ^nil });
//.postln
	}
		// useful for wrapper processes--set up a midi trigger to fire ONE child process
	triggerOneEvent { |argQuant, argClock, doReset|
		(this.exists and: { value.canStream }).if({
			value.eventStreamPlayer.notNil.if({
				this.stop(argQuant);
			}, {
				this.prepareForPlay(argQuant, argClock, doReset);
			});
			value.clock.schedAbs(this.eventSchedTime(argQuant), {
				value.eventStream.next(value.event).play;
				nil	// otherwise it will play again after ~delta beats
			});
		});
	}
			
	stop { |argQuant|
// is this a valid test? midi wrapper does not have an eventStreamPlayer
// removing check b/c some processes might have a finite pattern but require stopCleanup later
		var	time = this.eventSchedTime(argQuant);
		this.exists.if({
			value[\eventSchedTime] = time;
			try {
				value.clock.schedAbs(time - 0.05, {
					value[\isPlaying].not.if({
						this.stopNow(nil, argQuant);
					});
				});
			} {
				this.stopNow(nil, argQuant);
			};
			value.put(\isPlaying, false).put(\isWaiting, false);
			this.changed(\stop);
		});
	}
	
		// for rewrapping/replacing -- specify an AdhocClass to use
		// there may be cases where I don't want to notify dependents
	stopNow { |adhoc, quant, notify = true, doCleanup = true|
		var	child;	// to iterate down the chain of child processes
//this.dumpBackTrace;
//"BP-stopNow".postln;
		this.exists.if({
			adhoc = adhoc ? value;
			adhoc.eventStreamPlayer.stop;
			doCleanup.if({ adhoc.stopCleanup(false, quant); });
			child = adhoc;
			{ (child = child.child).notNil }.while({
				child.eventStreamPlayer.stop;
					// false = not auto-stop (see CleanupStream in asStream for true condition)
				doCleanup.if({ child.stopCleanup(false, quant); });
				child.put(\isPlaying, false);
			});
			value.put(\isPlaying, false).put(\isWaiting, false);
			notify.if({ this.changed(\stop); });
		});
	}
	
	asStream {
//"BP-asStream".postln;
		var	stream;
		this.exists.if({ 
			stream = value.use({
					// by entering the environment for asStream, PR/BP code can be simpler
				this.asPattern.asStream;
			});
			^value.put(\eventStream, CleanupEventStream(stream,
				this.streamCleanupFunc(this, value))).eventStream
		}, { ^nil });
	}
	asPattern { ^value.asPattern }
	asEventStreamPlayer {
		var event;
		this.exists.if({
			value.preparePlay;
			value[\event] = event = this.prepareEvent;
			value.put(\eventStreamPlayer, 
				PausableEventStreamPlayer(this.asStream, event).refresh);
			^value[\eventStreamPlayer]
		}, { ^nil });
	}
	
	prepareEvent {
		(value[\event][\eventKey].notNil
			and: { ProtoEvent(value[\event][\eventKey]).exists }).if({
			^value[\event].copy.put(\parent, ProtoEvent(value[\event][\eventKey]).v);
		}, {
			^value[\event]
		});
	}

		// replay needs to change the cleanup func in the old stream before stopping
		// so I need to refer to other adhocs than my own in that case
	streamCleanupFunc { |self, adhoc|
		^{
//this.debug("streamCleanupFunc");
				// if a sequence stops of its own accord, eventStreamPlayer needs to be nil
				// so that stream will be recreated on next play
			adhoc.put(\eventStreamPlayer, nil);
				// optional post-stop activity--true means stopped automatically
			adhoc.stopCleanup(true);
			adhoc.put(\isPlaying, false);
			self.changed(\stop);	// if self is nil, this is still OK
		}
	}
		
		// re-pattern, and restart stream if playing
		// should I follow the naming convention of stopNow / stop?
		// would break code
	reset {
		var	oldPlayer;
		this.exists.if({
			(oldPlayer = value[\eventStreamPlayer]).notNil.if({
				value.reset;
				this.prepareForPlay(doReset:true);
			});
			this.isPlaying.if({
				value.clock.schedAbs(oldPlayer.nextBeat, value[\eventStreamPlayer].refresh);
					// must kill the old cleanup func
				oldPlayer.stream.tryPerform(\cleanup_, nil);				oldPlayer.stop;
			});
			this.changed(\reset);
		});
	}
	
	resetq { |argQuant|
		this.exists.if({
			value[\clock].notNil.if({
				value[\clock].schedAbs(this.eventSchedTime(argQuant) - 0.05, {
					this.reset;
				});
			}, { this.reset; });	// no clock, must do it now
		});
	}
	
		// tryPerform because otherwise Object-isPlaying is invoked (always false)
	isPlaying { ^value.tryPerform(\isPlaying) ? false }
	isDriven { ^value.tryPerform(\isDriven) ? false }

	wrap { |process, doReplay = true|
		var	saveAdhoc;
		process.canWrap.if({
			process = process.copy;
				// maybe not necessary to put in the adhoc?
			process.put(\child, value)
				.put(\clock, value[\clock]);
			saveAdhoc = value;
//			saveEventStreamPlayer = value[\eventStreamPlayer];
			value = process;
			this.recalcPropagateKeys;
			doReplay.if({ this.replay(saveAdhoc[\eventStreamPlayer], saveAdhoc) });
		}, {
			"This is not a wrapper process. Wrapping not done.".warn;
		});
	}

	rewrap { |process, doReplay = true|
		var	saveAdhoc;
		process.canWrap.if({
				// if there is no child, must delegate to -wrap
			value[\child].isNil.if({ ^this.wrap(process, doReplay) });
			process = process.copy;
				// maybe not necessary to put in the adhoc?
			process.put(\child, value[\child])
				.put(\clock, value[\clock]);
			process[\event].put(\child, value[\child])
				.put(\clock, value[\clock]);
				// transfer other required keys
			process.rewrapKeys.do({ |key|
				process[key].isNil.if({
					process.put(key, value[key])
				});
			});
			saveAdhoc = value;
			BP(\saved).value = nil;
				// clear child so saved process can't be replayed
				// and save; typical chucking use precludes saving, so I do it automatically
				// user should do BP(\saved) => BP(\myWrapper) after rewrap
				// this should also be used for freeing resources
			saveAdhoc/*.put(\child, nil)*/ =>.overwrite BP(\saved);
			value = process;
			this.recalcPropagateKeys;
			doReplay.if({ this.replay(saveAdhoc[\eventStreamPlayer], saveAdhoc) });
		}, {
			"This is not a wrapper process. Wrapping not done.".warn;
		});
	}
	
		// pop one wrapper process off the stack (lifo)
		// can use BP(\x).unwrap => BP(\y) to save state of wrapper, or to call freeCleanup
	unwrap { |doReplay = true|
			// extract the child process
		var	oldWrapper;
		value.canWrap.if({
			oldWrapper = value;
			value = value.child;
			this.recalcPropagateKeys;
			doReplay.if({ this.replay(oldWrapper[\eventStreamPlayer], oldWrapper) });
		}, {
			"Can't unwrap a child process.".warn
		});
		^oldWrapper	// returns nil on failure
	}
	
// What the hell was I thinking with this one (replace)? I can't even remember.
// maybe replace the child with the child from another process? But this code isn't right

// throw everything out and start with this incoming process
// this method is not fully supported yet
	overwrite { |process|
		var	saveAdhoc, esp;
		saveAdhoc = value;		// to preserve this adhocclass through the scheduling
// this test is bad
		((esp = value[\eventStreamPlayer]).notNil
			and: { esp.isPlaying and: { esp.nextBeat.notNil } }).if({
				// let it play through this event
			value[\clock].schedAbs(esp.nextBeat - 0.05, {
				esp.stream.tryPerform(\cleanup_, this.streamCleanupFunc(nil, saveAdhoc));
				this.stopNow(saveAdhoc)
			});
		});
		value = process;
	}
	
	replay { |oldEventStreamPlayer, oldAdhoc|	// process must be playing
		var	nextTime;
//"enter replay".debug;
		(oldEventStreamPlayer.isPlaying and: { oldEventStreamPlayer.nextBeat.notNil }).if({
			nextTime = oldEventStreamPlayer.nextBeat;
//nextTime.debug("old player was playing; nextTime");
//"stopping old player".debug;
				// to prevent the stream's cleanup from messing up my flags
			oldEventStreamPlayer.stream.tryPerform(\cleanup_,
				this.streamCleanupFunc(nil, oldAdhoc));
			oldEventStreamPlayer.stop;
			value[\isPlaying] = true;
				// make a new one and schedule it for the next event time
//"scheduling new player; inspecting".debug;
			value[\clock].schedAbs(nextTime, this.asEventStreamPlayer.refresh);
		});
	}
	
		// to be stored in play event
		// assumes we're in the environment already
	propagateDownFunc {
		^#{	~child.notNil.if({
				~child.event[\allKeysToPropagate].do({ |key|
					~child.event.put(key, key.envirGet);
				});
				~child.event.put(\child, ~child.child)
					.put(\clock, ~clock)
						// i.e., this func!
					.put(\propagateDownward, ~propagateDownward);
			});
		}
	}
	
		// this should be called on wrapping/unwrapping
		// I don't want to do this per note!
	recalcPropagateKeys {
		var	keys, event, child;
		keys = IdentitySet.new;
			// first descend into all child events to get a distinct set of keysToPropagate
			// from each
		child = value;	// start with outermost
		{ child.notNil }.while({
			(event = child[\event]).notNil.if({
				keys = keys.union(event[\keysToPropagate])
					.union(ProtoEvent(event[\eventKey]).v[\keysToPropagate]);
			});
			child = child[\child];
		});
			// now populate all child events with allKeysToPropagate
		child = value;
		{ child.notNil }.while({
			child.event.tryPerform(\put, \allKeysToPropagate, keys);
			child = child.child;
		});
	}
	
	promoteChildEventObjects {
		var	childEvent, event;
		(value[\child].notNil and: { (childEvent = value[\child][\event]).notNil
			and: { (event = value[\event]).notNil } }).if({
			childEvent[\allKeysToPropagate].do({ |k|
				(childEvent[k].notNil and: { event[k].isNil }).if({
					event[k] = childEvent[k];
				});
			});
		});
	}
	
		// GUI support
	
		// midi trigger support
	draggedIntoMTGui { |gui, index|
		gui.model.add(this, index);
		^true	// success flag
	}
	
	draggedIntoVoicerGUI { |gui|
		var	voicer;
		(voicer = gui.model).active.if({
			useVoicerProxy.if({
				this.voicer = gui.model;
			}, {
				this.voicer = gui.model.voicer;
			});
			Post << "Changed BP(\\" << collIndex << ")'s voicer to " << voicer.asString << "\n";
		}, {
				// otherwise, if this process has a useGui method, use it
				// delegate because this is more complex than it sounds
			value[\useGui].notNil.if({ VP.doUseGuiOnBP(this, gui) });
		});
	}
	
	draggedIntoMixerGUI { |gui|
		value[\chan].notNil.if({
			gui.mixer_(value[\chan])
		}, {
			"No mixerchannel defined in this process. Cannot assign to MCGui.".warn;
		});
	}
			
}

// microrhythms
// contains a pattern that returns a 4-item array: [delta, length, gate, args]
// "args" is optional
// maybe somePattern.collect({ |val| [val, val*2, someOtherVal] })
// or Ptuple([deltaPat, lengthPat, gatePat], repeats)
// if finite, this determines the length of a chord gesture triggered by macrorhythm player
// chord prototype has a microrhythm pattern that outputs symbolic keys into this collection

	// should this allow new? no new, allows default
MicRh : AbstractChuckNewDict {
	bindPattern { |pattern|
			// first, test pattern to see that it's valid
//		(pattern.asStream.next.size < 3).if({
//			"Invalid microrhythm pattern. Output must be an array of at least 3 items.
//Microrhythm not added.".warn;
//		}, {
			value = pattern;
//		});
	}
	
		// #{ populate some variables, then return a pattern based on the vars } => MicRh(...)
		// also, #{ |notePattern| Pfin(notePattern.estimateLength, notePattern) }
		// to allow gestures to retain their integrity across note-pattern changes
	bindFunction { |func|
		value = func
	}
	
		// will be called at play time
		// if value is already a Pattern, .value will return the pattern unmodified
		// a function will be executed -- notePattern is for pattern length estimation
	asPattern { |... args|
//args.debug('micrh aspattern args');
//this.dumpBackTrace;
		^value.value(*args)
	}

	bindClassName { ^\MacRh }
}

// macrorhythms
// used in wrapper processes
// should output either a simplenumber (delta) or an array [delta, length]
// if length is not nil, it is the maximum time a child process will be allowed to play

// process should pattern as Pbind(#[\delta, \length], pattern)

MacRh : MicRh {
	bindPattern { |pattern|
		var temp;
		((temp = pattern.asStream.next).size < 1 and: temp.isNumber.not).if({
			"Invalid macrorhythm pattern. Output must be an array of at least 1 item,
or a SimpleNumber. Macrorhythm not added.".warn;
		}, {
			value = pattern;
		});
	}
	
	// also inherits bindFunction
}

// proto-event holder
ProtoEvent : AbstractChuckNewDict {
	bindEvent { |event|
		event[\play].isNil.if({
			"Event is missing a play function. Not added.".warn;
		}, {
				// consolidate keys from event and its parent into one parent event
			value = Event(parent: ());
			event.parent.notNil.if({ value.parent.putAll(event.parent) });
			value.parent.putAll(event);
		});
	}
}

ArpegPat : MacRh {
	bindPattern { |pattern|
		value = pattern
	}
	// inherits bindFunction and asPattern(notes)
}

SA : AbstractChuckNewDict {
	classvar	<>defaultEvent;

	var	<argKeys;

	*initClass { defaultEvent = () }

//	bindPattern { |pattern|
//		value = pattern;
//		this.setArgKeys;
//	}

		// array is of the form [pattern, keys]
	bindArray { |array|
		#value, argKeys = array;
//		this.setArgKeys;
	}
	
	// bindFunction?
	
//	setArgKeys {
//			// poll the stream once to fetch the keys
//		argKeys = value.asStream.next(()).keys.asArray
//	}
	
	asPattern { ^value }
}

// dynamically assignable mixer gui slots

MCG : AbstractChuckArray {
	bindMixerChannelGUI { |gui| value = gui }
	
	bindMixerChannel { |channel|
		value.mixer = channel;
	}
	
	bindVC { |vc|
		{ value.mixer_(vc.env.target) }.try({
			"VC's target is not a MixerChannel. Can't bind into MCG.".postln;
		});
	}
	
	bindBP { |bp|
		(bp.exists and: { bp.v[\chan].notNil }).if({
			try { value.mixer_(bp.v[\chan]) }
				{ "Error during MCG-bindBP.".postln }
		});
	}
	
	bindVP { |vp|
		var	mix = vp.v.tryPerform(\voicer).tryPerform(\bus).tryPerform(\asMixer);
		mix.notNil.if({
			{ value.mixer_(mix) }.try({
				"VC's target is not a MixerChannel. Can't bind into MCG.".postln;
			});
		}, { "VC's target is not a MixerChannel. Can't bind into MCG.".postln; });
	}		
}

// overall modalspec for piece
// make sure Mode(\default) is populated!
Mode : AbstractChuckNewDict {
	bindModalSpec { |mode|
		value = mode
	}
	
	bindArray { |modeNames|
		value = modeNames;  // .collect({ |name| Mode(name).v });
	}
	
	bindSymbol { |modeName|
		value = Mode(modeName).copy;
	}
	
// for compatibility with ModalSpec
//	prMap { |degree|
//		(value.size == 0).if({
//			^value.prMap(degree)
//		}, {
//			^value[0].prMap(degree)
//		});
//	}
//	
//	prUnmap { |key|
//		(value.size == 0).if({
//			^value.prUnmap(key)
//		}, {
//			^value[0].prUnmap(key)
//		});
//	}
	
	asMode { ^this }
	
		// delegation to referenced object
	doesNotUnderstand { |selector ... args|
		var mode;
		mode = (value.size == 0).if({ value }, { value[0].asMode });
		^mode.perform(selector, *args)	// if this throws an error, then it should
	}
}

// holder for adaptation functions -- may be material manipulators or support funcs
Func : AbstractChuckNewDict {
	var	<>nilProtect = true;
	bindFunction { |func, adverb|
		value = func;
		(adverb == \allowNil).if({ nilProtect = false });
	}
	
		// args will usually include source material and material to crossbreed with it
	doAction { |... args|
		nilProtect.if({
			^value.value(*args) ? args[0]  // this may be a bad idea, but I'll go with it
									// if somebody calls a Func that doesn't exist,
									// result should be the original...???
		}, { ^value.value(*args) });
	}
	
	listArgs {
		this.streamArgs(Post);
//		Post << "Func(" << collIndex.asCompileString << ").doAction(";
//		value.def.argNames.do({ |name, i|
//			(i > 0).if({ Post << ", " });
//			Post << name;
//		});
//		Post << ");\n";
	}
	
	proto {
		var	stream;
		stream = CollStream(String.new(256));
		this.streamArgs(stream);
		Document.current.selectedString_("\n" ++ stream.collection)
	}
	
	streamArgs { |collstream|
		collstream << "Func(" <<< collIndex << ").doAction(";
		value.streamArgs(collstream);	// add function args
		collstream << ");\n";
	}
}

// chucking MIDI controls

// start/stop processes using MIDI keys

// Midi Trigger -- corresponds to a socket in the MIDI hierarchy
// contains a dictionary of note nums -> processes
MT : AbstractChuckNewDict {
	classvar	<>default;	// used in ChuckBrowserKeyController
	classvar	<>readyThreshold = 5;	// how long to hold a process in ready state before clearing
	classvar	<>defaultMinNote = 48, <>defaultMaxNote = 72;  // integer note numbers
	var	/*<>nextAddNote,*/ <lastBP, noteAllocator;
	
	var	<socket, <>minNote, <>maxNote;	// midi responder, passes messages here

		// all indices have to be converted to channel objects
	*prNew { |index|
		var	temp;
		this.put(index = index.asChannelIndex, temp = super.prNew(index));
		^temp
	}

	*new { |index|
		var	collTemp;
		^collection[this.name][index = index.asChannelIndex] ?? { this.prNew(index) }
	}
	
		// needed b/c MIDIChannelIndices can be == but not ===
	*collectionType { ^Dictionary }

	init {
		value = IdentityDictionary.new; // note num -> MTNoteInfo
		minNote = /*nextAddNote =*/ defaultMinNote;
		maxNote = defaultMaxNote;
		socket = MTSocket(collIndex, this);
		noteAllocator = ContiguousBlockAllocator(maxNote+1, minNote);
		default = this;
	}
	
	free {
		this.changed(\free);
		socket.free;
		value.do(_.free);
		this.class.collection.removeAt(collIndex);
		collIndex = value = socket = /*nextAddNote =*/ noteAllocator = minNote = maxNote = nil;
		this.releaseDependants;	// remove from dependants dictionary
		this.removeFromCollection;
	}

	bindPR { |pr, adverb|
		BP(pr.collIndex).free;
		this.add(pr => BP(pr.collIndex), adverb);
	}

	bindBP { |bp, adverb|
		this.add(bp, adverb);
	}
	
	bindFact { |fact, adverb|
		(fact.isBP).if({
			this.add(fact.makev, adverb)
		}, {
			"%'s subtype is %; must be 'bp' to chuck into MT.".format(fact, fact.v[\type]).warn;
		});
	}
	
		// support for BP([\pr1, \pr2, \pr3, \pr4]) => MT(0) syntax
		// you give up control over where they go
	bindArray { |ar|
		ar.do({ |bp| bp => this });
	}
	
	add { |bp, adverb, updateGUI = true|
		var new, nextAddNote;
		lastBP = bp;
		nextAddNote = this.convertAdverb(adverb);
		this.adverbIsValidNote(adverb).if({
			noteAllocator.reserve(adverb);
		});
		value[nextAddNote].notNil.if({ value[nextAddNote].free });
		value.put(nextAddNote, new = MTNoteInfo(bp, false, nextAddNote, this));
//		nextAddNote = (nextAddNote + 1).wrap(defaultMinNote, defaultMaxNote);
		updateGUI.if({ this.changed(new) });	// tell the gui
	}
	
	removeAt { |notenum|
		noteAllocator.free(notenum);
		this.changed(notenum);
		^value.removeAt(notenum).free;
	}

	adverbIsValidNote { |adverb|
		^(adverb = adverb.tryPerform(\asInteger)).notNil
			and: { adverb.inclusivelyBetween(minNote, maxNote) }
	}

	convertAdverb { |adverb|
		(this.adverbIsValidNote(adverb)).if({
			^adverb.asInteger
		}, {
			^noteAllocator.alloc(1)
		});
	}
	
	noteOn { |num|
		var	entry, clock;
			// entry must exist
		(entry = value[num]).notNil.if({
				// if ready to fire, do play
			entry.ready.if({
				entry.ready = false;
				entry.bp.isPlaying.if({
					entry.bp.stop;
				}, {
					entry.bp.play;
				});
				this.changed(entry);
			}, {		// else make ready and schedule non-ready
				entry.ready = true;
				AppClock.sched(readyThreshold, {
					entry.ready = false;
					this.changed(entry);
				});
				this.changed(entry);
			});
		});
	}
	
	guiClass { ^MTGui }
}

MTNoteInfo {
	var	<>bp, <>ready, <>noteNum, <owner;
	
	*new { |bp, ready, noteNum, owner|
		var new;
		new = super.newCopyArgs(bp, ready, noteNum, owner);
		bp.addDependant(new);
		^new
	}
	
	asString { ^(noteNum.asMIDINote ++ ": " ++ bp.collIndex) }
	
	playState {
		^case { ready == true } { \ready } // ready takes precedence
			{ bp.isDriven } { \driven }
			{ bp.isPlaying } { \playing }
			{ \idle }
	}
	
	free { bp.removeDependant(this) }
	
	update { |obj, changer|
		(changer == \free).if({
			owner.removeAt(noteNum);
		});
		#[\play, \stop, \driven].includes(changer).if({  // ignore other messages
			owner.changed(this);	// this should call the gui
		});
	}
}
