
// these are near-identical copies of some of the stream subclasses
// people keep making changes to streams that break compatibility,
// so this is to make sure my stuff keeps working

PauseStreamHJH : Stream
{
	var <stream, <originalStream, <clock, <nextBeat, <streamHasEnded=false;
	var <isWaiting = false;
	
	*new { arg argStream, clock; 
		^super.newCopyArgs(nil, argStream, clock ? TempoClock.default) 
	}
	
	isPlaying { ^stream.notNil }
	
	play { arg argClock, doReset = (false), quant=0.0;
		if (stream.notNil, { "already playing".postln; ^this });
		if (doReset, { this.reset });
		clock = argClock ? clock ? TempoClock.default;
		streamHasEnded = false;
		stream = originalStream; 
		isWaiting = true;	// make sure that accidental play/stop/play sequences
						// don't cause memory leaks
		clock.play({
			if(isWaiting and: { nextBeat.isNil }) {
				clock.sched(0, this);
				isWaiting = false;
			};
			nil
		}, quant);
		^this
	}
	reset { ^originalStream.reset }
	stop {
		stream = nil;
		isWaiting = false;
	}
	removedFromScheduler { this.stop }
	wasStopped { ^streamHasEnded.not and: { stream.isNil } }
	
	pause {
		stream = nil;
		isWaiting = false;
	}
	resume { arg argClock, quant=1.0; 
		^this.play(clock ? argClock, false, quant) 
	}
	
	refresh {
		stream = originalStream
	}

	start { ^this.play(clock, true) }
		
	stream_ { arg argStream; 
		originalStream = argStream; 
		if (stream.notNil, { stream = argStream; streamHasEnded = argStream.isNil; });
	}

	next { arg inval; 
		var nextTime = stream.next(inval);
		if (nextTime.isNil) { streamHasEnded = stream.notNil; stream = nextBeat = nil }
			{ nextBeat = inval + nextTime };	// inval is current logical beat
		^nextTime
	}
	awake { arg beats, seconds, inClock;
		stream.beats = beats;
		^this.next(beats)
	}
}

CleanupEventStream : Stream {
	var <stream, <>cleanup;
	
	*new { arg stream, cleanup;
		^super.newCopyArgs(stream, cleanup)
	}
	next { arg inval;
		var outval;
		inval.notNil.if({
			outval = stream.next(inval);
		});
		if (outval.isNil) {
			try {
				cleanup.value(this, inval);
				cleanup = nil;
			} { |error|
				error.reportError;
				"CleanupStream cleanup failed. Continuing to stop stream.".warn;
			}
		}
		^outval
	}
	reset {
		stream.reset
	}
}

PausableEventStreamPlayer : PauseStreamHJH {
	var <>event, <>muteCount = 0;
	
	*new { arg stream, event;
		^super.new(stream).event_(event ? Event.default);
	}

		// note, I took out stream.next(nil) from stop -- this breaks Pmono & Pfx
	stop { stream = nil; isWaiting = false; }

	mute { muteCount = muteCount + 1; }
	unmute { muteCount = muteCount - 1; }
	
	next { arg inTime;
		var nextTime;
		var outEvent = stream.next(event);
		if (outEvent.isNil) {
			streamHasEnded = stream.notNil;
			stream = nextBeat = nil;
			^nil
		}{
			if (muteCount > 0) { outEvent.put(\freq, \rest) };
			if ((nextTime = outEvent.play).isNil) { stream = nil };
			nextBeat = inTime + nextTime;	// inval is current logical beat
			^nextTime
		};
	}
	
	asEventStreamPlayer { ^this }
}
