
TempoClick {
	// takes a tempoclock, plays impulses on a kr bus based on the clock
	
	// this latency adjustment works on my iBook 700mHz
	// if your routines don't play in sync with BufferSeq's,
	// adjust this number. Smaller values mean the BufferSeq's
	// will play slightly later.
	// not yet determined if the correct latencyFudge depends on desired tempo
	classvar	<>latencyFudge = 0,
			nodeIDStart = 500, lastIDUsed;	// for bouncing between 2 reserved node IDs
	
	var	<server, <clock, <bus, <subdiv, <nodeID, nodeIDBounce;
	var	aliveThread;
	
	*initClass {
		StartUp.add({
			SynthDef.writeOnce("TempoClick", { arg tempo = 1, subd = 1, bus = 0;
					// kill the synth every beat - this helps the client enforce sync
				Line.kr(0, 1, tempo.reciprocal, doneAction:2);
				Out.kr(bus, Impulse.kr(tempo * subd));
			});
		});
	}
	
	*new { arg server, clock, bus, subdiv = 1;
		^super.newCopyArgs(server, clock ? TempoClock.default,
			bus, subdiv).init;
	}
	
	init {
		clock.isKindOf(TempoClock).not.if({
			("Clock must be a TempoClock. This is " ++ clock.asString).die;
		});
		clock.addDependant(this);
		nodeID = nodeID.isNil.if({ nodeIDStart }, { lastIDUsed = lastIDUsed + 2 });
		nodeIDBounce = nodeID * 2 + 1;
		this.play;
	}
	
	play {
		server.waitForBoot({	// when server is booted
			bus.isNil.if({ bus = Bus.control(server, 1) });
//			synth.notNil.if({ synth.free });
			this.startAliveThread;
		});
	}
	
	startAliveThread {
		aliveThread = Routine({
			{	nodeID = nodeIDBounce - nodeID;
					// schedule the onset using latency to coincide with the clock
					// so server-side and client-side sequencers are in sync
				server.sendBundle(0.2, [\s_new, \TempoClick, nodeID,
					0, server.defaultGroup.nodeID, 
					\tempo, clock.tempo, \subd, subdiv, \bus, bus.index]);
				1.0.wait;
			}.loop;
		});
		clock.schedAbs(clock.elapsedBeats.roundUp - (0.2*clock.tempo*latencyFudge),
			aliveThread);
	}
	
	stop {
		aliveThread.stop; aliveThread = nil;
	}
	
	remove { this.free }
	
	free {
		this.stop;
		bus.free;
		clock.removeDependant(this);
	}
	
//	resync {
//			// tempo changes can cause the impulses to go out of sync with the client-side clock
//			// this repairs the problem
//			// if subdivision > 1, some beats will be missed
//		this.stop;
//		this.play;
//	}
	
	update {
		aliveThread.notNil.if({ server.sendMsg(\n_set, nodeID, \tempo, clock.tempo) });
	}
	
	tempo_ { arg tempo;
		clock.tempo = tempo;	// TempoClock calls changed, which calls update
	}
	
	subdiv_ { arg s;
		subdiv = s;
		aliveThread.notNil.if({ server.sendMsg(\n_set, nodeID, \subd, subdiv) });
	}
	
	index { ^bus.index }
	
}
