
	// these classes allow mixing and matching of components
	// of recorded or programmed note sequences
	// SeqBuilder exists primarily as the model for the gui--the gui is the meat of it

SeqBuilder {
		// holds slots for pitch seq, delta seq, length seq and arg seq
		// when dragged into a VoicerProxyGUI, these components are assembled
		// into a sequence
	
		// dragging several items sequentially into one of the slots concatenates at the end
		// clicking "fix" takes this concatenated sequence and enters it into the pattern
		// as a NoteSequence
		// dragging SeqBuilder into VoicerGUI compiles the NoteSequence pattern
		// into a LinkedVoicerSequencer
		// clicking "clear" empties master pattern to start again
		// (slots remain intact)
		
		// todo: how to change seq type (Pseq, Pxrand, P......) ??

	var	<pitch, <dur, <length, <absLength, <maxPlayBeats, <args,
		<>seq,	// pattern returning NoteSequences
		<>patProto = "Pseq(x, inf)",	// see fix {}
		<>view;	// so changes here can be reflected graphically
	
	*new { arg pitch, delta, length, absLength = true, args;
		^super.new.all_(pitch, delta, length, absLength, args)
			.seq_(Array.new)	// initialize seq
	}
	
//	init { arg ... in;
//		#pitch, delta, length, absLength, args = in;
//	}

	fix {	// makes a notesequence out of current contents
		var sq;
		(sq = this.asNoteSequence).notNil.if({
			seq = seq.add(sq);
		}, {
			"SeqBuilder-fix: No sequence to add.".warn;
		});
		view.notNil.if({ view.refresh });
	}
	
	clear {
		seq = Array.new;
		view.notNil.if({ view.refresh });		
	}
	
		// these handle dragging the new sequence somewhere else
		// here, into a player
	draggedIntoVoicerGUI { arg dest;
		var	v, seqPat, w, screenBounds;		// w is for name pop-up
		v = dest.model;
		seqPat = this.asSeqPat;
		(seqPat.list.size > 0).if({
			screenBounds = SCWindow.screenBounds;
			w = SCWindow("Name this sequence",
				Rect(screenBounds.width/2 - 150, screenBounds.height/2 - 40, 300, 90));
			SCStaticText(w, Rect(0, 30, 300, 20))
				.string_("Enter a name for this sequence:")
				.align_(\center);
			SCTextField(w, Rect(40, 60, 220, 20))
				.action_({ arg txt;
					var sq;
// Do I always want to stop others?
					sq = [txt.string, { arg p, g; g.stopOthers(p); p.reset.play },
						LinkedVoicerSequencer(v, seqPat)];
					(v.processes.size == 0).if({	// must make a new process group
						v.addProcess([
							sq,
							["Stop", { arg p, g; g.stopAll }]
						]);
							// auto-add to VPTrigSocket? what's current TrigSocket?
					}, {
							// how to determine current group? for now just use vpg 0
						v.processes.at(0).add(sq);
							// check that stop is there?
					});
					w.close;
				})
				.focus(true);
			w.front;
		});
	}
	
		// here, into the midi buffer repository so that the composite sequence
		// can be used as a source for other sequences
		// not sure if this is possible
//	draggedIntoMIDIRecGUI { arg dest;
//	}
//	
	all_ { arg ... in;		// how to concatenate here?
		#[\pitch_, \dur_, \length_, \absLength_, \maxPlayBeats_, \args_].do({ |meth, i|
			this.perform(meth, in[i]);
		});
//		#pitch, dur, length, absLength, maxPlayBeats = in;
//		this.args_(in[5]);
		absLength = absLength ? true;
		view.notNil.if({ view.refresh });
	}		
	
// NEED TO CHANGE THESE METHOD NAMES	

	pitch_ { arg p;
		p.isNil.if({
			pitch = nil;
		}, {
			pitch = pitch ++ p;
		});
		view.notNil.if({ view.refresh });
	}
	
	dur_ { arg d;
		d.isNil.if({
			dur = nil;
		}, {
			dur = dur ++ d;
		});
		view.notNil.if({ view.refresh });
	}
	
	length_ { arg l;
		l.isNil.if({
			length = nil;
		}, {
			length = length ++ l;
		});
		view.notNil.if({ view.refresh });
	}
	
	absLength_ { arg absL;
		absLength = absL;
		view.notNil.if({ view.refresh });
	}
	
	maxPlayBeats_ { arg mpb;
		maxPlayBeats = mpb;
		view.notNil.if({ view.refresh });
	}
	
	args_ { arg a;	// how to concatenate?
		var i;
		a.isNil.if({
			args = nil;
		}, {
			a.isKindOf(Dictionary).if({		// autoconvert dictionary to array of pairs
				a = a.asSortedArray.flatten(1);
			});
a.insp;
				// make args[0] == Array of keys, args[1] = array of patterns
			(args.isNil || (args.size == 0)).if({
				args = a;
			}, {
				args = args.clump(2).flop;
				a.clump(2).do({ |pair|
					(i = args[0].indexOf(pair[0])).notNil.if({
							// concat new info from appropriate key
						args[1].put(i, args[1][i] ++ pair[1])
					}, {
							// if key not present initially, add it to whole list
						args = args.add(pair);
					});
				});
				args = args.flop.flatten(1);	// reverse clump/flop from earlier
			});
		});
		view.notNil.if({ view.refresh });
	}
	
	asSeqPat {
		var sq;
		(seq.size == 0).if({	// if seq is empty, make sequence shortcut
				// is there something in the notesequence?
				// if not, give warning and do nothing
				// if so, make a Pseq from it
			(sq = this.asNoteSequence).isNil.if({
"SeqBuilder-draggedIntoVoicerGUI: SeqBuilder pattern is empty. Sequence not built.".warn;
			}, {
				seq = [sq];
			});
		});
			// interpret pattern prototype and drop in sequence
		^("{ arg x; " ++ patProto ++ " }").interpret.value(seq);
	}
	
	asNoteSequence {
		(pitch.notNil and: { dur.notNil }).if({
			^NoteSequence.performList(\new, [pitch, dur, length, absLength, maxPlayBeats]
				++ args)
		}, {
			^nil		// no data, return something that will cause error
		});
	}
	
	asFormattedString {
		var str;
		str = "[ ";
		seq.do({ |s|
			str = str ++ s.asEditString ++ "\n\n";
		});
		^(str ++ "]")
	}
	
	guiClass { ^SeqBuilderGui }

}

//---------------------------------//

SeqBuilderGui : HJHObjectGui {
		// view objects
	var	allSink, pitchSink, durSink, lenSink, argsSink,
		durToLenButton, absLenButton, fixButton, clearButton, seqSize,
		patProtoBox,	// default Pseq(x, inf) - this gets interpreted with x as list of NoteSeq
		<mainView;
	
	guiBody { arg lay;
//"SeqBuilderGui-guiBody".postln;
//this.dumpBackTrace;
		layout = lay;
		
		seqSize = SCStaticText(layout, Rect(0, 0, 100, 20))
			.string_("Seq. size: " ++ model.seq.size);
		
		fixButton = SCButton(layout, Rect(0, 0, 30, 20))
			.states_([["fix"]])
			.action_({
				model.fix;
			});
		
		clearButton = SCButton(layout, Rect(0, 0, 40, 20))
			.states_([["clear"]])
			.action_({
				model.clear;
			});
		
			// so you can drag nil into a slot
		SCDragSource(layout, Rect(0, 0, 30, 20)).object_(nil).align_(\center);
		
			// if Pfunc or Prout is anywhere in the seq, it will be broken after editing
		SCButton(layout, Rect(0, 0, 40, 20))
			.states_([["edit"]])
			.action_({
				var oldseq, d;
				oldseq = model.seq;	// save sequence so object cache can be purged later
				d = Document.new("Edit SeqBuilder contents", model.asFormattedString)
					.syntaxColorize;
				d.onClose_({
"UPDATING SEQBUILDER FROM EDIT WINDOW".postln;
						model.seq_(d.string.interpret.insp);  // set seq to edited contents
						this.refresh;	// reflect changes
						oldseq.purgeObjectCache;
					});
			});

		patProtoBox = SCTextField(layout, Rect(0, 0, 200, 20))
			.string_(model.patProto)
			.action_({ arg txt;
				model.patProto_(txt.string)
			});
		
		SCButton(layout, Rect(0, 0, 20, 20))		// a clear button for prototype
			.states_([["X"]])
			.action_({
				patProtoBox.string_("").focus(true);	// should not do action
			});
		
		mainView = FlowView.new(layout, Rect(0, 0, 700, 100));
		
		SCStaticText(mainView, Rect(0, 0, 50, 20)).string_("All:")
			.align_(\right);
		allSink = SCDragSink(mainView, Rect(0, 0, 30, 20)).action_({ arg rec;
			rec.object.tryPerform(\draggedIntoSeqBuilder, model, \all_)
		});
		
		SCStaticText(mainView, Rect(0, 0, 50, 20)).string_("Pitch:")
			.align_(\right);
		pitchSink = SCDragSink(mainView, Rect(0, 0, 30, 20)).action_({ arg rec;
			rec.object.tryPerform(\draggedIntoSeqBuilder, model, \pitch_)
		});

		SCStaticText(mainView, Rect(0, 0, 50, 20)).string_("Dur:")
			.align_(\right);
		durSink = SCDragSink(mainView, Rect(0, 0, 30, 20)).action_({ arg rec;
			rec.object.tryPerform(\draggedIntoSeqBuilder, model, \dur_)
		});
		
		durToLenButton = SCButton(mainView, Rect(0, 0, 30, 20))
			.states_([[">>"]])
			.action_({ arg b;
				model.length_(model.dur)
			});

		SCStaticText(mainView, Rect(0, 0, 50, 20)).string_("Length:")
			.align_(\right);
		lenSink = SCDragSink(mainView, Rect(0, 0, 30, 20)).action_({ arg rec;
			rec.object.tryPerform(\draggedIntoSeqBuilder, model, \length_)
		});
		
		absLenButton = SCButton(mainView, Rect(0, 0, 30, 20))
			.states_([["rel"], ["abs"]])
			.action_({ arg b;
				model.absLength_(b.value > 0)
			})
			.value_(1);	// default is absolute length

		SCStaticText(mainView, Rect(0, 0, 70, 20)).string_("Args:")
			.align_(\right);
		argsSink = SCDragSink(mainView, Rect(0, 0, 30, 20)).action_({ arg rec;
			rec.object.tryPerform(\draggedIntoSeqBuilder, model, \args_)
		});
		
		masterLayout.recursiveResize;
		model.view = this;
		this.refresh;
	}
	
	remove {
		model.view = nil;	// break connection
	}
	
	refresh {
		pitchSink.object = model.pitch;	// fix dragsink objects
		durSink.object = model.dur;
		lenSink.object = model.length;
		argsSink.object = model.args;
			// update absLenButton: user might change it in the model programmatically
		{	absLenButton.value_(model.absLength.binaryValue);
			[allSink, pitchSink, durSink, lenSink, argsSink].do({ |sink|
					// nils will be trapped and displayed as empty string
				sink.string_((sink.object ?? { "" }).asString.copyRange(0, 9));
			});
			seqSize.string_("Seq. size: " ++ model.seq.size);
		}.defer;
	}

}