
// this class stores parameters needed to define a mixer channel
// mixers should be customizable for any number of channels, any panning methodology
// define all that here
// h. james harkins -- jamshark70@dewdrop-world.net

MixerChannelDef {

		// the setters are there for the sake of inheritance
		// to make a similar mcdef, .copy and existing one and replace the instance
		// variables you need to change

	var	<>name, <>inChannels, <>outChannels,
		<>basicFader, <>postSendReadyFader,
		<>controls, <>guidef;

	*initClass {
		Class.initClassTree(Collection);
		Library.global.put(\mixerdefs, IdentityDictionary.new);
		MixerChannelDef(\mix1x1, 1, 1,
			basicFader: SynthDef("mixers/Mix1x1", {	// mono-to-stereo mixer channel
						arg busin, busout, level;
						Out.ar(busout, In.ar(busin, 1) * level);
					}), 
			postSendReadyFader: SynthDef("mixers/Mxb1x1", {						arg busin, busout, level;
						var in;
						in = In.ar(busin, 1) * level;
							// so that mixerchan bus can be used as postsendbus
						ReplaceOut.ar(busin, in);
						Out.ar(busout, in);
					}), 
			controls: (level: { |name| MixerControl(name, nil, 0.75, \amp) })
		);
		
		MixerChannelDef(\mix1x2, 1, 2,
			basicFader: SynthDef("mixers/Mix1x2", {	// mono-to-stereo mixer channel
						arg busin, busout, level, pan;
						Out.ar(busout, Pan2.ar(In.ar(busin, 1), pan, level));
					}), 
			postSendReadyFader: SynthDef("mixers/Mxb1x2", {						arg busin, busout, level, pan;
						var in, out;
						in = In.ar(busin, 1);
						out = Pan2.ar(in, pan, level);
							// so that mixerchan bus can be used as postsendbus
						ReplaceOut.ar(busin, out);
						Out.ar(busout, out);
					}), 
			controls: (level: { |name| MixerControl(name, nil, 0.75, \amp) },
						pan: { |name| MixerControl(name, nil, 0, \bipolar) }
			)
		);
		
		MixerChannelDef(\mix2x2, 2, 2,
			basicFader: SynthDef("mixers/Mix2x2", {	// stereo-to-stereo mixer channel
						arg busin, busout, level, pan;
						var l, r;
						#l, r = In.ar(busin, 2);
						Out.ar(busout, Balance2.ar(l, r, pan, level));
					}), 
			postSendReadyFader: SynthDef("mixers/Mxb2x2", {						arg busin, busout, level, pan;
						var l, r, out;
						#l, r = In.ar(busin, 2);
						out = Balance2.ar(l, r, pan, level);
						ReplaceOut.ar(busin, out);
						Out.ar(busout, out);
					}), 
			controls: (level: { |name| MixerControl(name, nil, 0.75, \amp) },
						pan: { |name| MixerControl(name, nil, 0, \bipolar) }
			)
		);

	}

	*new { |name, inChannels, outChannels, basicFader, postSendReadyFader, controls, guidef|
		inChannels.isNil.if({
			^this.at(name)
		}, {
			^super.newCopyArgs(name, inChannels, outChannels, basicFader, postSendReadyFader,
				controls, guidef).init
		});
	}
	
		// validate inputs and store in Library
	init {
		name = name.asSymbol;
		basicFader = basicFader.asSynthDef;
		postSendReadyFader = postSendReadyFader.asSynthDef;
		controls.respondsTo(\keys).not.if({
			MethodError("Error constructing MixerChannelDef - controls must be a Dictionary",
				this).throw
		});
		
		Library.global.put(\mixerdefs, name, this);
	}
	
	*at { |name|
		^Library.global[\mixerdefs, name]
	}
	
	*all { ^Library.global[\mixerdefs] }

		// instance methods

	synthdef { |postSendReady = false|
		^postSendReady.if({ postSendReadyFader }, { basicFader })
	}
	
	defName { |postSendReady = false|
		^this.synthdef(postSendReady).name
	}

	makeControlArray { |mixerName|
		var out = IdentityDictionary.new;
		controls.keysValuesDo({ |key, gcfunc|
			out.put(key, gcfunc.value("% %".format(mixerName, key)));
		});
		^out
	}

}

