
PeakMonitor {
		// back end for graphic peak follower

	classvar	<all;	// kr bus # -> monitor object
//			<oldResponder,	// server -> previous inhabitant of OSCresponder for c_setn
//			<serversInited;

	var	<target,	// may be bus or mixer
		<synth,	// writes peak level(s) onto control bus(es)
		<bus,
		<synthTarget,
		<freq,
		<>peaks,
		resp,	// OSCresponderNode
		<updater,
		<mixer;	// used only when PM is hitting a mixerchannel
		
	*initClass {
		8.do({ arg i;
			SynthDef.writeOnce("sys-PeakMon" ++ (i+1), {
				arg bus, kbus, t_trig;  // t_trig lets client control timing/reset of Peak
				var sig;
				sig = Peak.ar(In.ar(bus, i+1), t_trig);
				Out.kr(kbus, sig);
			});
		});
		all = IdentityDictionary.new;
		CmdPeriod.add(this);
	}

	*new { arg target = 0, freq = 2, layout;
		^super.new.init(target, freq, layout)
	}
	
	init { arg t, f = 2, layout;
		freq = f ? 2;
		t.isKindOf(MixerChannel).if({
			mixer = t;	// save for free
			t.postSendReady.if({
				target = t.inbus;	// if it's postsendready, fader's output is on the mc's bus
			}, {
					// need to isolate the fader's output on a new bus
				t.xfer.isNil.if({ t.makeXfer });
				t.xfer.addClient(this);
				target = t.xfer.bus;
			});
			synthTarget = t.fadergroup;	// place at tail of fadergroup
		}, {
				// convert other types of args
			target = t.asBus(\audio, 2, t.tryPerform(\server) ? Server.default);
			synthTarget = target.server.asTarget;  // go to rootnode
		});
		
		this.prInit;
		
		bus = Bus.control(target.server, target.numChannels);
		peaks = Array.fill(target.numChannels, 0);
		synth = Synth.tail(synthTarget, "sys-PeakMon" ++ (target.numChannels),
			[\bus, target.index, \kbus, bus.index, \freq, freq]);
		
		all.put(bus.index, this);	// so OSCresponder can find me
		updater = Routine({
					// first get the results from the last period
					// then reset the Peak ugen immediately after -- bundling ensures timing
			{	target.server.sendBundle(nil, [\c_getn, bus.index, bus.numChannels], 
					[\n_set, synth.nodeID, \t_trig, 1]);
				freq.reciprocal.wait;
			}.loop
		}).play(SystemClock);
		
		this.gui(layout);

	}
	
	freq_ { arg f;
		freq = f;
//		synth.notNil.if({
//			synth.set(\freq, freq);
//		});
	}
	
	free {
		updater.stop;
			// if there's a gui, drop it
		this.dependants.do({ arg d; d.free });
		resp.remove;
		synth !? { synth.free };
		bus !? {
			all.removeAt(bus.index);
			bus.free;
		};
		mixer.notNil.if({
			mixer.removeClient(this);
		});
		synth = bus = peaks = target = synthTarget = updater = mixer = nil;
	}
	
	*freeAll {
		all.do({ arg p; p.free });
	}
	
	*cmdPeriod {
		this.freeAll;
	}
	
	numChannels { ^target.numChannels }
	
	guiClass { ^PeakMonitorGui }

	asString { ^"PeakMonitor(" ++ target.asString ++ ", " ++ freq ++ ")" }


/// private	
	prInit {	// called first time
			// target should already be set

				// make my responder
		resp = OSCresponderNode(synthTarget.server.addr, "/c_setn", { arg t, r, m;
			PeakMonitor.all.at(m.at(1)).tryPerform(\peaks_, m.copyRange(3, m.size-1)).changed;
		}).add;
	}
}


PeakMonitorGui : ObjectGui {
	var	leftFlow, rightFlow;
	var	multiSl, clipButtons, maxRecentView;
	var	maxRecent, recentSize, maxClip;
	
	guiBody { arg lay;
		lay.startRow;
		leftFlow = FlowView(lay, Rect(0, 0, model.numChannels * 13 + 10, 210));
		multiSl = SCMultiSliderView(leftFlow, Rect(0, 0, model.numChannels * 13 + 2, 200))
			.size_(model.numChannels);
		
		rightFlow = FlowView(lay, Rect(0, 0,
				// ax + b(x-1) = ax + bx - b = (a+b)x - b
			max((lay.decorator.gap.x + 50) * model.numChannels - lay.decorator.gap.x + 10, 210),
			210));	// height

		clipButtons = Array.fill(model.numChannels, { arg i;
			SCButton(rightFlow, Rect(0, 0, 50, 20))
				.states_([
					["ok", Color.black, Color.grey],
					["CLIP", Color.white, Color.red]
				])
				.action_({ arg b;
					Post << "Channel " << i << " maximum clip: " << maxClip[i] << "\n";
					maxClip[i] = 0;
					b.value_(0)
				});
		});
		rightFlow.startRow;
		maxRecentView = SCTextView(rightFlow, Rect(0, 0, 200, 100));
		
		maxClip = 0 ! model.numChannels;
		recentSize = model.freq*2;
		maxRecent = Array.new(recentSize);
	}
	
	update {
		var newpeaks, str;
		multiSl.isActive.if({	// if you close the window at the wrong time, this prevents a choke
			newpeaks = model.peaks.collect({ arg p, i; 
				((p = p.abs) > 1).if({
					(p > maxClip[i]).if({ maxClip[i] = p });
					{ clipButtons.at(i).value_(1) }.defer;
				});
				p.clip(0,1)
			});
				// update "recent peak" display
			(maxRecent.size >= recentSize).if({ maxRecent.removeAt(0) });
			maxRecent.add(newpeaks);
			str = "Channel peaks:\n\n";
			maxRecent.flop.do({ |chan, i|
				str = str ++ format("%: % dB\n", i+1, chan.maxItem.ampdb.round(0.01));
			});
			{	multiSl.value_(newpeaks.sqrt);
				maxRecentView.setString(str, 0, maxRecentView.string.size);
			}.defer;
		});
	}
	
	remove {
		model.free;
	}
	
}