//
//NoteSequence {
//		// maintains several streams relevant to note-based sequencing
//		// not a Stream subclass because I need it NOT to be embedded in streams
//		
//	var	<freqs, <durs, <lengths, <argStreams,	// streams to generate values
//		<>origFreqs, <>origDurs, <>origLengths, <>origArgStreams,  // for reset
//		<absLengths,		// are lengths absolute (beats) or relative (multiply by dur)?
//		<maxPlayBeats,	// how many beats should this sequence run? (nil = no limit)
//		lengthFunc;		// internal: for efficiency calculating note lengths
//		
//	var	<stream;
//
//var <>clock;
//
//	*new { arg freqs, durs, lengths, absLengths = true, maxPlayBeats ... args;
//		^super.new.init(freqs, durs, lengths, absLengths, maxPlayBeats, args)
//	}
//	
//	storeArgs {
//		^[origFreqs, origDurs, origLengths, absLengths, maxPlayBeats,
//			origArgStreams.asSortedArray.flatten(1)]
//	}
//
//	init { arg f, d, l, absL, maxPlay, a;
//			// save the easy ones
//		origFreqs = f;
//		origDurs = d;
//		origLengths = l;
//		absLengths = absL ? true;
//		maxPlayBeats = maxPlay ? inf;		// if nil, no limit
//		
//			// other args have to be made into an IdentityDict
//		a = a.flat.clump(2);
//		origArgStreams = IdentityDictionary.new.addAll(
//			Array.fill(a.size, { arg i;
//				a.at(i).at(0) -> a.at(i).at(1)
//			})
//		);
//		this.reset;
//	}
//	
//	next { ^stream.next }
//	
//	value { ^stream.next }
//	
//	reset {
//		freqs = origFreqs.asStream;
//		durs = origDurs.asStream;
//		lengths = origLengths.asStream;
//		argStreams = origArgStreams.collect({ arg pat; pat.asStream });
//		lengthFunc = this.lengthFunction;
//		stream = this.asStream;		// because I'm not a stream, I have to implement .next
//	}
//	
//	freqs_ { arg newStream;
//		freqs = origFreqs = newStream.asStream;
//	}
//	
//	durs_ { arg newStream;
//		durs = origDurs = newStream.asStream;
//	}
//	
//	lengths_ { arg newStream;
//		lengths = origLengths = newStream.asStream;
//		lengthFunc = this.lengthFunction;
//	}
//	
//	absLengths_ { arg absL = true;
//		absLengths = absL;
//		lengthFunc = this.lengthFunction;
//	}
//	
//	maxPlayBeats_ { arg mpb;
//		maxPlayBeats = mpb ? inf;
//	}
//	
//	putArg { arg name, newStream;
//		argStreams.at(name).notNil.if({
//			argStreams.removeAt(name);
//		});
//		argStreams.put(name, newStream.asStream);
//	}
//	
//	asStream {
//		^Routine({
//			var noNils = true, f, d, l, a, beatsPlayed = 0;
//				// stop when any stream (except lengths) becomes nil
//				// also stop when maxPlayBeats has been reached
//			{ noNils and: { beatsPlayed < maxPlayBeats } }.while({
//				f = freqs.next ?? { noNils = false; nil };
//				d = durs.next ?? { noNils = false; nil };
//					// if playing this note will exceed maxPlayBeats...
////"\t".post; [clock.elapsedBeats.trunc(0.01), beatsPlayed, d].postln;
//				((beatsPlayed = beatsPlayed + (d ? 0)) > maxPlayBeats).if({
//					d = beatsPlayed - maxPlayBeats;	// ...shorten d to fill maxPlayBeats exactly
////"\t\t".post; ["Correcting dur to: ", d].postln;
//				});
//					// lengths may be nil; just use duration in that case
//				l = lengthFunc.value(lengths.next ? 1, d ? 1);
//				(argStreams.size > 0).if({
//					a = Array.new;
//					argStreams.keysValuesDo({ arg k, v;
//						v = v.next ?? { noNils = false; nil };
//						a = a ++ [k, v]
//					});
//				}, {
//					a = nil;
//				});
//				
//				noNils.if({
//					SequenceNote(f, d, l, a).yield
//				}, {
//					nil.yield
//				});
//			});
//			nil.yield;	// reached when noNils becomes false or maxPlayBeats is exceeded
//		});
//	}
//	
//	lengthFunction {
//		origLengths.isNil.if({
//			^{ arg l, d; d }
//		});
//		origLengths.isKindOf(SimpleNumber).if({
//			^{ arg l, d; d*l }
//		});
//		absLengths.if({
//			^{ arg l, d; l }
//		}, {
//			^{ arg l, d; d*l }
//		});
//	}
//
//	postNotes { arg numToPrint = 50;
//		var i = 0, note, stream;
//		stream = this.asStream;
//		{ (i < numToPrint) and: (note = this.next).notNil }.while({
//			(i.asString ++ "\t" ++ note.asCompileString).postln;
//			i = i+1;
//		});
//	}
//	
//	storeEditOn { arg stream, cacheKey;
//		var args;
//		stream << "NoteSequence(\n";
//		(args = this.storeArgs).do({ arg a, i;
//			stream << "\t";
//			a.storeEditOn(stream, cacheKey);
//			(i < (args.size - 1)).if({ stream << ",\n" });
//		});
//		stream << "\n)";
//	}
//	
////	asCompileString {
////			// does not represent Pfunc safely
////		^"NoteSequence( " ++ origFreqs.asCompileString ++ ", " ++ origDurs.asCompileString ++ ", "
////			++ origLengths.asCompileString ++ ", " ++ absLengths.asString ++ ", "
////			++ maxPlayBeats ++ ", " ++ origArgStreams.asCompileString ++ ")"
////	}
//	
//}

	// now a subclass of AbstractFunction so that you can do math on them
	// this is the simple variety: no grace notes or chord notes
	// can be played simply by using the vars as voicer parms for example
	// typically you're doing math on the frequencies
SequenceNote : AbstractFunction {
	var	<>freq, <>dur, <>length, <>args;		// values for this note

		// note, newCopyArgs will break if any instance variables are added to AbstractFunction
	*new { arg freq, dur, length, args;
		^super.newCopyArgs(freq, dur, length, args);
	}
	
	mapMode { |mode|
		^this.copy.freq_(freq.mapMode(mode))
	}

	unmapMode { |mode|
		^this.copy.freq_(freq.unmapMode(mode))
	}
	
	asString { ^this.asArray.asCompileString }
	
	storeArgs { ^[freq, dur, length, args] }
	
	composeUnaryOp { |selector|
		^this.copy.freq_(freq.perform(selector))
	}
	composeBinaryOp { arg aSelector, note;
		^this.copy.freq_(note.numPerformBinaryOpOnNumber(aSelector, freq))
	}

// this needs to be tested
	reverseComposeBinaryOp { arg aSelector, something;
		^something.perform(aSelector, this.asFloat)
	}
	composeNAryOp { arg aSelector, anArgList;
		^this.copy.freq_(freq.performList(aSelector, anArgList))
	}

	// double dispatch for mixed operations
	// this handles the situation of 5 + SequenceNote() -- a note should be returned
	performBinaryOpOnSimpleNumber { arg aSelector, aNumber;
		var result;
			
		(result = aNumber.perform(aSelector, this.freq)).respondsTo('+').if({
			^this.copy.freq_(result)	// note should be output only if the result is a number
		}, {
			^result
		});
	}
	// this handles note + number, but returns number so composeBinaryOp can wrap it correctly
	numPerformBinaryOpOnNumber { arg aSelector, aNumber;
		^aNumber.perform(aSelector, freq)
	}
		// have to think through this one	
//	performBinaryOpOnSeqColl { arg aSelector, theOperand;
//		theOperand.performBinaryOp(aSelector, freq)
//	}

		// compares frequency only. To compare the whole note use identity ===
		// these do not work if you do 3 == aNote
	== { |that| ^that.numPerformBinaryOpOnNumber('==', this.freq) }
	!= { |that| ^that.numPerformBinaryOpOnNumber('!=', this.freq) }
	< { |that| ^that.numPerformBinaryOpOnNumber('<', this.freq) }
	> { |that| ^that.numPerformBinaryOpOnNumber('>', this.freq) }
	<= { |that| ^that.numPerformBinaryOpOnNumber('<=', this.freq) }
	>= { |that| ^that.numPerformBinaryOpOnNumber('>=', this.freq) }
	
		// some freq calculations will require end result to be a number
	asInteger { ^freq.asInteger }
	asFloat { ^freq.asFloat }
	
		// support for more complex note types
	asPlayableNote { ^this }
	
//			// nothing interesting here but SeqChordNote and SequenceItem do other stuff
//	freq_recalc { |newFreq| freq = newFreq; }
	
	convertGates {
		args.isKindOf(SimpleNumber).if({
			^this.copy.args_([\gate, args])
		});
	}	// else, output this by default
	
		// simplify parsing
	isGraceNote { |deltaThresh = 0.1, overlapThresh = 0.1|
			// gracenotes must be short and have a small overlap
		^(dur < deltaThresh) and: { (length - dur) < overlapThresh }
	}

		// add chord notes -- this changes the type to SeqChordNote
	add { |note|
		^SeqChordNote(freq, dur, length, args).add(note)
	}
	++ { |notes|
		^SeqChordNote(freq, dur, length, args) ++ notes
	}
	
	addGraceNotes { |notes|
		^SequenceItem(freq, dur, length, args, notes)
	}

	asSequenceNote { ^this }
	asMelodyNote { ^this }
	asTopNote { ^this }
	
	asNoteArray { ^[this] }
	
	isChord { ^false }		// only 1 note
	
	asArray { ^[freq, dur, length, args] }
	
		// more complex types have to be read as a stream (to play grace notes)
		// this is for compatibility
	asPattern { ^Pn(this, 1) }
	asStream { ^this.asPattern.asStream }

	shortDur { ^dur }	// var is not needed here, but subclasses use it
					// all the different note types should be interchangeable

}

// the next 2 classes will allaw chords and gracenotes to be played into a melodic
// sequence while melodic process calculate off the main note
// the math implementation for all is the same, but the play implementation is different

SeqChordNote : SequenceNote {	// chord, no grace notes
	var	<>chordNotes,		// array of SequenceNotes - should not contain SeqChordNotes
		<>shortDur;		// shortDur = dur as played; dur = dur + dur of chordnotes
						// so that you can play w/o grace notes and keep rhythm	
						// chord notes are stored relative to freq for easy transposition

	*new { arg freq, dur, length, args, chordNotes;
		^super.new(freq, dur, length, args)
			.shortDur_(dur)	// initialize shortdur
			++ chordNotes
	}
	
	storeArgs { ^[freq, shortDur, length, args, chordNotes+freq] }
	
	copy { ^this.deepCopy }		// must copy everything or there will be trouble
	
	convertGates {
		chordNotes = chordNotes.collect({ |n| n.convertGates });
		^super.convertGates	// handle the main note
	}
	
		// works with only one note
	add { |note|
		dur = dur + note.dur;	// chord dur must be the *sum* of the note deltas
		chordNotes = chordNotes.asArray ++ (note.asMelodyNote - freq)
	}
	
	++ { |notes|
		chordNotes = chordNotes.asArray ++ notes.collect({ |n| 
			dur = dur + n.dur;  // faster than .collect({ |n| n.dur }).sum (then another collect)
			n.asMelodyNote - freq
		})
	}

	addGraceNotes { |notes|
		^SequenceItem(freq, dur, length, args, notes, chordNotes)
	}
	
//	freq_recalc { |newFreq|
//		var	freqAdj;
//		freqAdj = newFreq - freq;
//		chordNotes.do({ |n| n.freq_(n.freq - freqAdj) });
//		freq = newFreq;
//	}
//	
		// outputs a new object
	mapMode { |mode|
		var	newFreq;
		newFreq = freq.mapMode(mode);
		^SeqChordNote(newFreq, dur, length, args,
			chordNotes.isNil.if({ nil },
				{ (chordNotes + freq).mapMode(mode) }))
	}
	
	unmapMode { |mode|
		var	newFreq;
		newFreq = freq.unmapMode(mode);
		^SeqChordNote(newFreq, dur, length, args,
			chordNotes.isNil.if({ nil },
				{ (chordNotes + freq).unmapMode(mode) }))
	}
	
		// changing duration needs to scale all rhythmic values
	dur_ { |newDur|
		var	ratio;
		ratio = newDur / dur;
		chordNotes.do({ |n|
			n.dur_(n.dur * ratio); // .length_(n.length * ratio)
		});
		shortDur = shortDur * ratio;
		dur = newDur;
	}

	asPlayableNote {
		var	f, d, l, a, newSize;
		(chordNotes.size > 0).if({
			f = Array.new(newSize = chordNotes.size + 1).add(freq);
			l = Array.new(newSize).add(length);
			a = Array.new(newSize).add(args);
			chordNotes.do({ |n|
				f.add(n.freq);
				l.add(n.length);
				a.add(n.args);
			});
			^SequenceNote(f, dur, l, a)
		}, {
			^this
		});
	}
	
	asNoteArray { ^[this.asMelodyNote] ++ (chordNotes + freq) }
	
	isChord { ^chordNotes.size > 0 }
	
		// if this method is used, chord notes are not wanted but grace notes might be
	asMelodyNote { ^this.copy.chordNotes_(nil) }
	asSequenceNote { ^SequenceNote(freq, dur, length, args) }
	asTopNote {
		var	highNote;
			// chordNotes must exist for the test to be valid
		((highNote = chordNotes !? { chordNotes.maxItem }).notNil
			and: { highNote + freq > this }).if({
				^SequenceNote(highNote.freq + freq, dur, length, args)
			}, {
				^this.asSequenceNote;
			});
	}
			
	
	asPattern { ^Pseq([SequenceNote(freq, shortDur, length, args)] ++ (chordNotes + freq), 1) }
	embedInStream { |inval| ^this.asPattern.asStream.embedInStream(inval) }
}

SequenceItem : SeqChordNote {	// add grace notes
	var	<>graceNotes;
	
	*new { arg freq, dur, length, args, graceNotes, chordNotes;
		^(super.new(freq, dur, length, args) ++ chordNotes)
			.shortDur_(dur)	// initialize shortdur
			.addGraceNotes(graceNotes)
	}

	storeArgs { ^[freq, shortDur, length, args, graceNotes+freq, chordNotes+freq] }

	mapMode { |mode|
		var	newFreq;
		newFreq = freq.mapMode(mode);
		^SequenceItem(newFreq, dur, length, args,
			graceNotes.isNil.if({ nil },
				{ (graceNotes + freq).mapMode(mode) }),
			chordNotes.isNil.if({ nil },
				{ (chordNotes + freq).mapMode(mode) }))
	}
	
	unmapMode { |mode|
		var	newFreq;
		newFreq = freq.unmapMode(mode);
		^SequenceItem(newFreq, dur, length, args,
			graceNotes.isNil.if({ nil },
				{ (graceNotes + freq).unmapMode(mode) }),
			chordNotes.isNil.if({ nil },
				{ (chordNotes + freq).unmapMode(mode) }))
	}
	
//	freq_recalc { |newFreq|
//		var	freqAdj;
//		freqAdj = newFreq - freq;
//		chordNotes.do({ |n| n.freq_(n.freq - freqAdj) });
//		graceNotes.do({ |n| n.freq_(n.freq - freqAdj) });
//		freq = newFreq;
//	}
//	
	dur_ { |newDur|
		var	ratio;
		ratio = newDur / dur;
		chordNotes.do({ |n|
			n.dur_(n.dur * ratio); // .length_(n.length * ratio)
		});
			// could use super.dur_ but this saves a dispatch
		graceNotes.do({ |n|
			n.dur_(n.dur * ratio); // .length_(n.length * ratio)
		});
		shortDur = shortDur * ratio;
		dur = newDur;
	}

	convertGates {
		graceNotes = graceNotes.collect({ |n| n.convertGates });
		^super.convertGates	// handle the chord notes and main note
	}
	
	addGraceNotes { |notes|
		graceNotes = graceNotes.asArray ++ notes.collect({ |n| 
			dur = dur + n.dur;
			n.asSequenceNote - freq		// or asMelodyNote?
		});
	}
	
	// asPlayableNote should drop the grace notes because grace notes can't be represented
	// as one playable note - hence superclass implementation is sufficient

	asPattern { ^Pseq(
			(graceNotes.size > 0).if({ [Pseq(graceNotes + freq, 1)] }, { Array.new })
				++ SequenceNote(freq, shortDur, length, args)
				++ (chordNotes.size > 0).if({ Pseq(chordNotes + freq, 1) })
			, 1)
	}
	
}
