
#include "SC_PlugIn.h"

static InterfaceTable *ft;
#define RGET \
	RGen& rgen = *unit->mParent->mRGen; \
	uint32 s1 = rgen.s1; \
	uint32 s2 = rgen.s2; \
	uint32 s3 = rgen.s3; 

#define RPUT \
	rgen.s1 = s1; \
	rgen.s2 = s2; \
	rgen.s3 = s3;

struct MyDelay : public Unit
{
	uint32 mDelayLength;
	uint32 mPosition;
	float *mData; 
	float mDecay;
	float alpCoef;
	float *alpFF; // feedforward delay for allpass
	float alpFB; // feedback for allpass
	uint32 alpPos;  // pointer for alpass fb delay
	uint32 lPosition; // pointer for lowpass delay
	float *lData;		// lowpass delay
};
 
extern "C"
{
	void load(InterfaceTable *inTable);
	void MyDelay_next_notfull(MyDelay *unit, int inNumSamples);
	void MyDelay_next_full(MyDelay *unit, int inNumSamples);
	void MyDelay_Ctor(MyDelay* unit);
	void MyDelay_Dtor(MyDelay* unit);
};

//////////////////////////////////////////////////////////////////

void MyDelay_Ctor(MyDelay* unit)
{
	float d;
	SETCALC(MyDelay_next_notfull);

	unit->mDelayLength = (uint32)( SAMPLERATE/IN0(1));
	unit->mData = (float*)RTAlloc(unit->mWorld, unit->mDelayLength * sizeof(float));
	unit->mPosition = 0;
	unit->mDecay = IN0(2);
	unit->lData = (float*)RTAlloc(unit->mWorld, 2 * sizeof(float));
	unit->lPosition = 0;	

	d = (float)(SAMPLERATE/IN0(1)) - unit->mDelayLength; //fractional delay
	unit->alpCoef = (1.0 - d)/(1.0 + d);  // allpass coefficient
	unit->alpFF = (float*)RTAlloc(unit->mWorld, 2 * sizeof(float));
	if(unit->alpCoef == 1) unit->alpCoef = .99;
	unit->alpFB = 0;
	unit->alpPos = 0;
	bzero(unit->mData, unit->mDelayLength*4);	
	bzero(unit->lData, 8);	
	bzero(unit->alpFF, 8);	
	MyDelay_next_notfull(unit, 1);
}

//////////////////////////////////////////////////////////////////


void MyDelay_Dtor(MyDelay* unit)
{
	RTFree(unit->mWorld, unit->mData);
	RTFree(unit->mWorld, unit->lData);
	RTFree(unit->mWorld, unit->alpFF);
}


//////////////////////////////////////////////////////////////////

void MyDelay_next_notfull(MyDelay *unit, int inNumSamples)
{
	float *out = OUT(0);
	
	float *in = IN(0);

	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	float decay = unit->mDecay;
	float *ldata = unit->lData;
	uint32 lposition = unit->lPosition;
	
	float *alpFF = unit->alpFF;
	float alpFB = unit->alpFB;
	float alpCoef = unit->alpCoef;
	uint32 alpPos = unit->alpPos;
	RGET
 	bool wrapped = false;
	
	 	for (int i=0; i < inNumSamples; ++i)
	{
 		float z ;
		// comb
		z = frand(s1,s2,s3) + (data[position] * decay);
 		
		// lowpass
		ldata[lposition++] = z;
		lposition = lposition%2;
 		z = 0.5 *(z + ldata[lposition] );
		
		// allpass
		//alpFF[alpPos++] = z;
		//alpPos = alpPos%2;
		//z = alpFF[alpPos]  +alpCoef * (z - alpFB);
		//alpFB = z;

	  	data[position] = z;
		if (++position >= length) {
			position = 0; 
			wrapped = true;  
	 		SETCALC(MyDelay_next_full);

		};
 		out[i] = z;
	}
	unit->mPosition = position;
	unit->lPosition = lposition;
	unit->alpPos = alpPos;
	RPUT
}
	 
//////////////////////////////////////////////////////////////////

 void MyDelay_next_full(MyDelay *unit, int inNumSamples)
{
 	float *out = OUT(0);
	
 	float *in = IN(0);
	
 	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	float decay = unit->mDecay;
	
	float *ldata = unit->lData;
	uint32 lposition = unit->lPosition;
	
	float *alpFF = unit->alpFF;
	float alpFB = unit->alpFB;
	float alpCoef = unit->alpCoef;
	uint32 alpPos = unit->alpPos;
	
 	for (int i=0; i < inNumSamples; ++i)
	{
 		float z ;
		// comb
		//z = in[i] + (data[position] * decay);
		 z =    (data[position] * decay);
 		
		// lowpass
		ldata[lposition++] = z;
		lposition = lposition%2;
 		z = 0.5 *(z + ldata[lposition] );
		
		// allpass
	 	//alpFF[alpPos++] = z;
	 	//alpPos = alpPos%2;
	 	//z = alpFF[alpPos]  + (alpCoef * (z - alpFB));
	 	//alpFB = z;

	  	data[position] = z;
		if (++position >= length) {
			position = 0;  
		};
		
		out[i] = z;
	}

	unit->mPosition = position;
	unit->lPosition = lposition;
	unit->alpPos = alpPos;

}

 
void load(InterfaceTable *inTable)
{
	ft = inTable;

	DefineDtorUnit(MyDelay);
}

