/////////////// C++ code in a file named TestOsc.cpp ///////////////////

#include "SC_PlugIn.h"
#include <limits.h>

static InterfaceTable *ft;

struct BufUnit : public Unit
{
SndBuf *m_buf;
float m_fbufnum;
};

struct TableLookup : public BufUnit
{
double m_cpstoinc, m_radtoinc;
int32 mTableSize;
int32 m_lomask;
};

struct TestOsc : public TableLookup
{
int32 m_phase, m_phaseoffset;
float m_phasein;
};

//////////////////////////////////////////////////////////////////////////////////////////////////

extern "C"
{
void load(InterfaceTable *inTable);

void TestOsc_Ctor(TestOsc *unit);
void TestOsc_next_ikk(TestOsc *unit, int inNumSamples);
void TestOsc_next_ika(TestOsc *unit, int inNumSamples);
void TestOsc_next_iak(TestOsc *unit, int inNumSamples);
void TestOsc_next_iaa(TestOsc *unit, int inNumSamples);
}

//////////////////////////////////////////////////////////////////////////////////////////////////

#define GET_TABLE \
float fbufnum = ZIN0(0); \
if (fbufnum != unit->m_fbufnum) { \
uint32 bufnum = (uint32)fbufnum; \
World *world = unit->mWorld; \
if (bufnum >= world->mNumSndBufs) bufnum = 0; \
unit->m_buf = world->mSndBufs + bufnum; \
} \
SndBuf *buf = unit->m_buf; \
if(!buf) { \
ClearUnitOutputs(unit, inNumSamples); \
return; \
} \
float *bufData __attribute__((__unused__)) = buf->data; \
if (!bufData) { \
ClearUnitOutputs(unit, inNumSamples); \
return; \
} \
int tableSize = buf->samples;

////////////////////////////////////////////////////////////////////////////////////////////////////////

void TestOsc_Ctor(TestOsc *unit)
{
unit->m_fbufnum = -1e9f;
if (INRATE(1) == calc_FullRate) {
if (INRATE(2) == calc_FullRate) {
//Print("next_iaa\n");
SETCALC(TestOsc_next_iaa);
} else {
//Print("next_iak\n");
SETCALC(TestOsc_next_iak);
}
} else {
if (INRATE(2) == calc_FullRate) {
//Print("next_ika\n");
SETCALC(TestOsc_next_ika);
} else {
//Print("next_ikk\n");
SETCALC(TestOsc_next_ikk);
}
}
unit->m_buf = unit->mWorld->mSndBufs;
unit->mTableSize = -1;
unit->m_phasein = ZIN0(2);
unit->m_phaseoffset = (int32)(unit->m_phasein * unit->m_radtoinc);
if (INRATE(2) == calc_FullRate) {
unit->m_phase = 0;
} else {
unit->m_phase = unit->m_phaseoffset;
}
TestOsc_next_ikk(unit, 1);
}

//////////////!!!

void TestOsc_next_ikk(TestOsc *unit, int inNumSamples)
{
// get table
GET_TABLE

float *table0 = bufData;
float *table1 = table0 + 1;
if (tableSize != unit->mTableSize) {
unit->mTableSize = tableSize;
int tableSize2 = tableSize >> 1;
unit->m_lomask = (tableSize2 - 1) << 3; // Osc, OscN, COsc, COsc, COsc2, OscX4, OscX2
unit->m_radtoinc = tableSize2 * (rtwopi * 65536.); // Osc, OscN, PMOsc
// SigOsc, Osc, OscN, PMOsc, COsc, COsc2, OscX4, OscX2
unit->m_cpstoinc = tableSize2 * SAMPLEDUR * 65536.;
}

float *out = ZOUT(0);
float freqin = ZIN0(1);
float phasein = ZIN0(2);

int32 phase = unit->m_phase;
int32 lomask = unit->m_lomask;

int32 freq = (int32)(unit->m_cpstoinc * freqin);
int32 phaseinc = freq + (int32)(CALCSLOPE(phasein, unit->m_phasein) * unit->m_radtoinc);
unit->m_phasein = phasein;

LOOP(inNumSamples,
ZXP(out) = lookupi1(table0, table1, phase, lomask);
phase += phaseinc;
);
unit->m_phase = phase;

}

void TestOsc_next_ika(TestOsc *unit, int inNumSamples)
{
// get table
GET_TABLE
float *table0 = bufData;
float *table1 = table0 + 1;
if (tableSize != unit->mTableSize) {
unit->mTableSize = tableSize;
int tableSize2 = tableSize >> 1;
unit->m_lomask = (tableSize2 - 1) << 3; // Osc, OscN, COsc, COsc, COsc2, OscX4, OscX2
unit->m_radtoinc = tableSize2 * (rtwopi * 65536.); // Osc, OscN, PMOsc
// SigOsc, Osc, OscN, PMOsc, COsc, COsc2, OscX4, OscX2
unit->m_cpstoinc = tableSize2 * SAMPLEDUR * 65536.;
}

float *out = ZOUT(0);
float freqin = ZIN0(1);
float *phasein = ZIN(2);

int32 phase = unit->m_phase;
int32 lomask = unit->m_lomask;

int32 freq = (int32)(unit->m_cpstoinc * freqin);
float radtoinc = unit->m_radtoinc;
//Print("Osc_next_ika %d %g %d\n", inNumSamples, radtoinc, phase);
LOOP(inNumSamples,
int32 phaseoffset = phase + (int32)(radtoinc * ZXP(phasein));
ZXP(out) = lookupi1(table0, table1, phaseoffset, lomask);
phase += freq;
);
unit->m_phase = phase;
//unit->m_phasein = phasein;

}

void TestOsc_next_iaa(TestOsc *unit, int inNumSamples)
{
// get table
GET_TABLE
float *table0 = bufData;
float *table1 = table0 + 1;
if (tableSize != unit->mTableSize) {
unit->mTableSize = tableSize;
int tableSize2 = tableSize >> 1;
unit->m_lomask = (tableSize2 - 1) << 3; // Osc, OscN, COsc, COsc, COsc2, OscX4, OscX2
unit->m_radtoinc = tableSize2 * (rtwopi * 65536.); // Osc, OscN, PMOsc
// SigOsc, Osc, OscN, PMOsc, COsc, COsc2, OscX4, OscX2
unit->m_cpstoinc = tableSize2 * SAMPLEDUR * 65536.;
}

float *out = ZOUT(0);
float *freqin = ZIN(1);
float *phasein = ZIN(2);

int32 phase = unit->m_phase;
int32 lomask = unit->m_lomask;

float cpstoinc = unit->m_cpstoinc;
float radtoinc = unit->m_radtoinc;
//Print("Osc_next_iaa %d %g %g %d\n", inNumSamples, cpstoinc, radtoinc, phase);
LOOP(inNumSamples,
int32 phaseoffset = phase + (int32)(radtoinc * ZXP(phasein));
float z = lookupi1(table0, table1, phaseoffset, lomask);
phase += (int32)(cpstoinc * ZXP(freqin));
ZXP(out) = z;
);
unit->m_phase = phase;
//unit->m_phasein = ZX(phasein);

}

void TestOsc_next_iak(TestOsc *unit, int inNumSamples)
{
// get table
GET_TABLE
float *table0 = bufData;
float *table1 = table0 + 1;
if (tableSize != unit->mTableSize) {
unit->mTableSize = tableSize;
int tableSize2 = tableSize >> 1;
unit->m_lomask = (tableSize2 - 1) << 3; // Osc, OscN, COsc, COsc, COsc2, OscX4, OscX2
unit->m_radtoinc = tableSize2 * (rtwopi * 65536.); // Osc, OscN, PMOsc
// SigOsc, Osc, OscN, PMOsc, COsc, COsc2, OscX4, OscX2
unit->m_cpstoinc = tableSize2 * SAMPLEDUR * 65536.;
}

float *out = ZOUT(0);
float *freqin = ZIN(1);
float phasein = ZIN0(2);

int32 phase = unit->m_phase;
int32 lomask = unit->m_lomask;

float cpstoinc = unit->m_cpstoinc;
int32 phaseinc = (int32)(CALCSLOPE(phasein, unit->m_phasein) * unit->m_radtoinc);
int32 phaseoffset = unit->m_phaseoffset;

LOOP(inNumSamples,
float z = lookupi1(table0, table1, phase + phaseoffset, lomask);
phaseoffset += phaseinc;
phase += (int32)(cpstoinc * ZXP(freqin));
ZXP(out) = z;
);
unit->m_phase = phase;
unit->m_phasein = phasein;
unit->m_phaseoffset = phaseoffset;

}
 
// declare struct to hold unit generator state
struct MyDelay : public Unit
{
	uint32 mDelayLength;
	uint32 mPosition;
	float *mData; // delay buffer
};

// declare unit generator functions 
extern "C"
{
	void load(InterfaceTable *inTable);
	void MyDelay_next_notfull(MyDelay *unit, int inNumSamples);
	void MyDelay_next_full(MyDelay *unit, int inNumSamples);
	void MyDelay_Ctor(MyDelay* unit);
	void MyDelay_Dtor(MyDelay* unit);
};

//////////////////////////////////////////////////////////////////

// Ctor is called to initialize the unit generator. 
// It only executes once.

// A Ctor usually does 3 things.
// 1. set the calculation function.
// 2. initialize the unit generator state variables.
// 3. calculate one sample of output.
void MyDelay_Ctor(MyDelay* unit)
{

	// 1. set the calculation function.
	SETCALC(MyDelay_next_notfull);

	// 2. initialize the unit generator state variables.
	// get the delay length
	unit->mDelayLength = (uint32)(IN0(1) * SAMPLERATE);
	
	// allocate the buffer
	unit->mData = (float*)RTAlloc(unit->mWorld, unit->mDelayLength * sizeof(float));
	// RTAlloc allocates out of the real time memory pool of the server 
	// which is finite. The size of the real time memory pool is set using the
	// -m command line argument of the server.
	
	// initialize the position
	unit->mPosition = 0;
	
	// 3. calculate one sample of output.
	MyDelay_next_notfull(unit, 1);
}

//////////////////////////////////////////////////////////////////

// Dtor is called to perform any clean up for the unit generator. 

void MyDelay_Dtor(MyDelay* unit)
{
	// free the buffer
	RTFree(unit->mWorld, unit->mData);
}


//////////////////////////////////////////////////////////////////

// The calculation function executes once per control period 
// which is typically 64 samples.

// calculation function when the buffer has not yet been filled
void MyDelay_next_notfull(MyDelay *unit, int inNumSamples)
{
	// get the pointer to the output buffer
	float *out = OUT(0);
	
	// get the pointer to the input buffer
	float *in = IN(0);
	
	// get values from struct and store them in local variables.
	// The optimizer will cause them to be loaded it into a register.
	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	bool wrapped = false;
	
	// perform a loop for the number of samples in the control period.
	// If this unit is audio rate then inNumSamples will be 64 or whatever
	// the block size is. If this unit is control rate then inNumSamples will
	// be 1.
	for (int i=0; i < inNumSamples; ++i)
	{
		// get old value in delay line
		float z = data[position];
		// store new value in delay line
		data[position] = in[i];
		
		// see if the position went to the end of the buffer 
		if (++position >= length) {
			position = 0; // go back to beginning
			wrapped = true; // indicate we have wrapped.
			
			// change the calculation function
			// next time, the MyDelay_next_full function will be called
			SETCALC(MyDelay_next_full);
		}
		// if we have not yet wrapped, then z is garbage from the uninitialized 
		// buffer, so output zero. If we have wrapped, then z is a good value.
		out[i] = wrapped ? 0.f : z;
	}

	// store the position back to the struct
	unit->mPosition = position;
}

//////////////////////////////////////////////////////////////////

// calculation function when the buffer has been filled
void MyDelay_next_full(MyDelay *unit, int inNumSamples)
{
	// get the pointer to the output buffer
	float *out = OUT(0);
	
	// get the pointer to the input buffer
	float *in = IN(0);
	
	// get values from struct and store them in local variables.
	// The optimizer will cause them to be loaded it into a register.
	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	
	// perform a loop for the number of samples in the control period.
	// If this unit is audio rate then inNumSamples will be 64 or whatever
	// the block size is. If this unit is control rate then inNumSamples will
	// be 1.
	for (int i=0; i < inNumSamples; ++i)
	{
		// get old value in delay line
		float z = data[position];
		// store new value in delay line
		data[position] = in[i];
		// see if the position went to the end of the buffer 
		if (++position >= length) {
			position = 0; // go back to beginning
		}
		out[i] = z;
	}

	// store the position back to the struct
	unit->mPosition = position;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void load(InterfaceTable *inTable)
{
ft = inTable;

DefineSimpleUnit(TestOsc);
DefineDtorUnit(MyDelay);

}
/////////

 
////////////////////////////////////////////////////////////////////
