
#include "SC_PlugIn.h"

static InterfaceTable *ft;
#define RGET \
	RGen& rgen = *unit->mParent->mRGen; \
	uint32 s1 = rgen.s1; \
	uint32 s2 = rgen.s2; \
	uint32 s3 = rgen.s3; 

#define RPUT \
	rgen.s1 = s1; \
	rgen.s2 = s2; \
	rgen.s3 = s3;

struct PluckIn : public Unit
{
	uint32 mDelayLength;
	uint32 mPosition;
	float *mData; 
	float mDecay;
	uint32 lPosition;   // pointer for lowpass delay
	float *lData;		// lowpass delay
	float *alpFF;		// feedforward delay for allpass
	float alpFB;		// feedback for allpass
	uint32 alpFFPos;	// pointer for alpass FF delay
	float alpCoef;		// allpass coefficient

};
 
extern "C"
{
	void load(InterfaceTable *inTable);
	void PluckIn_next_notfull(PluckIn *unit, int inNumSamples);
	void PluckIn_next_full(PluckIn *unit, int inNumSamples);
	void PluckIn_Ctor(PluckIn* unit);
	void PluckIn_Dtor(PluckIn* unit);
};

//////////////////////////////////////////////////////////////////

void PluckIn_Ctor(PluckIn* unit)
{
	float d;
	
	SETCALC(PluckIn_next_notfull);

	unit->mDelayLength = (uint32)( SAMPLERATE/IN0(0));
	unit->mData = (float*)RTAlloc(unit->mWorld, unit->mDelayLength * sizeof(float));
	unit->mPosition = 0;
	unit->mDecay = IN0(1);
	unit->lData = (float*)RTAlloc(unit->mWorld, 2 * sizeof(float));
	unit->lPosition = 0;	

	d = (float)(SAMPLERATE/IN0(0)) - unit->mDelayLength; //fractional delay
	unit->alpCoef = (1.0 - d)/(1.0 + d);  // allpass coefficient
	//if(unit->alpCoef == 1) unit->alpCoef = .99;
	
	unit->alpFF = (float*)RTAlloc(unit->mWorld, 2 * sizeof(float));
	unit->alpFB = 0;
	unit->alpFFPos = 0;
	
	// empty out delay lines
	bzero(unit->mData, unit->mDelayLength*sizeof(uint32));	
	bzero(unit->lData, 2 * sizeof(float));	
	bzero(unit->alpFF, 2 * sizeof(float));
	
	PluckIn_next_notfull(unit, 1);
}

//////////////////////////////////////////////////////////////////
void PluckIn_Dtor(PluckIn* unit)
{
	RTFree(unit->mWorld, unit->mData);
	RTFree(unit->mWorld, unit->lData);
	RTFree(unit->mWorld, unit->alpFF);
}
//////////////////////////////////////////////////////////////////

void PluckIn_next_notfull(PluckIn *unit, int inNumSamples)
{
	float *out = OUT(0);
	

	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	float decay = unit->mDecay;
	float *ldata = unit->lData;
	uint32 lposition = unit->lPosition;
	
	float alpCoef =  unit->alpCoef;
	float *alpFF = unit->alpFF;
	uint32 alpFFPos = unit->alpFFPos;
	float alpFB = unit->alpFB;

	
	RGET
 	bool wrapped = false;
	
	for (int i=0; i < inNumSamples; ++i)
	{
 		float z ;
		// comb
		z = frand(s1,s2,s3) + (data[position] * decay); // fill one pitch period with noise
		// lowpass
		ldata[lposition++] = z;
		lposition = lposition%2;
 		z = 0.5 *(z + ldata[lposition] );
		
		// allpass
		alpFF[alpFFPos++] = z;
		alpFFPos = alpFFPos%2;
		z = alpFF[alpFFPos] + alpCoef * (z - alpFB);
		alpFB = z;

	 
	  	data[position] = z;
		if (++position >= length) {
			position = 0; 
			wrapped = true;  
	 		SETCALC(PluckIn_next_full);
		};
 		out[i] = z;
	}
	unit->mPosition = position;
	unit->lPosition = lposition;
	unit->alpFFPos = alpFFPos;
	unit->alpFB = alpFB;
	RPUT
}
	 
//////////////////////////////////////////////////////////////////

 void PluckIn_next_full(PluckIn *unit, int inNumSamples)
{
 	float *out = OUT(0);
		
 	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	float decay = unit->mDecay;
	
	float *ldata = unit->lData;
	uint32 lposition = unit->lPosition;
			
	float alpCoef =  unit->alpCoef;
	float *alpFF = unit->alpFF;
 	uint32 alpFFPos = unit->alpFFPos;
	float alpFB = unit->alpFB;
	
 	for (int i=0; i < inNumSamples; ++i)
	{
 		float z ;
		// comb
		 z = data[position] * decay;
 		
		// lowpass
		ldata[lposition++] = z;
		lposition = lposition%2;
 		z = 0.5 *(z + ldata[lposition] );

		// allpass
		alpFF[alpFFPos++] = z;
		alpFFPos = alpFFPos%2;
		z = alpFF[alpFFPos] + alpCoef * (z - alpFB);
		alpFB = z;

	  	data[position] = z;
		if (++position >= length) {
			position = 0;  
		};
 		out[i] = z;
	}

	unit->mPosition = position;
	unit->lPosition = lposition;
	unit->alpFFPos = alpFFPos;
	unit->alpFB = alpFB;
	
}

 
void load(InterfaceTable *inTable)
{
	ft = inTable;

	DefineDtorUnit(PluckIn);
}

