 
/*
Implementation of Perry Cook's Slide Flute Physical Model
Paul Lansky, 4/28/04
*/
/**************************** interface file

 
Sflute2 : UGen {
	*ar { arg   freq=440, pressure = 1, randamp = 0.04, dampcoef = 0.7,emb = 12, 
					jetstream = 0.2, fullwave = 0.9,  mul = 1.0, add = 0.0;
		^this.multiNew('audio', freq , pressure, randamp, dampcoef, emb, 
					jetstream, fullwave ).madd(mul, add)
	}
	*kr { arg   freq=440, pressure = 1,randamp = 0.04, dampcoef = 0.7,emb = 12, 
					jetstream = 0.2, fullwave = 0.9,  mul = 1.0, add = 0.0;
		^this.multiNew('control', freq, pressure, randamp, dampcoef, emb, 
					jetstream, fullwave ) .madd(mul, add)
	}
}

sample code

(
{LeakDC.ar 					// necessary because of natural DCoffset
   (Sflute2.ar(				// all arguments are modulateable			 
     freq:[FSinOsc.kr(5,0,(300-310)/2,(300+310)/2),305], // fmod from 300-310hz
     pressure:0.9, // effects loudness, but also timbre (e.g. blow harder) range:0.4-1
     randamp:0.081, // effects how much noise there is in the signal, range 0.01 - 0.9
     dampcoef:Line.kr(0.9,0.3,12), // effects lowpass at and of flute,
						//  also effects pitch somewhat, range 0-1
     emb: Line.kr(20,90,12),	// effects overblowing, (harmonics) range 1-99
     jetstream: 0.2,// percentage of jetstream (across mouth piece) to mix in,
     				// range 0.2-0.4
     fullwave: 0.9,		    // percentage of fullwave to mix in
     mul:0.6				// important number to watch, signal can easily clip
),0.99)}.play
)

***********************************/
#include "SC_PlugIn.h"

static InterfaceTable *ft;
#define RGET \
        RGen& rgen = *unit->mParent->mRGen; \
        uint32 s1 = rgen.s1; \
        uint32 s2 = rgen.s2; \
        uint32 s3 = rgen.s3; 

#define RPUT \
        rgen.s1 = s1; \
        rgen.s2 = s2; \
        rgen.s3 = s3;

 struct Sflute2 : public Unit
{
	uint32 delay1Length;
	float *delay1;
	float *delay2;
	float delay3;
	uint32 delay1Pos;
	uint32 delay2Pos;
	//bool lenwrapped;
	//bool lipwrapped;
};
 extern "C"
{
        void load(InterfaceTable *inTable);
        void Sflute2_next(Sflute2 *unit, int inNumSamples);
        void Sflute2_Ctor(Sflute2* unit);
        void Sflute2_Dtor(Sflute2* unit);
};

//////////////////////////////////////////////////////////////////

void Sflute2_Ctor(Sflute2* unit)
{
        
#define EMBLEN 200
#define FLUTELEN 900
        SETCALC(Sflute2_next);

	unit->delay1Length = (uint32)( SAMPLERATE/IN0(0));
	unit->delay1 = (float*)RTAlloc(unit->mWorld, FLUTELEN * sizeof(float));
	unit->delay1Pos = 0;
	unit->delay2 = (float*)RTAlloc(unit->mWorld, EMBLEN * sizeof(float)); // make larger to allow for change
	unit->delay2Pos = 0;
	unit->delay3 = 0;
	bzero(unit->delay1, FLUTELEN*sizeof(float));
	bzero(unit->delay2, EMBLEN*sizeof(float));
	Sflute2_next(unit, 1);
}

//////////////////////////////////////////////////////////////////

void Sflute2_Dtor(Sflute2* unit)
{
        RTFree(unit->mWorld, unit->delay1);
        RTFree(unit->mWorld, unit->delay2);
}
//////////////////////////////////////////////////////////////////

void Sflute2_next(Sflute2 *unit, int inNumSamples)
{
	float *out = OUT(0);

	uint32 delay1Length = unit->delay1Length;
	float *delay1 = unit->delay1;
	uint32 delay1Pos = unit->delay1Pos;
	uint32 delay2Length = (uint32)IN0(4);
	float *delay2 = unit->delay2 ;
	uint32 delay2Pos = unit->delay2Pos;
	float pressure = IN0(1) ;
	float randamp = IN0(2);
	float dampcoef = IN0(3);
	float delay3 = unit->delay3;
	float jetstream = IN0(5);
	float fullwave = IN0(6);
	//bool lipwrapped = unit->lipwrapped;
	//bool lenwrapped = unit->lenwrapped;
	delay1Length = (uint32)( SAMPLERATE/IN0(0));
		
	if(delay1Length > FLUTELEN) delay1Length = FLUTELEN;
	if(delay2Length > EMBLEN) delay2Length = EMBLEN;
 
	RGET
	for (int i=0; i < inNumSamples; ++i)
        {
		float y,lenval;
				
		y = (frand(s1,s2,s3)*2-1) * randamp * pressure; 
 
		y = y + pressure;	// breath pressure with random deviation
		lenval = delay1[delay1Pos];
 		y = y + (lenval * -.35);	// mix with pressure of returning wave
				
		delay2[delay2Pos] = y;
		if(++delay2Pos >= delay2Length) {
				delay2Pos = 0;
				//lipwrapped = true;
				};
 				
		y = delay2[delay2Pos];
 		y = (y * y * y) - y;	//transfer differential across mouth hole
		y = (jetstream * y) + (fullwave * lenval);   // jet stream plus full wave 
 				
		out[i] = y;	// catch output here
				
		y = (dampcoef * y) + (1.-dampcoef) * delay3; // lowpass at end of tube
 				
		delay3 = y;
           
		delay1[delay1Pos] = y;
                if (++delay1Pos >= delay1Length) {
                        delay1Pos = 0; 
						//lenwrapped = true;
                };
                 
        }
		unit->delay1Pos = delay1Pos;
		unit->delay2Pos = delay2Pos;
		unit->delay3 = delay3;
		//unit->lipwrapped = lipwrapped;
		//unit->lenwrapped = lenwrapped;
		RPUT
}
         
//////////////////////////////////////////////////////////////////

 
void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(Sflute2);
}
