#include "SC_PlugIn.h"
#include "BeeThree.h"
#include "Stk.h"
#include "STKAlloc.h"
 
static InterfaceTable *ft;
 struct StkBeeThree : public Unit
{
	BeeThree *beethree;
	float op4gain;
	float op3gain;
	float lfospeed;
	float lfodepth;
	float adsrtarget;
	float trig;
 

	};

extern "C"
{
        void load(InterfaceTable *inTable);
        void StkBeeThree_next(StkBeeThree *unit, int inNumSamples);
        void StkBeeThree_Ctor(StkBeeThree* unit);
        void StkBeeThree_Dtor(StkBeeThree* unit);
};

//////////////////////////////////////////////////////////////////

void StkBeeThree_Ctor(StkBeeThree* unit)
{
/********
 
    Control Change Numbers: 
       - Operator 4 (feedback) Gain = 2
       - Operator 3 Gain = 4
       - LFO Speed = 11
       - LFO Depth = 1
       - ADSR 2 & 4 Target = 128

*****/
  
	STKAlloc0(unit->mWorld,  unit->beethree, BeeThree);

	unit->beethree->controlChange((int)2,unit->op4gain=IN0(1));// operator4gain
	unit->beethree->controlChange((int)4,unit->op3gain=IN0(2));// operator3gain
	unit->beethree->controlChange((int)11,unit->lfospeed=IN0(3)); // lfospeed
	unit->beethree->controlChange((int)1,unit->lfodepth=IN0(4)); // lfodepth
	unit->beethree->controlChange((int)128,unit->adsrtarget=IN0(5)); // adsrtarget
 
 	unit->beethree->noteOn(IN0(0) ,1);
	unit->trig = IN0(6);
 	SETCALC(StkBeeThree_next);
	StkBeeThree_next(unit, 1);
}
  
void StkBeeThree_next(StkBeeThree *unit, int inNumSamples)
{
	float *out=OUT(0);
	float in00=IN0(0);
	float in01=IN0(1);
	float in02=IN0(2);
	float in03=IN0(3);
	float in04=IN0(4);
	float in05=IN0(5);
 
 	if(IN0(6) > 0) { 
		if(unit->trig < 0) {  
 				unit->beethree->noteOff( 0);
			    unit->beethree->noteOn(IN0(0) ,1);
			
	}; 
			};
	unit->trig = in05;
 
	if(in00 != unit->op4gain) unit->beethree->setFrequency(unit->op4gain = in00);	
	if(in01 != unit->op3gain) unit->beethree->controlChange(2,unit->op3gain =  in01); 	
	if(in02 != unit->lfospeed) unit->beethree->controlChange(4,unit->lfospeed = in02);
	if(in03 != unit->lfodepth) unit->beethree->controlChange(11,unit->lfodepth =  in03);		
	if(in04 != unit->adsrtarget) unit->beethree->controlChange(1, unit->adsrtarget = in04);
		   		
 	for (int i=0; i < inNumSamples; ++i)
	{
		out[i] = unit->beethree->tick();
        }
 }
 
void StkBeeThree_Dtor(StkBeeThree* unit)
{
	//unit->BeeThree->clear();
	//delete unit->BeeThree;
	unit->beethree->~BeeThree();
	RTFree(unit->mWorld, unit->beethree);
}

void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkBeeThree); 
}
 