#include "SC_PlugIn.h"
#include "/Users/paul/stk-4.1.3/include/SKINI.h"
#include "/Users/paul/stk-4.1.3/include/SKINI.tbl"
#include "STKAlloc.h"
#include "BlowHole.h"
 
static InterfaceTable *ft;
/*
       - Reed Stiffness = 2
       - Noise Gain = 4
       - Tonehole State = 11
       - Register State = 1
       - Breath Pressure = 128
*/
  struct StkBlowHole : public Unit
{
	BlowHole *blowhole;
	float freq;
	float reedstiffness;
	float noisegain;
	float tonehole;
	float rgister;
	float breathpressure;
	float trig;
};
 
extern "C"
{
        void load(InterfaceTable *inTable);
        void StkBlowHole_next(StkBlowHole *unit, int inNumSamples);
        void StkBlowHole_Ctor(StkBlowHole* unit);
        void StkBlowHole_Dtor(StkBlowHole* unit);
};

//////////////////////////////////////////////////////////////////

void StkBlowHole_Ctor(StkBlowHole* unit)
{
 	STKAlloc(unit->mWorld, unit->blowhole, BlowHole, 40);

	unit->freq = unit->reedstiffness = unit->noisegain = unit->tonehole = unit->rgister = unit->breathpressure = 0;
	unit->trig = 1;
	unit->blowhole->noteOn(IN0(0) ,1);
	SETCALC(StkBlowHole_next);
	StkBlowHole_next(unit, 1);
 
}

void StkBlowHole_next(StkBlowHole *unit, int inNumSamples)
{
	float *out OUT(0);
	float in00=IN0(0);
	float in01=IN0(1);
	float in02=IN0(2);
	float in03=IN0(3);
	float in04=IN0(4);
	float in05=IN0(5);

  	if(IN0(6) > 0) { 
		if(unit->trig < 0) {
				unit->blowhole->noteOff( 0);
			    unit->blowhole->noteOn(IN0(0) ,1);
				};
			};
	unit->trig = IN0(6);

 
	if(in00 != unit->freq) {
		unit->blowhole->setFrequency(in00);  
		unit->freq = in00;};
	
	if(in01 != unit->reedstiffness) {
		unit->blowhole->controlChange(2, in01);
		unit->reedstiffness = in01; };
 	
	if(in02 != unit->noisegain) {
		unit->blowhole->controlChange(4, in02);
		unit->noisegain = in02; };

	if(in03 != unit->tonehole) {
		unit->blowhole->controlChange(11, in03);
		unit->tonehole = in03; };
		
	if(in04 != unit->rgister) {
		unit->blowhole->controlChange(1, in04);
		unit->rgister = in04; };
		
	if(in04 != unit->breathpressure) {
		unit->blowhole->controlChange(128, in05);
		unit->breathpressure = in05; };

	
 	for (int i=0; i < inNumSamples; ++i)
	{
		out[i] = unit->blowhole->tick();
        }

}

void StkBlowHole_Dtor(StkBlowHole* unit)
{
	//delete unit->blowhole;
	unit->blowhole->~BlowHole();
    RTFree(unit->mWorld, unit->blowhole);

}
 
void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkBlowHole);
}
