#include "SC_PlugIn.h"
#include "/Users/paul/stk-4.1.3/include/SKINI.h"
#include "/Users/paul/stk-4.1.3/include/SKINI.tbl"
#include "STKAlloc.h"
#include "Clarinet.h"
 
static InterfaceTable *ft;
/*
    Control Change Numbers: 
       - Reed Stiffness = 2
       - Noise Gain = 4
       - Vibrato Frequency = 11
       - Vibrato Gain = 1
       - Breath Pressure = 128
*/
  struct StkClarinet : public Unit
{
	Clarinet *clarinet;
	float freq;
	float reedstiffness;
	float noisegain;
	float vibfreq;
	float vibgain;
	float breathpressure;
	float trig;
};
 
extern "C"
{
        void load(InterfaceTable *inTable);
        void StkClarinet_next(StkClarinet *unit, int inNumSamples);
        void StkClarinet_Ctor(StkClarinet* unit);
        void StkClarinet_Dtor(StkClarinet* unit);
};

//////////////////////////////////////////////////////////////////

void StkClarinet_Ctor(StkClarinet* unit)
{
 	STKAlloc(unit->mWorld, unit->clarinet, Clarinet, 40);

	
 	unit->clarinet->controlChange(2, unit->reedstiffness = IN0(1));
	unit->clarinet->controlChange(4, unit->noisegain = IN0(2));
	unit->clarinet->controlChange(11,unit->vibfreq = IN0(3));
	unit->clarinet->controlChange(1, unit->vibgain = IN0(4));
	unit->clarinet->controlChange(128, unit->breathpressure = IN0(5));
	unit->clarinet->noteOn(IN0(0) ,1);
	SETCALC(StkClarinet_next);
	unit->trig = 1;

	StkClarinet_next(unit, 1);
}

void StkClarinet_next(StkClarinet *unit, int inNumSamples)
{
	float *out=OUT(0);
	float in00=IN0(0);
	float in01=IN0(1);
	float in02=IN0(2);
	float in03=IN0(3);
	float in04=IN0(4);
	float in05=IN0(5);
  
 	if(IN0(6) > 0) { 
		if(unit->trig < 0) {  
 				unit->clarinet->noteOff(0);
 			    unit->clarinet->noteOn(IN0(0) ,1);
	}; 
			};
	unit->trig = IN0(6);
 	
	if(in00 != unit->freq) { 
		unit->clarinet->setFrequency(in00);
		unit->freq = in00; };
 	
	if(in01 != unit->reedstiffness) { 
		unit->clarinet->controlChange(2, in01);
		unit->reedstiffness = in01; };
 	
	if(in02 != unit->noisegain) { 
		unit->clarinet->controlChange(4, in02);
		unit->noisegain = in02; };

	if(in03 != unit->vibfreq) { 
		unit->clarinet->controlChange(11, in03);
		unit->vibfreq = in03; };
		
	if(in04 != unit->vibgain) { 
		unit->clarinet->controlChange(1, in04);
		unit->vibgain = in04; };
		
	if(in05 != unit->breathpressure) { 
		unit->clarinet->controlChange(128, in05);
		unit->breathpressure = in05; };

 	for (int i=0; i < inNumSamples; ++i)
	{
		out[i] = unit->clarinet->tick();
        }

}

void StkClarinet_Dtor(StkClarinet* unit)
{
	//delete unit->clarinet;
	unit->clarinet->~Clarinet();
    RTFree(unit->mWorld, unit->clarinet);

}
 
void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkClarinet);
}
