
 
#include "SC_PlugIn.h"
#include "OneZero.h"
#include "OnePole.h"
#include "DelayA.h"
#include "DelayL.h"
#include "JetTabl.h"
#include "PoleZero.h"
#include "ADSR.h"
#include "WaveLoop.h"
#include "Noise.h"
#include "Stk.h"
#include "Flute.h"

static InterfaceTable *ft;


  struct StkFlute : public Unit
{
  long length;
  DelayL *boreDelay;
  DelayL *jetDelay;
  JetTabl *jetTable;
  OnePole *filter;
  PoleZero *dcBlock;
  Noise *noise;
  ADSR *adsr;
  float endReflection;
  float jetReflection ;
  float noiseGain;           
  float jetRatio ;
  float maxPressure ;
  float lastFrequency;
  float outputGain;
  float frequency;


};
 
extern "C"
{
        void load(InterfaceTable *inTable);
        void StkFlute_next_notfull(StkFlute *unit, int inNumSamples);
        void StkFlute_Ctor(StkFlute* unit);
        void StkFlute_Dtor(StkFlute* unit);
};

//////////////////////////////////////////////////////////////////



void StkFlute_Ctor(StkFlute* unit)
{
  unit->length = (long) (SAMPLERATE/ 60 + 1);
  unit-> boreDelay = new DelayL( 100.0, unit->length );
  unit->length >>= 1;
  unit->jetDelay = new DelayL( 49.0, unit->length );
  unit->jetTable = new JetTabl();
  unit->filter = new OnePole();
  unit->dcBlock = new PoleZero();
  unit->dcBlock->setBlockZero();
  unit->noise = new Noise();
  unit->adsr = new ADSR();
 
  unit->filter->setPole( 0.7 - ((MY_FLOAT) 0.1 * 22050.0 / SAMPLERATE ) );
  unit->filter->setGain( -1.0 );
  unit->adsr->setAllTimes( 0.005, 0.01, 0.8, 0.010);
  unit->endReflection = (MY_FLOAT) 0.5;
  unit->jetReflection = (MY_FLOAT) 0.5;
  unit->noiseGain =  0.15;             // Breath pressure random component.
  unit->jetRatio = (MY_FLOAT) 0.32;
  //unit->jetRatio = (MY_FLOAT) IN0(1);

  unit->maxPressure = (MY_FLOAT) 0.0;

  unit->frequency = IN0(0);
  unit->lastFrequency = unit->frequency;
  unit->outputGain = 1;
  
  
  
  unit->lastFrequency *= 0.66666;
  // Delay = length - approximate filter delay.
  MY_FLOAT delay = SAMPLERATE / unit->lastFrequency - (MY_FLOAT) 2.0;
  if (delay <= 0.0) delay = 0.3;
  else if (delay > unit->length) delay = unit->length;

  unit->boreDelay->setDelay(delay);
  unit->jetDelay->setDelay(delay * unit->jetRatio);
  MY_FLOAT amplitude = 1;
  amplitude = 1.1 + (amplitude * 0.20);
   //startBlowing( 1.1 + (amplitude * 0.20), amplitude * 0.02);
  unit->adsr->setAttackRate(amplitude * 0.02);
  unit->maxPressure = amplitude / (MY_FLOAT) 0.8;
  unit->adsr->keyOn();
    
		SETCALC(StkFlute_next_notfull);
		StkFlute_next_notfull(unit, 1);
 
}

void StkFlute_next_notfull(StkFlute *unit, int inNumSamples)
{
      float *out = OUT(0);
	MY_FLOAT pressureDiff;
	MY_FLOAT breathPressure;
	MY_FLOAT temp,lastOutput;
	float frequency = IN0(0);
    unit->lastFrequency = frequency;
    unit->lastFrequency *= 0.66666;
    //MY_FLOAT delay = SAMPLERATE / unit->lastFrequency - (MY_FLOAT) 2.0;
    MY_FLOAT delay = SAMPLERATE / IN0(0) - (MY_FLOAT) 2.0;
    if (delay <= 0.0) delay = 0.3;
      else if (delay > unit->length) delay = unit->length;

   //unit->boreDelay->setDelay(delay);
   //unit->jetDelay->setDelay(delay * IN0(1));
	
	for (int i=0; i < inNumSamples; ++i)
	{

  // Calculate the breath pressure (envelope + noise)
  breathPressure = unit->maxPressure * unit->adsr->tick();
  breathPressure += breathPressure * unit->noiseGain * unit->noise->tick();
 
  temp = unit->filter->tick( unit->boreDelay->lastOut() );
  temp = unit->dcBlock->tick(temp); // Block DC on reflection.
  pressureDiff = breathPressure - (unit->jetReflection * temp);
  pressureDiff = unit->jetDelay->tick( pressureDiff );
  pressureDiff = unit->jetTable->tick( pressureDiff ) + (unit->endReflection * temp);
  lastOutput = (MY_FLOAT) 0.3 *unit->boreDelay->tick( pressureDiff );
 
  lastOutput *= unit->outputGain;
  out[i] = lastOutput;
        }
}

void StkFlute_Dtor(StkFlute* unit)
{

}
 
void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkFlute);
}
