#include "SC_PlugIn.h"
#include "Moog.h"
#include "Stk.h"
#include "STKAlloc.h"
 
static InterfaceTable *ft;
 struct StkMoog : public Unit
{
	Moog *moog;
	float filterQ;
	float sweeprate;
	float vibfreq;
	float vibgain;
	float gain;
	float freq;
	float trig;

	};

extern "C"
{
        void load(InterfaceTable *inTable);
        void StkMoog_next(StkMoog *unit, int inNumSamples);
        void StkMoog_Ctor(StkMoog* unit);
        void StkMoog_Dtor(StkMoog* unit);
};

//////////////////////////////////////////////////////////////////

void StkMoog_Ctor(StkMoog* unit)
{
/********
    Control Change Numbers: 
       - Filter Q = 2
       - Filter Sweep Rate = 4
       - Vibrato Frequency = 11
       - Vibrato Gain = 1
       - Gain = 128
*****/
  
	STKAlloc0(unit->mWorld,  unit->moog, Moog);

	unit->moog->controlChange((int)2, unit->filterQ = IN0(1));// filter Q 
	unit->moog->controlChange((int)4, unit->sweeprate = IN0(2));// sweep rate
	unit->moog->controlChange((int)11, unit->vibfreq = IN0(3)); // vibrato freq
	unit->moog->controlChange((int)1, unit->vibgain = IN0(4)); // vibrato gain
	unit->moog->controlChange((int)128, unit->gain = IN0(5)); // gain
 
 	unit->moog->noteOn(unit->freq = IN0(0) ,1);
 	SETCALC(StkMoog_next);
	unit->trig = 1;
	StkMoog_next(unit, 1);
}
  
void StkMoog_next(StkMoog *unit, int inNumSamples)
{
	float *out=OUT(0);
	float in00=IN0(0);
	float in01=IN0(1);
	float in02=IN0(2);
	float in03=IN0(3);
	float in04=IN0(4);
	float in05=IN0(5);
  
 	if(IN0(6) > 0) { 
		if(unit->trig < 0) {  
 				unit->moog->noteOff( 0);
			    unit->moog->noteOn(IN0(0) ,1);
			
	}; 
			};
	unit->trig = IN0(6);
 
	if(in00 != unit->freq) unit->moog->setFrequency(unit->freq = in00);
	
	if(in01 != unit->filterQ) unit->moog->controlChange(2,unit->filterQ =  in01);
 	
	if(in02 != unit->sweeprate) unit->moog->controlChange(4,unit->sweeprate = in02);

	if(in03 != unit->vibfreq) unit->moog->controlChange(11,unit->vibfreq =  in03);
		
	if(in04 != unit->vibgain) unit->moog->controlChange(1, unit->vibgain = in04);
		
	if(in05 != unit->gain) unit->moog->controlChange(128,unit->gain = in05);
 

 		
 	for (int i=0; i < inNumSamples; ++i)
	{
		out[i] = unit->moog->tick();
        }
 }
 
void StkMoog_Dtor(StkMoog* unit)
{
	//unit->Moog->clear();
	//delete unit->Moog;
	unit->moog->~Moog();
	RTFree(unit->mWorld, unit->moog);
}

void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkMoog); 
}
 