
 
#include "SC_PlugIn.h"
#include "OneZero.h"
#include "OnePole.h"
#include "DelayA.h"
#include "Noise.h"
#include "Stk.h"

static InterfaceTable *ft;


  struct StkPluck : public Unit
{
                uint32 length;
                float  loopGain;
                float  gain;
                DelayA *delayLine;
                OneZero *loopFilter;
                OnePole *pickFilter;
                Noise  *noise;

};
 
extern "C"
{
        void load(InterfaceTable *inTable);
        void StkPluck_next_notfull(StkPluck *unit, int inNumSamples);
        void StkPluck_Ctor(StkPluck* unit);
        void StkPluck_Dtor(StkPluck* unit);
};

//////////////////////////////////////////////////////////////////



void StkPluck_Ctor(StkPluck* unit)
{

        SETCALC(StkPluck_next_notfull);

                unit->pickFilter = new OnePole;
                unit->delayLine = new DelayA;
                unit->loopFilter = new OneZero;
                unit->noise = new Noise;
                unit->delayLine->clear();
                unit->loopFilter->clear();
                unit->pickFilter->clear();

                // set lowest frequency
                unit->length = (long) (SAMPLERATE/ 60 + 1);

                // set frequency
                MY_FLOAT delay = (SAMPLERATE/ IN0(0)) - (MY_FLOAT) 0.5;

                if (delay <= 0.0) delay = 0.3;
                        else if (delay > unit->length) delay = unit->length;
                unit->delayLine->setDelay(delay);
                unit->loopGain = IN0(1) + (IN0(0) * 0.000005);
 
                if ( unit->loopGain >= 1.0 ) unit->loopGain = (MY_FLOAT) 0.99999; 
 
                MY_FLOAT gain = 0.99;
                // set noiseburst
                unit->pickFilter->setPole((MY_FLOAT) 0.999 - (gain * (MY_FLOAT) 0.15));
                unit->pickFilter->setGain(gain * (MY_FLOAT) 0.5);
                for (uint32 i=0; i< unit->length; i++)
                        // Fill delay with noise additively with current contents.
                        unit->delayLine->tick( 0.6 * unit->delayLine->lastOut() + unit->pickFilter->tick( unit->noise->ti
ck() ) );
 
                StkPluck_next_notfull(unit, 1);
 
}

void StkPluck_next_notfull(StkPluck *unit, int inNumSamples)
{
      float *out = OUT(0);
          
        for (int i=0; i < inNumSamples; ++i)
        {
        float z,y;
        z = unit->delayLine->tick( 
                unit->loopFilter->tick(y= unit->delayLine->lastOut() * unit->loopGain ) ); 
 
           out[i] = z;
        }
} 

void StkPluck_Dtor(StkPluck* unit)
{
                delete unit->delayLine;
                delete unit->loopFilter;
                delete unit->pickFilter;
                delete unit->noise;

}
 
void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkPluck);
}
