#include "SC_PlugIn.h"
#include "Saxofony.h"
#include "Stk.h"
#include "STKAlloc.h"

static InterfaceTable *ft;
/*
    Control Change Numbers: 
       - Reed Stiffness = 2
       - Reed Aperture = 26
       - Noise Gain = 4
       - Blow Position = 11
       - Vibrato Frequency = 29
       - Vibrato Gain = 1
       - Breath Pressure = 128
*/
struct StkSaxofony : public Unit
{
	Saxofony *saxofony;
	float frequency;
	float reedstiffness;
	float reedaperture;
	float noisegain;
	float blowposition;
	float vibratofrequency;
	float vibratogain;
	float breathpressure;
	float trig;
	};

extern "C"
{
        void load(InterfaceTable *inTable);
        void StkSaxofony_next(StkSaxofony *unit, int inNumSamples);
        void StkSaxofony_Ctor(StkSaxofony* unit);
        void StkSaxofony_Dtor(StkSaxofony* unit);
};

//////////////////////////////////////////////////////////////////

void StkSaxofony_Ctor(StkSaxofony* unit)
{
	//Stk :: setRawwavePath("/Users/paul/stk-4.1.3/rawwaves");
	STKAlloc(unit->mWorld,unit->saxofony, Saxofony, 40);

	//unit->saxofony = new Saxofony((MY_FLOAT) 40); // 40 is lowest frequency
	unit->reedstiffness=unit->frequency=unit->reedaperture=unit->noisegain=
		unit->blowposition=unit->vibratofrequency=unit->vibratogain=unit->breathpressure=0;
	unit->saxofony->noteOn(IN0(0) ,1);
 	SETCALC(StkSaxofony_next);
	StkSaxofony_next(unit, 1);
	unit->trig = 1;
}

void StkSaxofony_next(StkSaxofony *unit, int inNumSamples)
{
	float *out OUT(0);
 
	float in00=IN0(0);
	float in01=IN0(1);
	float in02=IN0(2);
	float in03=IN0(3);
	float in04=IN0(4);
	float in05=IN0(5);
 	float in06=IN0(6);
 	float in07=IN0(7);

 	if(IN0(8) > 0) { 
		if(unit->trig < 0) {  
			unit->saxofony->noteOff( 0);
			unit->saxofony->noteOn(IN0(0) ,1);
			
		};
	};
	unit->trig = IN0(8);
	
	if(in00 != unit->frequency) {
		unit->saxofony->setFrequency(in00);
		unit->frequency = in00;};
	 
	if(in01 != unit->reedstiffness) {
		unit->saxofony->controlChange(2, in01);
		unit->reedstiffness = in01; };
	
	if(in02 != unit->reedaperture) {
		unit->saxofony->controlChange(26, in02);
		unit->reedaperture = in02; };
	
	if(in03 != unit->noisegain) {
		unit->saxofony->controlChange(4, in03);
		unit->noisegain = in03; };
		
	if(in04 != unit->blowposition) {
		unit->saxofony->controlChange(11, in04);
		unit->blowposition = in04; };
 
	if(in05 != unit->vibratofrequency) {
		unit->saxofony->controlChange(29, in05);
		unit->vibratofrequency = in05; };

	if(in06 != unit->vibratogain) {
		unit->saxofony->controlChange(1, in06);
		unit->vibratogain = in06; };
  		
	if(in07 != unit->breathpressure) {
		unit->saxofony->controlChange(128, in07);
		unit->breathpressure = in07; } ;

 		
 	for (int i=0; i < inNumSamples; ++i)
	{
		out[i] = unit->saxofony->tick();
        }
 }

void StkSaxofony_Dtor(StkSaxofony* unit)
{
	//unit->saxofony->clear();
	//delete unit->saxofony;	    
    unit->saxofony->~Saxofony();
    RTFree(unit->mWorld, unit->saxofony);

}

void load(InterfaceTable *inTable)
{
        ft = inTable;
        DefineDtorUnit(StkSaxofony); 
}
 